﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace DelegateExample
{
    // Declaration
    public delegate void myDelegate(string s);

    class myClass
    {
        public void myMethod1(string s)
        {
            Console.WriteLine(s + " from instance method1\n");
        }

        public static void myMethod2(string s)
        {
            Console.WriteLine(s + " from static method2\n");
        }
    } 

    class Program
    {
        static void Main(string[] args)
        {
            //Example illustrating different ways to work with 
            //delegate objects

            //(1) Working with instance methods.

            //Instantiate a null delegate object called delObj1.
            myDelegate delObj1 = null; //Delegate Instantiation

            //Instantiate an object called myObj from myClass 
            //class.
            myClass myObj = new myClass();

            //Assign instance method myMethod1 to delegate 
            //object delObj1. 
            delObj1 = myObj.myMethod1;

            //Use the delegate object delObj1 to pass some data 
            //to myMethod1.
            delObj1("Hello");  //Invocation

            //(2) Working with static methods.

            //Instantiate a null delegate object called delObj2.
            myDelegate delObj2 = null; //Delegate Instantiation

            //Assign static method myMethod2 to delegate object 
            //delObj2. Note that because myMethod2 is a static  
            //method, there is no need to first instantiate an  
            //object from myClass in order to then assign 
            //myMethod2 to delegage object delObj2.
            delObj2 = myClass.myMethod2;

            //Use the delegate object delObj2 to pass some 
            //data to myMethod2
            delObj2("Greetings");  //Invocation              

            //(3) Working with instance and static methods to  
            //build an invocation list

            //Create two delegate objects and directly assign 
            //them to their respective instance and static methods.

            //Delegate instantiation 
            myDelegate delObj3 = new myDelegate(myObj.myMethod1);
            //and invocation
            delObj3("Building invocation list, element 1");

            //Delegate instantiation 
            myDelegate delObj4 = new myDelegate(myClass.myMethod2);
            //and invocation
            delObj4("Building invocation list, element 2");

            //Build an invocation list from these two 
            //delegate objects
            myDelegate delObjTotal = delObj3 + delObj4;

            //Process the invocation list with some data
            delObjTotal("Processing entire invocation list.");

            //Remove a method from invocation list
            delObjTotal -= delObj4;

            //Process remaining methods in invocation list
            delObjTotal("Last message is");

            //Alternate way to build invocation list
            myDelegate delObj5 = null;
            delObj5 += new myDelegate(myObj.myMethod1);
            delObj5 += new myDelegate(myClass.myMethod2);

            //Process total invocation list with some data
            delObj5("bye...");

            //Remove a method from invocation list
            delObj5 -= myClass.myMethod2;

            //Process remaining methods in invocation list
            delObj5("Final message is");

            //Pause until user hits enter key
            Console.ReadLine(); 
        }
    }
}
