﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO;

namespace FileIOExample
{
    class Program
    {
    //Delegate to store references to some arbitrary 
    //function
    public delegate double Function(double x);        

    public class ConstructFunctionTable
    {
      public void MakeXYTable(Function f, double startValue, 
               double endValue, double increment, 
               ref List<double> xv, ref List<double> yv)
      {
        //Print table headings
        Console.WriteLine("x\tf(x)");
        Console.WriteLine("-------------");

        //Loop from the start to the end values of the   
        //(x,f(x)) table incrementing x with your value  
        //of choice. Calculate the corresponding function 
        //f(x) for each x value. Display results on the 
        //screen for reference.
        for (double x=startValue; x<=endValue; x+=increment)
        {
          xv.Add(Convert.ToDouble(x));
          yv.Add(Convert.ToDouble(f(x)));
          Console.WriteLine(String.Format("{0:f}\t{1:f}",
                                                x,f(x)));
        }
        Console.WriteLine("-------------");
      }

      public void SaveXYTable(string pathNfilename, ref List<double> xv, ref List<double> yv)
      {
        //Create a FileStream object and a StreamWriter object  
        //to write to a file specified by the variable 
        //pathNfilename.                
        FileStream fsWrite = new FileStream(pathNfilename, FileMode.Create, FileAccess.Write);
        StreamWriter sWriter = new StreamWriter(fsWrite);

        //Specify the available data delimeter characters
        char[] fileDelim = { ',', '|', '\t' };

        //Find out how many (x,y) data points you have. 
        int xyCount = xv.Count;

        //Loop through all the data points you have writing 
        //each of them to the text file referenced to by the 
        //StreamWriter object. 
        for (int i = 0; i < xyCount; i++)
	{
          sWriter.WriteLine(String.Format("{0:f}" + fileDelim[2] + "{1:f}", xv[i], yv[i]));
	}
        //Close the file StreamWriter object.
        sWriter.Close();
      }

      public void ReadXYTable(string pathNfilename, ref List<double> xv, ref List<double> yv)
      {
        //Create a FileStream object and a StreamReader object  
        //to read from a file specified by the variable 
        //pathNfilename. 
        FileStream fsRead = new FileStream(pathNfilename, FileMode.Open, FileAccess.Read);
        StreamReader sReader = new StreamReader(fsRead);

        //Specify the available data delimeter characters
        char[] fileDelim = { ',', '|', '\t' };

        //Declare a string array to hold incoming data read 
        //from file.
        string[] fields;

        //Read the first line of the file. 
        string xyDataline = sReader.ReadLine();

        //If the first line of the file is not null, then 
        //process data otherwise close the file StreamReader 
        //object.
        while (xyDataline != null)
        {
          //Using the delimeter of choice, split the incoming
          //data into two fields: one for the x values and the
          //other for the y values. 
          fields = xyDataline.Split(fileDelim[2]);

          //Store each value in its corresponding array.
          xv.Add(Convert.ToDouble(fields[0]));
          yv.Add(Convert.ToDouble(fields[1]));

          //Read another line of data from the file.
          xyDataline = sReader.ReadLine();
        }

        //Close the file StreamReader object.
        sReader.Close();
      }

      static void Main(string[] args)
      {
        //Give a name to some arbitrary text file to hold the 
        //(x,f(x)) data
        string fileName = "TestData.txt";

        //and assign it to the path of the current default 
        //directory
        string filePathNName = Directory.GetCurrentDirectory().ToString() + "\\" + fileName;
        //Note: The filename and path chosen here is arbitrary 
        //and these were selected only to illustrate the 
        //functionality of this example.

        //Declare a couple of generic collections of doubles. 
        //Collections were chosen instead of arrays because we 
        //do not know in advance how many data points we have. 
        //Unlike arrays, collections do not have a fixed size 
        //but can expand in size dynamically. Generics were  
        //chosen to allow a flexible choice of data types.
        List<double> xvalue = new List<double>();
        List<double> yvalue = new List<double>();

        //Create a function table object to run this example.
        ConstructFunctionTable fcntable = new ConstructFunctionTable();

        //Create an (x,y=f(x)) table of sines running from 0 
        //to 2, incrementing by 0.25 and return result inside two 
        //collections xvalue, and yvalue.
        Console.WriteLine("The original (x,y) data table:\n");                
        fcntable.MakeXYTable(Math.Sin, 0.0, 2.0, 0.25, ref xvalue, ref yvalue);

        //Write the data contained in collections xvalue and 
        //yvalue to the data text file described earlier.
        Console.Write("\nSaving data to file " + fileName + 
                      " please wait. ");
        fcntable.SaveXYTable(filePathNName, ref xvalue, ref yvalue);
        Console.WriteLine("Done!");

        //Clear out the collections xvalue and yvalue to 
        //receive fresh data.
        xvalue.Clear();
        yvalue.Clear();

        //Read the data contained in collections xvalue and 
        //yvalue from the data text file described earlier.
        Console.Write("\nReading data from file " + fileName + " please wait. ");
        fcntable.ReadXYTable(filePathNName, ref xvalue, ref yvalue);
        Console.WriteLine("Done!");
                
        //Display the data results just read on the screen
        Console.WriteLine("\nData just read from file " + fileName + "\n");
        Console.WriteLine("x\tf(x)");
        Console.WriteLine("-------------");
        int xydataCount = xvalue.Count;
        for (int i = 0; i < xydataCount; i++)
        {
          Console.WriteLine(String.Format("{0:f}\t{1:f}", 
                                  xvalue[i], yvalue[i]));
        }

        //Pause until user hits the ENTER key 
        Console.Write("\nPlease hit the ENTER key to terminate this program");
        Console.ReadLine();
      }
    }
  }
}

