﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace MinorMiscExamples
{
    class Program
    {
        static void swap<T>(ref T var1, ref T var2)
        {
            T temp;
            temp = var1;
            var1 = var2;
            var2 = temp;
        }

        public static void ChangeVarByValue(int x)
        {
           x = 0;
        }

        public static void ChangeVarByReference(ref int x)
        {
           x = 0;
        }

        public static void DoSum(int x, out int total)
        {
            Console.WriteLine("Adding 7 to the number {0}", x);
            total = x + 7;
            return;
        }

        public static bool AlmostEqualRelative(float x, float y, float maxRelativeError)
        {
           if (x == y)
              return true;
           float relativeError=Math.Abs(((x-y)/y));
           if (relativeError <= maxRelativeError)
              return true;
           return false;
        }

        public static bool AlmostEqualRelative2(float x, float y, float maxRelativeError)
        {
            if (x == Double.NaN || y == Double.NaN) 
                return false;     
            if (x == y)
                return true;
            float relativeError;
            if (Math.Abs(y) > Math.Abs(x))
                relativeError = Math.Abs((x - y) / y);
            else
                relativeError = Math.Abs((x - y) / x);
            if (relativeError <= maxRelativeError)
                return true;
            return false;
        }

        public static bool AlmostEqualRelativeOrAbsolute(float x, float y, float maxRelativeError, float maxAbsoluteError)
        {
            if (x == Double.NaN || y == Double.NaN) 
                return false;     
            if (Math.Abs(x - y) < maxAbsoluteError)
                return true;
            float relativeError=0.0f;
            if (Math.Abs(y) > Math.Abs(x))
                relativeError = Math.Abs((x - y) / y);
            else
                relativeError = Math.Abs((x - y) / x);
            if (relativeError <= maxRelativeError)
                return true;
            return false;
        }

        static void Main(string[] args)
        {
            Console.WriteLine("Testing the generic SWAP function\n");

            Console.Write("Type in an integer:");
            int i = int.Parse(Console.ReadLine());
            Console.WriteLine("You entered: i = {0}", i);

            Console.Write("Type in another integer:");
            int j = int.Parse(Console.ReadLine());
            Console.WriteLine("You entered: j = {0}", j);

            Console.WriteLine("Now swap them and display results\n");

            Console.WriteLine("Before swapping: i = {0} and j = {1}", i, j);
            swap<int>(ref i, ref j);
            Console.WriteLine("After swapping: i = {0} and j = {1}\n\n", i, j);
            
            i=0;
            Console.WriteLine("Testing i++ operation\n");
            Console.WriteLine("Before    : i = {0}", i);
            Console.WriteLine("After  i++: i = {0}\n\n", i++);

            i=0;
            Console.WriteLine("Testing ++i operation\n");
            Console.WriteLine("Before    : i = {0}", i);
            Console.WriteLine("After  ++i: i = {0}\n\n", ++i);

            Console.WriteLine("Testing foreach loop\n");
            string s = "Hello";
            foreach(char c in s)
                Console.Write("{0} ",c);

            Console.WriteLine("\n\nTesting arrays\n");
            int[] myInt1 = new int[3] {10,15,25};

            int[] myInt2 = {10,15,25};
            int[] numArray = { 1, 3, 5, 7, 9, 11, 13 };

            int[] myInt3 = new int[3];	
            myInt3[0] = 10;			
            myInt3[1] = 15;
            myInt3[2] = 25;

            for (int k = 0; k < myInt1.Length; k++)
                Console.WriteLine("myInt1[" + k.ToString() + "]={0}", myInt1[k].ToString());
            Console.WriteLine("\n\n");
            foreach (int m in myInt2)
            {
                Console.WriteLine(m.ToString());
            }
            Console.WriteLine("\n\n");
            for (int n = 0; n < myInt3.Length; n++)
                Console.WriteLine("myInt3[" + n.ToString() + "]={0}", myInt3[n].ToString());
            Console.WriteLine("\n\n");

            Console.WriteLine("Testing passing numbers by VALUE\n");
            int x = 10;
            Console.WriteLine("Before x = {0}", x);
            ChangeVarByValue(x);
            Console.WriteLine("After  x = {0}", x);

            Console.WriteLine("\n\nTesting passing numbers by REFERENCE\n");
            x = 10;
            Console.WriteLine("Before x = {0}", x);
            ChangeVarByReference(ref x);
            Console.WriteLine("After  x = {0}", x);

            Console.WriteLine("\n\nTesting passing numbers using the out keyword\n");
            x=5;
            int tot=0;
            Console.WriteLine("The original number is {0}", x);
            DoSum(x, out tot);
            Console.WriteLine("The final sum is {0}", tot);

            Console.WriteLine("\n\nTesting comparison of float numbers");
            float tenth = 0.1f;
            float one = 1f;
            float ten = 10f;

            //Goal: Compute 1 - 0.1*10
            Console.WriteLine("1 - (1/10)*10 = {0}", one - tenth * ten);
            //Expected output: 0
            //Actual output: 1.490116E-08 on 32-bit machines
            //Actual output: 0 on 64-bit machines

            //Goal: Compute 1.0 - 0.9
            Console.WriteLine("1 - 0.9 = {0}", one - 0.9f);
            //Expected output: 0.1
            //Actual output: 0.0999999999999996 on 32-bit machines
            //Actual output: 0.1 on 64-bit machines
            Console.ReadLine();
        }
    }
}
