﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace Chapter02_CodeExamples
{
    class Program
    {
        public static void PInE_Example()
        {
            Console.WriteLine("--- RUNNING Math.PI and Math.E EXAMPLE ---\n");
            Console.WriteLine("PI = {0}", Math.PI);
            Console.WriteLine("E  = {0}", Math.E);
            Console.WriteLine("\n------------------------------------------\n");
        }

        public static void MinMax_Example()
        {
            double x = 5.0, y = 10.0;

            Console.WriteLine("----- RUNNING Math.Min/Max EXAMPLE -----\n");
            Console.WriteLine("The smaller of {0} and {1} is {2}.", x, y, Math.Min(x, y));
            Console.WriteLine("The greater of {0} and {1} is {2}.", x, y, Math.Max(x, y));
            Console.WriteLine("\n----------------------------------------\n");
        }

        public static void PowExpLogLog10_Example()
        {
            double b = 2.75, x = 3.25, newBase = 8;

            Console.WriteLine("--- RUNNING Math.Pow/Exp/Log/Log10/Log(x,newBase) EXAMPLE ---\n");
            Console.WriteLine("Exp({0}) = {1}", x, Math.Exp(x));
            Console.WriteLine("Log({0}) = {1}\n", Math.Exp(x), Math.Log(Math.Exp(x)));

            Console.WriteLine("Pow({0},{1}) = {2}", b, x, Math.Pow(b, x));
            Console.WriteLine("Log10({0})/Log10({1}) = {2}\n", Math.Pow(b, x), b, Math.Log10(Math.Pow(b, x)) / Math.Log10(b));

            Console.WriteLine("Log{0}({1}) = {2}", newBase, x, Math.Log(x, newBase));
            Console.WriteLine("Pow({0},{1}) = {2}", newBase, Math.Log(x, newBase), Math.Pow(newBase, Math.Log(x, newBase)));
            Console.WriteLine("-------------------------------------------------------------\n");
        }

        public static void MultDivRem_Example()
        {
            int int1 = Int32.MaxValue;
            int int2 = Int32.MaxValue;
            int intResult;
            long longResult;
            double divisor, doubleResult;

            Console.WriteLine("\n--- RUNNING Math.BigMul EXAMPLE ---\n");
            longResult = Math.BigMul(int1, int2);
            Console.WriteLine("{0} * {1} = {2}\n", int1, int2, longResult);

            Console.WriteLine("\n--- RUNNING Math.DivRem EXAMPLE ---\n");
            intResult = Math.DivRem(int1, 2, out int2);
            Console.WriteLine("{0}/{1} = {2}, with a remainder of {3}.\n", int1, 2, intResult, int2);

            Console.WriteLine("\n--- RUNNING Math.IEEERemainder EXAMPLE ---\n");
            String str1 = "The IEEE remainder of {0:e}/{1:f} is {2:e}";
            divisor = 2.0;
            doubleResult = Math.IEEERemainder(Double.MaxValue, divisor);
            Console.WriteLine(str1, Double.MaxValue, divisor, doubleResult);

            divisor = 3.0;
            doubleResult = Math.IEEERemainder(Double.MaxValue, divisor);
            Console.WriteLine(str1, Double.MaxValue, divisor, doubleResult);
            Console.WriteLine("Note that two positive numbers can yield a negative remainder.");
            Console.WriteLine("\n--------------------------------------------\n");
        }

        public static void Misc_Examples()
        {
            double x = 5.0;
            double y = 0.0;
            double z = -5.0;

            Console.WriteLine("\n--- RUNNING Math.Sign EXAMPLE ---\n");
            Console.WriteLine("The sign of {0} is {1}", x, Math.Sign(x));
            Console.WriteLine("The sign of {0} is {1}", y, Math.Sign(y));
            Console.WriteLine("The sign of {0} is {1}", z, Math.Sign(z));
            Console.WriteLine("\n--------------------------------------------\n");

            Console.WriteLine("\n--- RUNNING Math.Abs EXAMPLE ----\n");
            Console.WriteLine("Before: {0,-5} After: {1,-5}", z, Math.Abs(z));
            Console.WriteLine("\n--------------------------------------------\n");

            Console.WriteLine("\n--- RUNNING Math.Sqrt EXAMPLE ---\n");
            Console.WriteLine("The square root of {0} = {1}", x, Math.Sqrt(x));
            Console.WriteLine("The square root of {0} = {1}", y, Math.Sqrt(y));
            Console.WriteLine("The square root of {0} = {1}", z, Math.Sqrt(z));
            Console.WriteLine("\n--------------------------------------------\n");
        }

        public static double DMStoDD(double DMSDeg, double DMSMin, double DMSSec)
        {
            //Converts an angle expressed in the DMS (Degree-Minute-Second)
            //format to its equivalent value in the DD (Degree-Decimal) format.            
            double DD = DMSDeg + (DMSMin / 60.0) + (DMSSec / 3600.0);
            return DD;
        }

        public static void DDtoDMS(double DD, out double d,
             out double m, out double s, out string strDMS)
        {
            //Converts an angle expressed in the DD (Degree-Decimal) format
            //to its equivalent value in the DMS (Degree-Minute-Second) format.
            //Output is both numerical format: degree/minute/second and a 
            //string in the format: xx°xx'xx" 

            //Extract the degree component
            double deg = Math.Floor(DD);
            DD -= deg;
            DD *= 60;
            //Extract the minute component
            double min = Math.Floor(DD);
            DD -= min;
            DD *= 60;
            //Extract the second component
            double sec = Math.Round(DD);

            d = deg;
            m = min;
            s = sec;

            //Create padding character
            char pad;
            char.TryParse("0", out pad);

            //Create degree/minute/second strings
            string str_deg = deg.ToString();
            string str_min = min.ToString().PadLeft(2, pad);
            string str_sec = sec.ToString().PadLeft(2, pad);

            //Append degree/minute/second strings together            
            strDMS = string.Format("{0}° {1}' {2}\"", str_deg, str_min, str_sec);
        }

        public static double DegreesToRadians(double degrees)
        {
            return (Math.PI / 180.0) * degrees;
        }

        public static double RadiansToDegrees(double radians)
        {
            return (180.0 / Math.PI) * radians;
        }

        public static double GradsToRadians(double grads)
        {
            return (grads / 200.0) * Math.PI;
        }

        public static double RadiansToGrads(double radians)
        {
            return (radians / Math.PI) * 200.0;
        }

        public static double DegreesToGrads(double degrees)
        {
            return (degrees / 9.0) * 10.0;
        }

        public static double GradsToDegrees(double grads)
        {
            return (grads / 10.0) * 9.0;
        }

        public static void AngleMeas_Examples()
        {
            Console.WriteLine("\n--- RUNNING ANGLE UNITS EXAMPLE ---\n");

            double xdegs = 45.0;
            double xrads = DegreesToRadians(xdegs);

            //Console.WriteLine("{0}° = {1} radians.", xdegs, xrads);
            Console.WriteLine("{0}\xb0 = {1} radians.", xdegs, xrads);
            Console.WriteLine("{0} radians = {1}°.\n", xrads, RadiansToDegrees(xrads));

            double xdd = DMStoDD(5.0, 23.0, 55.0);
            Console.WriteLine("5D 23M 55S = {0}°", xdd);

            string xdmsstr;
            double xdeg, xmin, xsec;
            DDtoDMS(xdd, out xdeg, out xmin, out xsec, out xdmsstr);
            Console.WriteLine("{0}° = {1}D {2}M {3}S = {4}\n", xdd, xdeg, xmin, xsec, xdmsstr);

            double xgrads = 100.0;
            double xrads2 = GradsToRadians(xgrads);
            Console.WriteLine("{0} grads = {1} radians.", xgrads, xrads2);
            Console.WriteLine("{0} radians = {1} grads.\n", xrads2, RadiansToGrads(xrads2));

            double xdegs2 = 90.0;
            double xgrads2 = DegreesToGrads(xdegs2);
            Console.WriteLine("{0}° = {1} grads.", xdegs2, xgrads2);
            Console.WriteLine("{0} grads = {1}°.\n", xgrads2, GradsToDegrees(xgrads2));

            Console.WriteLine("-------------------------------------\n");
        }

        public static double Sec(double x)
        {
            return (1.0 / Math.Cos(x));
        }

        public static double Csc(double x)
        {
            return (1.0 / Math.Sin(x));
        }

        public static double Cot(double x)
        {
            return (Math.Cos(x) / Math.Sin(x));
        }

        public static void TrigFunctions_Example()
        {
            Console.WriteLine("\n--- RUNNING TRIG FUNCTIONS EXAMPLE ---\n");
            for (double angleDEG = 0.0; angleDEG <= 360.0; angleDEG += 45.0)
            {
                double angleRAD = DegreesToRadians(angleDEG);
                Console.WriteLine("Angle = {0}\xb0", angleDEG);
                Console.WriteLine("cos({0}\xb0) = {1}", angleDEG, Math.Cos(angleRAD));
                Console.WriteLine("sin({0}\xb0) = {1}", angleDEG, Math.Sin(angleRAD));
                Console.WriteLine("tan({0}\xb0) = {1}", angleDEG, Math.Tan(angleRAD));
                Console.WriteLine("sec({0}°) = {1}", angleDEG, Sec(angleRAD));
                Console.WriteLine("csc({0}°) = {1}", angleDEG, Csc(angleRAD));
                Console.WriteLine("cot({0}°) = {1}", angleDEG, Cot(angleRAD));
                Console.WriteLine("--------------------------------------------\n");
            }
            Console.WriteLine("--------------------------------------------\n");
        }

        public static double Acos2(double x)
        {
            if (x == 1.0) return 0.0;
            else if (x == -1.0) return Math.PI;
            else
                return ((Math.PI / 2.0) - Math.Atan(x / Math.Sqrt(1.0 - x * x)));
        }

        public static double Asin2(double x)
        {
            if (x == 1.0) return (Math.PI / 2.0);
            else if (x == -1.0) return (-Math.PI / 2.0);
            else
                return (Math.Atan(x / Math.Sqrt(1.0 - x * x)));
        }

        //The following routine gives a "raw" version of the built-in
        //library code for Math.Atan2(double y, double x)
        public static double Atan2A(double y, double x)
        {
            if (x == 0.0)
            {
                if (y == 0.0)
                    return double.NaN;
                else
                    return (Math.Sign(y) * (Math.PI / 2.0));
            }
            else if (x < 0.0)
            {
                if (y < 0.0)
                {
                    return (-Math.PI + Math.Atan(y / x));
                }
                else
                {
                    return (Math.PI + Math.Atan(y / x));
                }
            }
            else
                return (Math.Atan(y / x));
        }

        public static double Asec(double x)
        {
            return Math.Acos(1.0 / x);
        }

        public static double Acsc(double x)
        {
            return Math.Asin(1.0 / x);
        }

        public static double Acot(double x)
        {
            return Math.Atan(1.0 / x);
        }

        public static void InverseTrigFunctions_Example()
        {
            Console.WriteLine("\n--- RUNNING INVERSE TRIG FUNCTIONS EXAMPLE ---\n");
            for (double angleDEG = 0.0; angleDEG <= 360.0; angleDEG += 45.0)
            {
                double angleRAD = DegreesToRadians(angleDEG);
                Console.WriteLine("Angle = {0}°", angleDEG);
                Console.WriteLine("ArcCos({0}) = {1}\xb0", Math.Cos(angleRAD), RadiansToDegrees(Math.Acos(Math.Cos(angleRAD))));
                Console.WriteLine("ArcCos2({0}) = {1}\xb0", Math.Cos(angleRAD), RadiansToDegrees(Acos2(Math.Cos(angleRAD))));

                Console.WriteLine("ArcSin({0}) = {1}\xb0", Math.Sin(angleRAD), RadiansToDegrees(Math.Asin(Math.Sin(angleRAD))));
                Console.WriteLine("ArcSin2({0}) = {1}\xb0", Math.Sin(angleRAD), RadiansToDegrees(Asin2(Math.Sin(angleRAD))));

                Console.WriteLine("ArcTan({0}) = {1}\xb0", Math.Tan(angleRAD), RadiansToDegrees(Math.Atan(Math.Tan(angleRAD))));
                Console.WriteLine("ArcTan2({0}) = {1}\xb0", Math.Tan(angleRAD), RadiansToDegrees(Math.Atan2(Math.Sin(angleRAD), Math.Cos(angleRAD))));
                Console.WriteLine("ArcTan2A({0}) = {1}\xb0", Math.Tan(angleRAD), RadiansToDegrees(Atan2A(Math.Sin(angleRAD), Math.Cos(angleRAD))));

                Console.WriteLine("ArcSec({0}) = {1}°", Sec(angleRAD), RadiansToDegrees(Asec(Sec(angleRAD))));
                Console.WriteLine("ArcCsc({0}) = {1}°", Csc(angleRAD), RadiansToDegrees(Acsc(Csc(angleRAD))));
                Console.WriteLine("ArcCot({0}) = {1}°", Cot(angleRAD), RadiansToDegrees(Acot(Cot(angleRAD))));
                Console.WriteLine("--------------------------------------------\n");
            }
            Console.WriteLine("--------------------------------------------\n");
        }

        public static double Sech(double x)
        {
            return (1.0 / Math.Cosh(x));
        }

        public static double Csch(double x)
        {
            return (1.0 / Math.Sinh(x));
        }

        public static double Coth(double x)
        {
            return (Math.Cosh(x) / Math.Sinh(x));
        }

        public static void HyperbolicFunctions_Example()
        {
            Console.WriteLine("\n--- RUNNING HYPERBOLIC FUNCTIONS EXAMPLE ---\n");
            for (double angleDEG = 0.0; angleDEG <= 360.0; angleDEG += 45.0)
            {
                double angleRAD = DegreesToRadians(angleDEG);
                Console.WriteLine("Angle = {0}\xb0", angleDEG);
                Console.WriteLine("Cosh({0}\xb0) = {1}", angleDEG, Math.Cosh(angleRAD));
                Console.WriteLine("Sinh({0}\xb0) = {1}", angleDEG, Math.Sinh(angleRAD));
                Console.WriteLine("Tanh({0}\xb0) = {1}", angleDEG, Math.Tanh(angleRAD));
                Console.WriteLine("Sech({0}°) = {1}", angleDEG, Sech(angleRAD));
                Console.WriteLine("Csch({0}°) = {1}", angleDEG, Csch(angleRAD));
                Console.WriteLine("Coth({0}°) = {1}", angleDEG, Coth(angleRAD));
                Console.WriteLine("--------------------------------------------\n");
            }
            Console.WriteLine("--------------------------------------------\n");
        }

        public static double ASinh(double x)
        {
            return (Math.Log(x + Math.Sqrt(x * x + 1.0)));
        }

        public static double ACosh(double x)
        {
            return (Math.Log(x + Math.Sqrt((x * x) - 1.0)));
        }

        public static double ATanh(double x)
        {
            return (Math.Log((1.0 + x) / (1.0 - x)) / 2.0);
        }

        public static double ACoth(double x)
        {
            //return (Math.Log((x + 1.0) / (x - 1.0)) / 2.0);
            return (ATanh(1.0 / x));
        }

        public static double ASech(double x)
        {
            return (ACosh(1.0 / x));
        }

        public static double ACsch(double x)
        {
            return (ASinh(1.0 / x));
        }

        public static void InverseHyperbolicFunctions_Example()
        {
            Console.WriteLine("\n--- RUNNING INVERSE HYPERBOLIC FUNCTIONS EXAMPLE ---\n");
            for (double angleDEG = 0.0; angleDEG <= 360.0; angleDEG += 45.0)
            {
                double angleRAD = DegreesToRadians(angleDEG);
                Console.WriteLine("Angle = {0}\xb0", angleDEG);
                Console.WriteLine("ACosh({0}) = {1}\xb0", Math.Cosh(angleRAD), RadiansToDegrees(ACosh(Math.Cosh(angleRAD))));
                Console.WriteLine("ASinh({0}) = {1}\xb0", Math.Sinh(angleRAD), RadiansToDegrees(ASinh(Math.Sinh(angleRAD))));
                Console.WriteLine("ATanh({0}) = {1}\xb0", Math.Tanh(angleRAD), RadiansToDegrees(ATanh(Math.Tanh(angleRAD))));
                Console.WriteLine("ASech({0}) = {1}°", Sech(angleRAD), RadiansToDegrees(ASech(Sech(angleRAD))));
                Console.WriteLine("ACsch({0}) = {1}°", Csch(angleRAD), RadiansToDegrees(ACsch(Csch(angleRAD))));
                Console.WriteLine("ACoth({0}) = {1}°", Coth(angleRAD), RadiansToDegrees(ACoth(Coth(angleRAD))));
                Console.WriteLine("--------------------------------------------\n");
            }
        }

        public static void CeilingFloorFunctions_Example()
        {
            double[] values = { 7.03, 7.64, 0.12, -0.12, -7.1, -7.6 };
            Console.WriteLine("Value   Ceiling   Floor   Truncate\n");
            foreach (double value in values)
                Console.WriteLine("{0,5} {1,5} {2,8} {3,8}", value,
                    Math.Ceiling(value), Math.Floor(value), Math.Truncate(value));
        }

        public static double RoundUp(double x)
        {
            return Math.Floor(x + 0.5);
        }

        public static double RoundDown(double x)
        {
            double floorValue = Math.Floor(x);
            if ((x - floorValue) > .5)
            {
                return (floorValue + 1);
            }
            else
            {
                return (floorValue);
            }
        }

        public static void RoundingFunctions_Example()
        {
            Console.WriteLine("Example of Using Method: Round(x)");
            Console.WriteLine("Round(4.4)={0}", Math.Round(4.4));
            Console.WriteLine("Round(4.5)={0}", Math.Round(4.5));
            Console.WriteLine("Round(4.6)={0}", Math.Round(4.6));
            Console.WriteLine("Round(-4.4)={0}", Math.Round(-4.4));
            Console.WriteLine("Round(-4.5)={0}", Math.Round(-4.5));
            Console.WriteLine("Round(-4.6)={0}\n", Math.Round(-4.6));

            Console.WriteLine("Example of Using Method: Round(x,n)");
            Console.WriteLine("Round(4.44,1)={0}", Math.Round(4.44, 1));
            Console.WriteLine("Round(4.45,1)={0}", Math.Round(4.45, 1));
            Console.WriteLine("Round(4.46,1)={0}", Math.Round(4.46, 1));
            Console.WriteLine("Round(-4.44,1)={0}", Math.Round(-4.44, 1));
            Console.WriteLine("Round(-4.45,1)={0}", Math.Round(-4.45, 1));
            Console.WriteLine("Round(-4.46,1)={0}\n", Math.Round(-4.46, 1));

            double result = 0.0;
            double posValue = 3.45;
            double negValue = -3.45;

            Console.WriteLine("Example of Using Method: Round(x,n)");
            // By default, round a positive and a negative value 
            // to the nearest even number. 
            // The precision of the result is 1 decimal place.

            result = Math.Round(posValue, 1);
            Console.WriteLine("Math.Round({1,5},1)={0,4}", result, posValue);
            result = Math.Round(negValue, 1);
            Console.WriteLine("Math.Round({1,5},1)={0,4}\n", result, negValue);

            Console.WriteLine("Example of Using Method: Round(x,n,mode)");
            // Round a positive value to the nearest even number, 
            // then to the nearest number away from zero. 
            // The precision of the result is 1 decimal place.

            result = Math.Round(posValue, 1, MidpointRounding.ToEven);
            Console.WriteLine("Math.Round({1,5},1,MidpointRounding.ToEven)={0,4}", result, posValue);
            result = Math.Round(posValue, 1, MidpointRounding.AwayFromZero);
            Console.WriteLine("Math.Round({1,5},1,MidpointRounding.AwayFromZero)={0,4}\n", result, posValue);

            // Round a negative value to the nearest even number, 
            // then to the nearest number away from zero. 
            // The precision of the result is 1 decimal place.

            result = Math.Round(negValue, 1, MidpointRounding.ToEven);
            Console.WriteLine("Math.Round({1,5},1,MidpointRounding.ToEven)={0,4}", result, negValue);
            result = Math.Round(negValue, 1, MidpointRounding.AwayFromZero);
            Console.WriteLine("Math.Round({1,5},1,MidpointRounding.AwayFromZero)={0,4}", result, negValue);

            Console.WriteLine("\n\nMath.Round(1.5,MidpointRounding.AwayFromZero)={0}", Math.Round(1.5, MidpointRounding.AwayFromZero));
            Console.WriteLine("Math.Round(2.5,MidpointRounding.AwayFromZero)={0}", Math.Round(2.5, MidpointRounding.AwayFromZero));

            Console.WriteLine("\n\nMath.Round(1.5,MidpointRounding.ToEven)={0}", Math.Round(1.5, MidpointRounding.ToEven));
            Console.WriteLine("Math.Round(2.5,MidpointRounding.ToEven)={0}", Math.Round(2.5, MidpointRounding.ToEven));

            //Testing the RoundUp Method
            Console.WriteLine("\n\nRoundUp(.4) = {0}", RoundUp(.4));
            Console.WriteLine("RoundUp(.5) = {0}", RoundUp(.5));
            Console.WriteLine("RoundUp(.6) = {0}", RoundUp(.6));
            Console.WriteLine("RoundUp(1.4) = {0}", RoundUp(1.4));
            Console.WriteLine("RoundUp(1.5) = {0}", RoundUp(1.5));
            Console.WriteLine("RoundUp(1.6) = {0}", RoundUp(1.6));
            Console.WriteLine("RoundUp(2.4) = {0}", RoundUp(2.4));
            Console.WriteLine("RoundUp(2.5) = {0}", RoundUp(2.5));
            Console.WriteLine("RoundUp(2.6) = {0}", RoundUp(2.6));

            //Testing the RoundDown Method
            Console.WriteLine("\nRoundDown(.4) = {0}", RoundDown(.4));
            Console.WriteLine("RoundDown(.5) = {0}", RoundDown(.5));
            Console.WriteLine("RoundDown(.6) = {0}", RoundDown(.6));
            Console.WriteLine("RoundDown(1.4) = {0}", RoundDown(1.4));
            Console.WriteLine("RoundDown(1.5) = {0}", RoundDown(1.5));
            Console.WriteLine("RoundDown(1.6) = {0}", RoundDown(1.6));
            Console.WriteLine("RoundDown(2.4) = {0}", RoundDown(2.4));
            Console.WriteLine("RoundDown(2.5) = {0}", RoundDown(2.5));
            Console.WriteLine("RoundDown(2.6) = {0}", RoundDown(2.6));
        }

        static void Main(string[] args)
        {
            PInE_Example();
            Console.ReadLine();

            MinMax_Example();
            Console.ReadLine();

            PowExpLogLog10_Example();
            Console.ReadLine();

            MultDivRem_Example();
            Console.ReadLine();

            Misc_Examples();
            Console.ReadLine();

            AngleMeas_Examples();
            Console.ReadLine();

            TrigFunctions_Example();
            Console.ReadLine();

            InverseTrigFunctions_Example();
            Console.ReadLine();

            HyperbolicFunctions_Example();
            Console.ReadLine();

            InverseHyperbolicFunctions_Example();
            Console.ReadLine();

            CeilingFloorFunctions_Example();
            Console.ReadLine();

            RoundingFunctions_Example();
            Console.ReadLine();
        }
    }
}
