﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Text.RegularExpressions;
using System.Collections;
using System.IO;

namespace Chapter06_CodeExamples
{
    class Program
    {

        #region Bitwise Methods

        //Converts a number stored as string from one base to another
        public static string ConvertToBase(string number, int start_base, int end_base)
        {
            long base10 = ConvertToBase10(number, start_base);
            string result = ConvertFromBase10(base10, end_base);
            return result;
        }

        //Converts a number stored as a string into a long data type value expressed in base 10
        public static long ConvertToBase10(string number, int start_base)
        {
            if (start_base == 10) return Convert.ToInt64(number);
            if (start_base < 2 || start_base > 36) return 0;

            char[] CharDataArray = number.ToCharArray();
            int i = CharDataArray.Length - 1;
            int j = start_base;
            int k = 0;
            long result = 0;

            foreach (char c in CharDataArray)
            {
                if (char.IsNumber(c))
                    k = int.Parse(c.ToString());
                else
                    k = Convert.ToInt32(c) - 55;
                result += k * (Convert.ToInt64(Math.Pow(j, i)));
                i--;
            }
            return result;
        }

        //Converts a long data type value into a number expressed as a string in some specified base
        public static string ConvertFromBase10(long number, int end_base)
        {
            if (end_base == 10) return number.ToString();
            if (end_base < 2 || end_base > 36) return "";

            long quotient = number;
            long remainder;
            string result = "";

            while (quotient >= end_base)
            {
                remainder = quotient % end_base;
                quotient = quotient / end_base;

                if (remainder < 10)
                    result = remainder.ToString() + result;
                else
                    result = Convert.ToChar(remainder + 55).ToString() + result;
            }

            if (quotient < 10)
                result = quotient.ToString() + result;
            else
                result = Convert.ToChar(quotient + 55).ToString() + result;

            return result;
        }

        //Utility to print out test results
        public static void TestBaseConversion(string numinp, int base1, int base2)
        {
            string outstr = "{0} in base {1} = {2} in base {3}\n";
            string numout = ConvertToBase(numinp, base1, base2);
            Console.WriteLine(outstr, numinp, base1, numout, base2);
            return;
        }

        //Converts an int32 number value into binary, octal or a hexadecimal string
        //using only bitwise manipulations of input data
        public static string DecToBinOctOrHex(int number, int baseval)
        {
            int n;
            int bitcounter;
            int bitstoshift;
            int startbit;
            string digit = "0123456789ABCDEF";

            switch (baseval)
            {
                case 2:
                    startbit = 31;
                    bitstoshift = 1;
                    break;
                case 8:
                    startbit = 30;
                    bitstoshift = 3;
                    break;
                case 16:
                    startbit = 28;
                    bitstoshift = 4;
                    break;
                default:
                    startbit = 0;
                    bitstoshift = 0;
                    break;
            }

            StringBuilder output = new StringBuilder();

            for (bitcounter = startbit, n = number; bitcounter >= 0; bitcounter -= bitstoshift)
            {
                output.Append(digit[(n >> bitcounter) & (baseval - 1)]);
            }
            return output.ToString();
        }

        //Converts an double number value, INCLUDING FRACTIONAL PARTS of the number, into binary,
        //octal or a hexadecimal string using only bitwise manipulations of input data.
        public static string DecToBinOctOrHex(double number, int baseval)
        {
            int n;
            int bitcounter;
            int bitstoshift;
            int startbit;
            string digit = "0123456789ABCDEF";

            switch (baseval)
            {
                case 2:
                    startbit = 31;
                    bitstoshift = 1;
                    break;
                case 8:
                    startbit = 30;
                    bitstoshift = 3;
                    break;
                case 16:
                    startbit = 28;
                    bitstoshift = 4;
                    break;
                default:
                    startbit = 0;
                    bitstoshift = 0;
                    break;
            }

            StringBuilder output = new StringBuilder();

            for (bitcounter = startbit, n = (int)Math.Truncate(number); bitcounter >= 0; bitcounter -= bitstoshift)
            {
                output.Append(digit[(n >> bitcounter) & (baseval - 1)]);
            }
            output.Append(".");
            for (int i = 0; i < 12; i++)
            {
                number = (number - Math.Floor(number)) * baseval;
                output.Append(digit[(int)Math.Truncate(number)]);
            }
            return output.ToString();
        }

        //Routine to display int32 input data type values into equivalent bit values for a specified number of bits
        public static string DisplayBits(int number, int nBitsToDisplay)
        {
            int mask = 1 << nBitsToDisplay - 1;
            StringBuilder output = new StringBuilder();

            // Get each bit, add space every 8 bits 
            // for display formatting
            for (int bitcounter = 1; bitcounter <= nBitsToDisplay; bitcounter++)
            {
                // Append 0 or 1 depending on result of masking
                output.Append(
                   (number & mask) == 0 ? "0" : "1");
                // Shift left so that mask will find bit of 
                // next digit during next iteration of loop
                number <<= 1;

                if (bitcounter % 8 == 0)
                    output.Append(" ");
            }
            return output.ToString();
        }

        public static long BitCount1(BitArray ba)
        {
            //Count number of bits directly from BitArray
            int bitCount = 0;
            foreach (bool b in ba)
            {
                if (b) bitCount += 1;
            }
            return bitCount;
        }

        public static long BitCount2(int value)
        {
            //Count number of bits directly from input value
            int bitCount = 0;
            while (value != 0)
            {
                bitCount++;
                value &= (value - 1);
            }
            return bitCount;
        }

        public static long ConvertEndianess(long i)
        {
            long temp = 0;
            temp |= (i & 0xFF000000) >> 24;
            temp |= (i & 0x00FF0000) >> 8;
            temp |= (i & 0x0000FF00) << 8;
            temp |= (i & 0x000000FF) << 24;
            return temp;
        }

        // Create a byte array from a decimal numeric value.
        public static byte[] DecimalToByteArray(decimal src)
        {
            // Create a MemoryStream as a buffer to hold the binary data.
            using (MemoryStream stream = new MemoryStream())
            {
                // Create a BinaryWriter to write binary data to the stream.
                using (BinaryWriter writer = new BinaryWriter(stream))
                {
                    // Write the decimal to the BinaryWriter/MemoryStream.
                    writer.Write(src);

                    // Return the byte representation of the decimal.
                    return stream.ToArray();
                }
            }
        }

        // Create a decimal numeric value from a byte array.
        public static decimal ByteArrayToDecimal(byte[] src)
        {
            // Create a MemoryStream containing the byte array.
            using (MemoryStream stream = new MemoryStream(src))
            {
                // Create a BinaryReader to read the decimal from the stream.
                using (BinaryReader reader = new BinaryReader(stream))
                {
                    // Read and return the decimal from the 
                    // BinaryReader/MemoryStream.
                    return reader.ReadDecimal();
                }
            }
        }

        enum RotateDirection { LEFT, RIGHT };
        //Returns the mask for masking out the bits which will be displaced in RotateBits
        private static int CalcRotationMask(int BitsToRotate, int Direction)
        {
            int i;
            int mask = 0;

            if (BitsToRotate == 0)
                return 0;

            i = Int32.MinValue;
            //i = 0x80000000 in hex or
            //i = 10000000000000000000000000000000 in binary
            mask = (i >> BitsToRotate);
            if (Direction == (int)RotateDirection.RIGHT)
            {
                mask = (i >> (32 - BitsToRotate));
                mask = ~mask;
            }
            else
                mask = (i >> BitsToRotate);
            return mask;
        }

        //value = value to be rotated 
        //BitsToRotate = number of bits to rotate
        //BitsizeOfResult = Bitsize of the resultant value = (8, 16, or 32)
        //Returns the value rotated left/right by the number
        //of BitsToRotate bits

        private static int RotateBits(int value, int BitsToRotate, int BitsToDisplay, int Direction)
        {
            int temp = 0;
            int mask = 0; ;
            int target = 0;

            BitsToRotate %= BitsToDisplay;

            if (Direction == (int)RotateDirection.RIGHT)
            {
                target = value;
                // Determine which bits will be affected by the rotation
                mask = CalcRotationMask(BitsToRotate, Direction);
                // Save those affected bits
                temp = value & mask;
                // Do the rotatation 
                target = (value >> BitsToRotate);
                // Shift the saved bits back to their proper place
                temp <<= (BitsToDisplay - BitsToRotate);
            }
            else if (Direction == (int)RotateDirection.LEFT)
            {
                // Determine which bits will be affected by the rotation
                mask = CalcRotationMask(BitsToRotate, Direction);
                // Shift mask into the proper place
                mask >>= (32 - BitsToDisplay);
                // Save those affected bits
                temp = value & mask;
                // Do the rotatation 
                target = (value << BitsToRotate);
                // Shift the saved bits back to their proper place
                temp >>= (BitsToDisplay - BitsToRotate);
            }
            // Add the rotated bits back into their proper location
            target |= temp;
            // and return the result
            return target;
        }

        //Extract the High Word value from an int
        public static int GetHighWord(int value)
        {
            return (value & (0xFFFF << 16));
        }

        //Extract the Low Word value from an int
        public static int GetLowWord(int value)
        {
            return (value & 0x0000FFFF);
        }

        //Extract the High Byte value from a short
        public static short GetHighByte(short shortValue)
        {
            return (short)(shortValue & (0xFF << 8));
        }

        //Extract the Low Byte value from a short
        public static short GetLowByte(short shortValue)
        {
            return (short)(shortValue & (short)0xFF);
        }

        //Returns value with the mask bits set. 
        public static int setBits(int value, int mask)
        {
            return value | mask;
        }

        //Returns value with the all the bits set except the mask bits. 
        public static int setBitsExcept(int value, int mask)
        {
            return value | ~mask;
        }

        //Returns true if any of the bits in mask is set in value. 
        public static bool isAnyBitSet(int value, int mask)
        {
            return (value & mask) != 0;
        }

        //Returns true if all the bits in mask are set in value. 
        public static bool areAllBitsSet(int value, int mask)
        {
            return (value & mask) == mask;
        }

        //Returns value with the nth bit set. 
        public static int setBitByPos(int value, int bitNumber)
        {
            return value | (1 << bitNumber);
        }

        //Returns true if value has the the nth bit set. 
        public static bool isBitSetByPos(int value, int bitNumber)
        {
            return (value & (1 << bitNumber)) != 0;
        }

        //Returns value with the mask bits set or cleared depending on the value of set.
        public static int setBits(int value, int mask, int bSet)
        {
            return bSet != 0 ? setBits(value, mask) : clearBits(value, mask);
        }

        //Returns value with the mask bits cleared. 
        public static int clearBits(int value, int mask)
        {
            return value & ~mask;
        }

        //Returns true if all the bits in mask are cleared in value. 
        public static bool areAllBitsClear(int value, int mask)
        {
            return (value & mask) == 0;
        }

        //Flips bit at position n 
        public static int BitFlip(int value, int n)
        {
            return ((value) ^ (1 << (n)));
        }

        //Toggles bit according to mask
        public static void BitToggle(int value, int mask)
        {
            mask ^= value;
        }

        //Returns value with the all the bits cleared except the mask bits. 
        public static int clearBitsExcept(int value, int mask)
        {
            return value & mask;
        }

        //Returns value with the nth bit cleared. 
        public static int clearBitByPos(int value, int bitNumber)
        {
            return value & ~(1 << bitNumber);
        }

        //Returns true if value has the the nth bit cleared. 
        public static bool isBitClearByPos(int value, int bitNumber)
        {
            return isBitSetByPos(value, bitNumber) == false;
        }

        //Returns value with the add bits set and the remove bits cleared. 
        public static int setClearBits(int value, int add, int remove)
        {
            return (value | add) & ~remove;
        }

        //Returns the one's complement of an input int value
        public static int OnesComplement(int value)
        {
            return (~value);
        }

        //Returns the two's complement of an input int value
        public static int TwosComplement(int value)
        {
            return (~value + 1);
        }

        //Returns x with the n bits that begin at position  
        //p inverted leaving the others bits unchanged. 
        private static uint Invert(uint x, int p, int n)
        {
            return x ^ (~(~0U << n) << p);
        }

        // Reverse the bits in a single byte 
        public static byte ReverseBitsInAByte(byte inByte)
        {
            byte result = 0x00;
            byte mask = 0x00;

            for (mask = 0x80; Convert.ToInt32(mask) > 0; mask >>= 1)
            {
                result >>= 1;
                byte tempbyte = (byte)(inByte & mask);
                if (tempbyte != 0x00)
                    result |= 0x80;
            }
            return (result);
        }

        //Reverses the byte order in an int value (version 1)
        public static int reverseBytes1(int value)
        {
            int b1 = (value >> 0) & 0xff;
            int b2 = (value >> 8) & 0xff;
            int b3 = (value >> 16) & 0xff;
            int b4 = (value >> 24) & 0xff;

            return b1 << 24 | b2 << 16 | b3 << 8 | b4 << 0;
        }

        //Reverses the byte order in an int value (version 2)
        public static int reverseBytes2(int value)
        {
            Byte[] bytes = BitConverter.GetBytes(value);
            Array.Reverse(bytes);
            return BitConverter.ToInt32(bytes, 0);
        }

        //Reverses the bit order in an int value
        public static int reverseBits(int value)
        {
            int n = 0;
            int i;
            // loop through all the bits
            for (i = 0; i < 32; i++)
            {
                // add bit from value to 1 bit left shifted variable
                n = (n << 1) + (value & 1);
                // right shift bits by 1
                value >>= 1;
            }
            return n;
        }

        public static string ExtractHexDigits(string input)
        {
            // Removes any characters that are not digits (like @)
            Regex isHexDigit = new Regex("[abcdefABCDEF\\d]+", RegexOptions.Compiled);
            string newnum = "";
            foreach (char c in input)
            {
                if (isHexDigit.IsMatch(c.ToString()))
                    newnum += c.ToString();
            }
            return newnum;
        }

        //Puts a string into a byte array
        public static byte[] StringToByteArr(string str)
        {
            System.Text.ASCIIEncoding encoding = new System.Text.ASCIIEncoding();
            return encoding.GetBytes(str);
        }

        public static int FindPatternInsideBitArray(byte[] ba, string pattern)
        {
            //Finds bit patterns inside a bit array
            int temp = 0;
            StringBuilder sb = new StringBuilder(8 * ba.Length);
            foreach (byte b in ba)
            {
                temp = int.Parse(Convert.ToString(b, 2));
                sb.Append(temp.ToString("00000000"));
            }
            string bitArray = sb.ToString();
            //Console.Clear();
            Console.WriteLine("bitArray is {0}", bitArray);
            Console.WriteLine("pattern  is {0}\n", pattern);
            int index = bitArray.IndexOf(pattern);
            return index;
        }

        public static string ConvertHexValuesToASCII(string HexValue)
        {
            //Converts input data in hex values to actual readable characters
            string StrValue = "";
            while (HexValue.Length > 0)
            {
                // Use ToChar() to convert each ASCII value (two hex digits) to the actual character
                StrValue += System.Convert.ToChar(System.Convert.ToUInt32(HexValue.Substring(0, 2), 16)).ToString();
                // Remove from the hex object the converted value
                HexValue = HexValue.Substring(2, HexValue.Length - 2);
            }
            return StrValue;
        }

        public static String ConvertASCIIValuesToHEX(string ASCIIValue)
        {
            StringBuilder sBuffer = new StringBuilder();
            for (int i = 0; i < ASCIIValue.Length; i++)
            {
                sBuffer.Append(Convert.ToInt32(ASCIIValue[i]).ToString("x"));
            }
            return sBuffer.ToString().ToUpper();
        }

        //Returns the radix of the given number. For example:
        //1st radix of 79981 is 1, 2nd radix of 79981 is 8
        public static int getRadix(int number, int radix)
        {
            return (int)(number / Math.Pow(10, radix - 1)) % 10;
        }
        #endregion
        
        static void Main(string[] args)
        {
            Console.Clear();
            Console.WriteLine("\nTesting Bit and Byte Routines for Chapter 06\n\n");

            //Test 1 - Number conversion between different bases
            Console.WriteLine("Testing number conversion between different bases\n");

            //Convert 25 in base 10 into binary and vice versa
            TestBaseConversion("25", 10, 2);
            TestBaseConversion("11001", 2, 10);

            //Convert 25 in base 10 into octal representation and vice versa
            TestBaseConversion("25", 10, 8);
            TestBaseConversion("31", 8, 10);

            //Convert 28 in base 10 into a hexadecimal and vice versa
            TestBaseConversion("28", 10, 16);
            TestBaseConversion("1C", 16, 10);

            //Convert a binary number to hexadecimal and vice versa
            TestBaseConversion("11001", 2, 16);
            TestBaseConversion("19", 16, 2);

            //Convert integer to binary
            int intdata = 28;
            string msg = "Integer  {0} in decimal = {1} in binary.\n";
            Console.WriteLine(msg, intdata, DecToBinOctOrHex(intdata, 2));
            intdata = -28;
            Console.WriteLine(msg, intdata, DecToBinOctOrHex(intdata, 2));

            //Convert double to binary
            double dbldata = 3.141592653589793;
            msg = "Double  {0} in decimal = \n{1} in binary.\n";
            Console.WriteLine(msg, dbldata, DecToBinOctOrHex(dbldata, 2));
            dbldata = -3.141592653589793;
            Console.WriteLine(msg, dbldata, DecToBinOctOrHex(dbldata, 2));

            //Convert integer to octal
            intdata = 28;
            msg = "Integer  {0} in decimal = {1} in octal.\n";
            Console.WriteLine(msg, intdata, DecToBinOctOrHex(intdata, 8));
            intdata = -28;
            Console.WriteLine(msg, intdata, DecToBinOctOrHex(intdata, 8));

            //Convert double to octal
            dbldata = 3.141592653589793;
            msg = "Double  {0} in decimal = {1} in octal.\n";
            Console.WriteLine(msg, dbldata, DecToBinOctOrHex(dbldata, 8));
            dbldata = -3.141592653589793;
            Console.WriteLine(msg, dbldata, DecToBinOctOrHex(dbldata, 8));

            //Convert integer to hexadecimal
            intdata = 28;
            msg = "Integer  {0} in decimal = {1} in hexadecmal.\n";
            Console.WriteLine(msg, intdata, DecToBinOctOrHex(intdata, 16));
            intdata = -28;
            Console.WriteLine(msg, intdata, DecToBinOctOrHex(intdata, 16));

            //Convert double to hexadecimal
            dbldata = 3.141592653589793;
            msg = "Double  {0} in decimal = {1} in hexadecmal.\n";
            Console.WriteLine(msg, dbldata, DecToBinOctOrHex(dbldata, 16));
            dbldata = -3.141592653589793;
            Console.WriteLine(msg, dbldata, DecToBinOctOrHex(dbldata, 16));

            Console.Write("Press enter to continue...\n");
            Console.ReadLine();
            Console.Clear();
            //Test 2 - Basic bitwise operations

            int a = 0x005A; // in binary = 0000 0000 0101 1010
            int b = 0x3C5A; // in binary = 0011 1100 0101 1010

            Console.WriteLine("Testing basic bitwise operations\n");

            Console.WriteLine("a = {0} = \t{1}", a, DisplayBits(a, 32));
            Console.WriteLine("b = {0} = \t{1}\n", b, DisplayBits(b, 32));

            Console.WriteLine("a & b = {0:x} = \t{1}", a & b, DisplayBits(a & b, 32));
            Console.WriteLine("a | b = {0:x} = \t{1}", a | b, DisplayBits(a | b, 32));
            Console.WriteLine("a ^ b = {0:x} = \t{1}", a ^ b, DisplayBits(a ^ b, 32));
            Console.WriteLine("~a = {0:x} = {1}\n", ~a, DisplayBits(~a, 32));

            Console.WriteLine("a>>1 = {0} = \t{1}", a >> 1, DisplayBits(a >> 1, 32));
            Console.WriteLine("a>>5 = {0} = \t{1}\n", a >> 5, DisplayBits(a >> 5, 32));

            Console.WriteLine("a<<1 = {0} = \t{1}", a << 1, DisplayBits(a << 1, 32));
            Console.WriteLine("a<<5 = {0} = \t{1}\n\n", a << 5, DisplayBits(a << 5, 32));

            Console.WriteLine("Ones complement of {0}\n = {1}\n", DisplayBits(a, 32), DisplayBits(OnesComplement(a), 32));
            Console.WriteLine("Twos complement of {0}\n = {1}\n", DisplayBits(a, 32), DisplayBits(TwosComplement(a), 32));
            Console.WriteLine("Twos Complement of {0} in decimal format = {1}\n", a, TwosComplement(a));

            Console.Write("Press enter to continue...\n");
            Console.ReadLine();
            Console.Clear();
            //Example 3 - Testing BitCount methods

            int intVal = 25;
            byte[] byteArrIntVal = BitConverter.GetBytes(intVal);
            BitArray ba = new BitArray(byteArrIntVal);

            Console.WriteLine("Bit counter for {0} = {1}", intVal, DisplayBits(intVal, 32));
            Console.WriteLine("Bit count (method 1) = {0}", BitCount1(ba));
            Console.WriteLine("Bit count (method 2) = {0}", BitCount2(intVal));

            Console.Write("\nPress enter to continue...\n");
            Console.ReadLine();
            Console.Clear();

            //Example 4 - Testing Little/Big Endians

            long d = 0x1234;
            Console.WriteLine("Testing Little/Big Endians memory scheme\n");

            if (BitConverter.IsLittleEndian)
            {
                Console.WriteLine("Original data (Little Endian) = {0}", DisplayBits((int)d, 32));
                Console.WriteLine("Reversed data (Big    Endian) = {0}", DisplayBits((int)ConvertEndianess(d), 32));
                Console.WriteLine("\n");
                byte[] dArray = BitConverter.GetBytes(d);
                Console.WriteLine("Original data (Little Endian) = {0}", BitConverter.ToString(dArray));
                dArray = BitConverter.GetBytes(ConvertEndianess(d));
                Console.WriteLine("Reversed data (Big    Endian) = {0}", BitConverter.ToString(dArray));
            }
            else
            {
                Console.WriteLine("Reversed data (Little Endian) = {0}", DisplayBits((int)ConvertEndianess(d), 32));
                Console.WriteLine("Original data (Big    Endian) = {0}", DisplayBits((int)d, 32));
            }

            Console.Write("\nPress enter to continue...\n");
            Console.ReadLine();
            Console.Clear();

            //Test 5 - Examples using the BitConverter class

            string formatter = "{0,10}{1,10}{2,25}";

            double aDoubl = 3.1415;
            float aSingl = 137.5F;
            long aLong = 123456789;
            int anInt = 9732;
            short aShort = 25;
            char aChar = 'M';
            bool aBool = true;
            decimal aDec = 842.696m;

            //Convert given data values into corresponding byte arrays
            byte[] byteArrDouble = BitConverter.GetBytes(aDoubl);
            byte[] byteArrSingl = BitConverter.GetBytes(aSingl);
            byte[] byteArrLong = BitConverter.GetBytes(aLong);
            byte[] byteArrInt = BitConverter.GetBytes(anInt);
            byte[] byteArrShort = BitConverter.GetBytes(aShort);
            byte[] byteArrChar = BitConverter.GetBytes(aChar);
            byte[] byteArrBool = BitConverter.GetBytes(aBool);
            byte[] byteArrDecimal = DecimalToByteArray(aDec);

            //Convert byte arrays values back into corresponding data values
            double valueDbl = BitConverter.ToDouble(byteArrDouble, 0);
            Single valueSingl = BitConverter.ToSingle(byteArrSingl, 0);
            long valueLong = BitConverter.ToInt64(byteArrLong, 0);
            int valueInt = BitConverter.ToInt32(byteArrInt, 0);
            short valueShort = BitConverter.ToInt16(byteArrShort, 0);
            char valueChar = BitConverter.ToChar(byteArrChar, 0);
            bool valueBool = BitConverter.ToBoolean(byteArrBool, 0);
            decimal valueDecimal = ByteArrayToDecimal(byteArrDecimal);

            Console.WriteLine(
            "Example of using methods in the BitConverter class\n" +
            "to convert values of specified data types into\n" +
            "corresponding byte arrays\n");
            Console.WriteLine(formatter, "Data Type", "Value", "Byte Array");
            Console.WriteLine(formatter, "--------", "----------", "----------");
            Console.WriteLine(formatter, "Double", aDoubl,
                BitConverter.ToString(byteArrDouble));
            Console.WriteLine(formatter, "Single", aSingl,
                BitConverter.ToString(byteArrSingl));
            Console.WriteLine(formatter, "Long", aLong,
                BitConverter.ToString(byteArrLong));
            Console.WriteLine(formatter, "Int", anInt,
                BitConverter.ToString(byteArrInt));
            Console.WriteLine(formatter, "Short", aShort,
                BitConverter.ToString(byteArrShort));
            Console.WriteLine(formatter, "Char", aChar,
                BitConverter.ToString(byteArrChar));
            Console.WriteLine(formatter, "bool", aBool,
                BitConverter.ToString(byteArrBool));
            Console.WriteLine(formatter, "Decimal", aDec,
                BitConverter.ToString(byteArrDecimal));

            Console.WriteLine(
            "\n\nExample of using methods in the BitConverter class\n" +
            "go convert byte arrays back to values of a\n" +
            "corresponding specified data type\n");
            formatter = "{0,25}{1,15}{2,15}";
            Console.WriteLine(formatter, "Byte Array", "Value", "Data Type");
            Console.WriteLine(formatter, "--------", "----------", "----------");
            Console.WriteLine(formatter,
            BitConverter.ToString(byteArrDouble), valueDbl, "Double");
            Console.WriteLine(formatter,
            BitConverter.ToString(byteArrSingl), valueSingl, "Single");
            Console.WriteLine(formatter,
            BitConverter.ToString(byteArrLong), valueLong, "Long");
            Console.WriteLine(formatter,
            BitConverter.ToString(byteArrInt), valueInt, "Int");
            Console.WriteLine(formatter,
            BitConverter.ToString(byteArrShort), valueShort, "Short");
            Console.WriteLine(formatter,
            BitConverter.ToString(byteArrChar), valueChar, "Char");
            Console.WriteLine(formatter,
            BitConverter.ToString(byteArrBool), valueBool, "Boolean");
            Console.WriteLine(formatter,
            BitConverter.ToString(byteArrDecimal), valueDecimal, "Decimal");

            Console.Write("\nPress enter to continue...\n");
            Console.ReadLine();
            Console.Clear();

            //Test 6 - Examples using the BitArray class

            Console.WriteLine("Examples of using the BitArray class\n");

            //Create a BitArray of bits
            BitArray bits = new BitArray(5);

            //And populate it with data
            for (int i = 0; i < bits.Length; i += 2)
            {
                bits[i] = true;
            }

            //Display original contents of bits BitArray
            foreach (bool bit in bits)
            {
                Console.Write(bit + " ");
            }
            Console.WriteLine();

            //Convert original contents in bits BitArray from 
            //boolean to corresponding integer values and 
            //redisplay contents of bits BitArray as integers
            foreach (bool bit in bits)
            {
                Console.Write(Convert.ToInt32(bit) + " ");
            }
            Console.WriteLine();

            //Convert each bit in bits BitArray into a byte array and
            //redisplay contents of bits BitArray as an array of bytes
            byte[] byteArrInt32;
            foreach (bool bit in bits)
            {
                //Convert bit data into an array of bytes and display
                byteArrInt32 = BitConverter.GetBytes(bit);
                Console.Write(BitConverter.ToString(byteArrInt32) + " ");
            }
            Console.Write("\n\nPress enter to continue...\n");
            Console.ReadLine();
            Console.Clear();

            //Example 7 - Testing bit rotation both LEFT & RIGHT by specified number of bits

            int bRotData = 0x2345;
            int bToShift = 4;
            int bToDisplay = 32;
            Console.WriteLine("Original Number To Test: " + DisplayBits(bRotData, bToDisplay));
            Console.WriteLine("Rotate  LEFT by  {0} bits: {1}", bToShift, DisplayBits(RotateBits(bRotData, bToShift, bToDisplay, (int)RotateDirection.LEFT), bToDisplay));
            Console.WriteLine("Rotate RIGHT by  {0} bits: {1}", bToShift, DisplayBits(RotateBits(bRotData, bToShift, bToDisplay, (int)RotateDirection.RIGHT), bToDisplay));

            Console.Write("\n\nPress enter to continue...\n");
            Console.ReadLine();
            Console.Clear();

            //Example 8 - Obtaining the High Word or Low Word of a Number

            int number = 0x12332344;
            Console.WriteLine("Original Data {0} = {1}", number, DisplayBits(number, 32));
            Console.WriteLine("HighWord of   {0} = {1}", number, DisplayBits(GetHighWord(number), 32));
            Console.WriteLine("LowWord of    {0} = {1}\n", number, DisplayBits(GetLowWord(number), 32));

            short num = 25;
            Console.WriteLine("Original Data {0} = {1}", num, DisplayBits(num, 32));
            Console.WriteLine("HighByte of   {0} = {1}", num, DisplayBits(GetHighByte(num), 32));
            Console.WriteLine("LowByte of    {0} = {1}", num, DisplayBits(GetLowByte(num), 32));

            Console.Write("\nPress enter to continue...\n");
            Console.ReadLine();

            int bitposition = 4;
            bToDisplay = 32;
            int bitsAvailable = bToDisplay - bitposition;

            Console.WriteLine("Testing function to invert bits starting at some specified position\n");
            Console.WriteLine("Original data:                            {0}", DisplayBits(number, bToDisplay));
            Console.WriteLine("Inverted data starting at bit position {0}: {1}",
                bitposition, DisplayBits((int)Invert((uint)number, bitposition, bitsAvailable), bToDisplay));

            Console.Write("\nPress enter to continue...\n");
            Console.ReadLine();
            Console.Clear();

            Console.WriteLine("Testing function to reverse Bytes and Bits of an int value\n");
            Console.WriteLine("Original Data  {0}\n", DisplayBits(number, 32));
            Console.WriteLine("Reverse Bytes1 {0}\n", DisplayBits(reverseBytes1(number), 32));
            Console.WriteLine("Reverse Bytes2 {0}\n", DisplayBits(reverseBytes2(number), 32));
            Console.WriteLine("Reverse Bits   {0}\n", DisplayBits(reverseBits(number), 32));

            Console.Write("\nPress enter to continue...\n");
            Console.ReadLine();
            Console.Clear();

            Console.WriteLine("Original Data (1 byte max) {0}\n", DisplayBits(54, 8));
            Console.WriteLine("Reverse Bits In A Byte     {0}\n", DisplayBits((int)ReverseBitsInAByte(54), 8));

            Console.Write("\nPress enter to continue...\n");
            Console.ReadLine();
            Console.Clear();

            string[] hex = new string[3];
            hex[0] = "$FFF@FF";
            hex[1] = "xVWAE";
            hex[2] = "'783EFKW#F'H";
            string str = "";

            foreach (string s in hex)
                str += s + " ----> " + ExtractHexDigits(s) + "\n";

            Console.WriteLine("{0}", str);

            Console.Write("\nPress enter to continue...\n");
            Console.ReadLine();
            Console.Clear();

            Console.WriteLine("Testing single bit flip at specified position\n");
            Console.WriteLine("Original Data                    {0}\n", DisplayBits(25, 32));
            Console.WriteLine("Original Data with bit 2 flipped {0}\n", DisplayBits(BitFlip(25, 2), 32));
            Console.WriteLine("Original Data with bit 3 flipped {0}\n", DisplayBits(BitFlip(25, 3), 32));

            Console.Write("\nPress enter to continue...\n");
            Console.ReadLine();
            Console.Clear();

            //Demo to illustrate how to find bit pattern inside a bit array
            Console.WriteLine("Example illustrating how to find a bit pattern embbeded inside an array of bits\n");
            byte[] baData = new byte[5] { 1, 2, 39, 4, 5 };
            string pattern = "0100111"; // decimal 39
            int foundPatternIndex = FindPatternInsideBitArray(baData, pattern);
            if (foundPatternIndex == -1)
                Console.WriteLine("Pattern wasn't found");
            else
                Console.WriteLine("Pattern first occurred at bit {0}", foundPatternIndex);

            Console.Write("\nPress enter to continue...\n");
            Console.ReadLine();
            Console.Clear();

            Console.WriteLine("Example illustrating how to convert hex values to actual readable characters\n");
            string HexValue = "48454C4C4F205448455245";
            Console.WriteLine("Input HEX data = {0}", HexValue);
            string strValue = ConvertHexValuesToASCII(HexValue);
            Console.WriteLine("Output result  = {0}\n", strValue);
            Console.WriteLine("and input data back into HEX characters...");
            Console.WriteLine("Output result  = {0}\n", ConvertASCIIValuesToHEX(strValue));

            Console.Write("\nPress enter to continue...\n");

            Console.Clear();
            Console.Write("\nRunning Radix Test...\n\n");
            Console.WriteLine("Radix 1 of 79981 = {0}", getRadix(79981,1));
            Console.WriteLine("Radix 2 of 79981 = {0}", getRadix(79981,2));
            Console.WriteLine("Radix 3 of 79981 = {0}", getRadix(79981,3));
            Console.WriteLine("Radix 4 of 79981 = {0}", getRadix(79981,4));
            Console.WriteLine("Radix 5 of 79981 = {0}", getRadix(79981,5));

            Console.ReadLine();
        }
    }
}
