﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

using System.Threading;
using System.Collections;
using System.Security.Cryptography;

namespace Chapter10_CodeExamples
{
    #region MersenneTwister
    public class MersenneTwister
    {
        // Class MersenneTwister generates random numbers 
        // from a uniform distribution using the Mersenne
        // Twister algorithm.
        private const int N = 624;
        private const int M = 397;
        private const uint MATRIX_A = 0x9908b0dfU;
        private const uint UPPER_MASK = 0x80000000U;
        private const uint LOWER_MASK = 0x7fffffffU;
        private const int MAX_RAND_INT = 0x7fffffff;
        private uint[] mag01 = { 0x0U, MATRIX_A };
        private uint[] mt = new uint[N];
        private int mti = N + 1;

        public MersenneTwister()
        { init_genrand((uint)DateTime.Now.Millisecond); }

        public MersenneTwister(int seed)
        { init_genrand((uint)seed); }

        public MersenneTwister(int[] init)
        {
            uint[] initArray = new uint[init.Length];
            for (int i = 0; i < init.Length; ++i)
                initArray[i] = (uint)init[i];
            init_by_array(initArray, (uint)initArray.Length);
        }

        public static int MaxRandomInt
        { get { return 0x7fffffff; } }

        public int Next()
        { return genrand_int31(); }

        public int Next(int maxValue)
        { return Next(0, maxValue); }

        public int Next(int minValue, int maxValue)
        {
            if (minValue > maxValue)
            {
                int tmp = maxValue;
                maxValue = minValue;
                minValue = tmp;
            }
            return (int)(Math.Floor((maxValue - minValue + 1) * genrand_real1() + minValue));
        }

        public float NextFloat()
        { return (float)genrand_real2(); }

        public float NextFloat(bool includeOne)
        {
            if (includeOne)
            {
                return (float)genrand_real1();
            }
            return (float)genrand_real2();
        }

        public float NextFloatPositive()
        { return (float)genrand_real3(); }

        public double NextDouble()
        { return genrand_real2(); }

        public double NextDouble(bool includeOne)
        {
            if (includeOne)
            {
                return genrand_real1();
            }
            return genrand_real2();
        }

        public double NextDoublePositive()
        { return genrand_real3(); }

        public double Next53BitRes()
        { return genrand_res53(); }

        public void Initialize()
        { init_genrand((uint)DateTime.Now.Millisecond); }

        public void Initialize(int seed)
        { init_genrand((uint)seed); }

        public void Initialize(int[] init)
        {
            uint[] initArray = new uint[init.Length];
            for (int i = 0; i < init.Length; ++i)
                initArray[i] = (uint)init[i];
            init_by_array(initArray, (uint)initArray.Length);
        }

        private void init_genrand(uint s)
        {
            mt[0] = s & 0xffffffffU;
            for (mti = 1; mti < N; mti++)
            {
                mt[mti] =
                  (uint)(1812433253U * (mt[mti - 1] ^ (mt[mti - 1] >> 30)) + mti);
                mt[mti] &= 0xffffffffU;
            }
        }

        private void init_by_array(uint[] init_key, uint key_length)
        {
            int i, j, k;
            init_genrand(19650218U);
            i = 1; j = 0;
            k = (int)(N > key_length ? N : key_length);
            for (; k > 0; k--)
            {
                mt[i] = (uint)((uint)(mt[i] ^ ((mt[i - 1] ^ (mt[i - 1] >> 30)) * 1664525U)) + init_key[j] + j);
                mt[i] &= 0xffffffffU;
                i++; j++;
                if (i >= N) { mt[0] = mt[N - 1]; i = 1; }
                if (j >= key_length) j = 0;
            }
            for (k = N - 1; k > 0; k--)
            {
                mt[i] = (uint)((uint)(mt[i] ^ ((mt[i - 1] ^ (mt[i - 1] >> 30)) * 1566083941U)) - i);
                mt[i] &= 0xffffffffU;
                i++;
                if (i >= N) { mt[0] = mt[N - 1]; i = 1; }
            }
            mt[0] = 0x80000000U;
        }

        uint genrand_int32()
        {
            uint y;
            if (mti >= N)
            {
                int kk;
                if (mti == N + 1)
                    init_genrand(5489U);
                for (kk = 0; kk < N - M; kk++)
                {
                    y = (mt[kk] & UPPER_MASK) | (mt[kk + 1] & LOWER_MASK);
                    mt[kk] = mt[kk + M] ^ (y >> 1) ^ mag01[y & 0x1U];
                }
                for (; kk < N - 1; kk++)
                {
                    y = (mt[kk] & UPPER_MASK) | (mt[kk + 1] & LOWER_MASK);
                    mt[kk] = mt[kk + (M - N)] ^ (y >> 1) ^ mag01[y & 0x1U];
                }
                y = (mt[N - 1] & UPPER_MASK) | (mt[0] & LOWER_MASK);
                mt[N - 1] = mt[M - 1] ^ (y >> 1) ^ mag01[y & 0x1U];
                mti = 0;
            }
            y = mt[mti++];
            y ^= (y >> 11);
            y ^= (y << 7) & 0x9d2c5680U;
            y ^= (y << 15) & 0xefc60000U;
            y ^= (y >> 18);
            return y;
        }

        private int genrand_int31()
        { return (int)(genrand_int32() >> 1); }

        double genrand_real1()
        { return genrand_int32() * (1.0 / 4294967295.0); }

        double genrand_real2()
        { return genrand_int32() * (1.0 / 4294967296.0); }

        double genrand_real3()
        { return (((double)genrand_int32()) + 0.5) * (1.0 / 4294967296.0); }

        double genrand_res53()
        {
            uint a = genrand_int32() >> 5, b = genrand_int32() >> 6;
            return (a * 67108864.0 + b) * (1.0 / 9007199254740992.0);
        }
    }
    #endregion    

    class Program
    {
        #region Testing RNGs
        ///////////////////////////////////////////////// 
        static void TestRNG()
        {
            Console.WriteLine("Testing C# Internal Random Number Generators\n\n");

            Console.WriteLine("Testing 6 random int32 from 0 to int32.MAX");
            int seed = 234;
            Random randObj1 = new Random(seed);
            for (int j = 0; j < 6; j++)
                Console.Write("{0,11} ", randObj1.Next());
            Console.WriteLine("\n");
            Console.ReadLine();

            Console.WriteLine("Testing 6 random int32 from 0 to upper bound = 25");
            int upper1 = 25;
            Random randObj2 = new Random(seed);
            for (int j = 0; j < 6; j++)
                Console.Write("{0,11} ", randObj2.Next(upper1));
            Console.WriteLine("\n");
            Console.ReadLine();

            Console.WriteLine("Testing 6 random int32 from lower to upper bound: -55 to 25");
            int lower2 = -55;
            int upper2 = 25;
            Random randObj3 = new Random(seed);
            for (int j = 0; j < 6; j++)
                Console.Write("{0,11} ", randObj3.Next(lower2, upper2));
            Console.WriteLine("\n");
            Console.ReadLine();

            Console.WriteLine("Testing 10 random bytes: 0 to 255");
            Random rnd = new Random();
            Byte[] b = new Byte[10];
            rnd.NextBytes(b);
            Console.WriteLine("The Random bytes are: ");
            for (int i = 0; i < 10; i++)
            {
                Console.Write(i);
                Console.Write(":");
                Console.WriteLine(b[i]);
            }
            Console.WriteLine("\n");
            Console.ReadLine();

            Console.WriteLine("Testing 6 random doubles from 0 to 1");
            int seedd = 234;
            Random randObj4 = new Random(seedd);
            for (int j = 0; j < 6; j++)
                Console.WriteLine("{0,11} ", randObj4.NextDouble());
            Console.WriteLine("\n");
            Console.ReadLine();

            Console.WriteLine("Testing 6 random doubles from -55 to 25 bound");
            int min = -55;
            int max = 25;
            Random randObj6 = new Random(seedd);
            for (int j = 0; j < 6; j++)
                Console.WriteLine("{0,11} ", (min + (max - min) * randObj6.NextDouble()));
            Console.WriteLine();
            Console.ReadLine();
        }        
        
        static void TestMersenneTwister()
        {
            Console.WriteLine("Testing the Mersenne Twister random number generator\n");
            MersenneTwister randGen = new MersenneTwister();
            Console.WriteLine("\n100 uniform random integers in [0,{0}]:\n", MersenneTwister.MaxRandomInt);
            int i;
            for (i = 0; i < 100; ++i)
            {
                Console.Write("{0} ", randGen.Next());
                if (i % 5 == 4) Console.WriteLine("");
            }

            Console.WriteLine("\n100 uniform random doubles in [0,1):\n");
            for (i = 0; i < 100; ++i)
            {
                Console.Write("{0} ", randGen.NextDouble().ToString("F8"));
                if (i % 5 == 4) Console.WriteLine("");
            }
            Console.ReadLine();
        }
        ///////////////////////////////////////////////// 
        #endregion

        #region Random Passwords
        static string CreateRandomPassword(int passwordLength)
        {
            string allowedChars = "abcdefghijkmnopqrstuvwxyzABCDEFGHJKLMNOPQRSTUVWXYZ0123456789!@$?_-";
            char[] chars = new char[passwordLength];
            Random rd = new Random();
            for (int i = 0; i < passwordLength; i++)
            {
                chars[i] = allowedChars[rd.Next(0, allowedChars.Length)];
            }
            return new string(chars);
        }

        static void TestRandomPassword()
        {
            Console.WriteLine("Testing for generation of random passwords");
            for (int i = 0; i < 6; i++)
            {
                Console.WriteLine("Password {0} = {1}", i, CreateRandomPassword(10));
                Thread.Sleep(2000);
            }
            Console.ReadLine();
        }
        ///////////////////////////////////////////////// 

        public static void TestBetterRandomString()
        {
            // create a stronger hash code using RNGCryptoServiceProvider
            byte[] random = new byte[64];
            RNGCryptoServiceProvider rng = new RNGCryptoServiceProvider();
            // populate with random bytes
            rng.GetBytes(random);

            // convert random bytes to string
            string randomBase64 = Convert.ToBase64String(random);
            // display
            Console.WriteLine("Random 64-bit string: {0}\r\n ", randomBase64);
            Console.ReadLine();
        }
        ///////////////////////////////////////////////// 

        public static string CreateRandomEncryptedPassword(int PasswordLength)
        {
            String _allowedChars = "abcdefghijkmnopqrstuvwxyzABCDEFGHJKLMNOPQRSTUVWXYZ23456789";
            Byte[] randomBytes = new Byte[PasswordLength];
            RNGCryptoServiceProvider rng = new RNGCryptoServiceProvider();
            rng.GetBytes(randomBytes);
            char[] chars = new char[PasswordLength];
            int allowedCharCount = _allowedChars.Length;
            for (int i = 0; i < PasswordLength; i++)
            {
                chars[i] = _allowedChars[(int)randomBytes[i] % allowedCharCount];
            }
            return new string(chars);
        }

        static void TestRandomEncryptedPassword()
        {
            for (int i = 0; i < 6; i++)
            {
                Console.WriteLine("Encrypted Password {0} = {1}", i, CreateRandomEncryptedPassword(10));
                Thread.Sleep(2000);
            }
            //Console.ReadLine();
        }
        #endregion

        #region Generating Unique Random Numbers
        ///////////////////////////////////////////////// 
        static int[] UniqueRandom1(int max)
        {
          Random rand = new Random();  
          int[] array = new int[max];

          // Initialize the array to the ints 0-max
          for (int i = 0; i < array.Length; i++)
          {
              array[i] = i;
          }

          for (int i = array.Length - 1; i > 0; i--)
          {
              int randomPosition = rand.Next(i+1);
              int temp = array[i];
              array[i] = array[randomPosition];
              array[randomPosition] = temp;
          }
          return array;
        }

        ///////////////////////////////////////////////// 

        static ArrayList UniqueRandom2(int max)
        {
            // Create an ArrayList object that will hold the numbers
            ArrayList lstNumbers = new ArrayList();
            // The Random class will be used to generate numbers
            Random rndNumber = new Random();

            // Generate a random number between 1 and the Max
            int number = rndNumber.Next(1, max + 1);
            // Add this first random number to the list
            lstNumbers.Add(number);
            // Set a count of numbers to 0 to start
            int count = 0;

            do // Repeatedly...
            {
                // ... generate a random number between 1 and the Max
                number = rndNumber.Next(1, max + 1);

                // If the newly generated number in not yet in the list...
                if (!lstNumbers.Contains(number))
                {
                    // ... add it
                    lstNumbers.Add(number);
                }
                // Increase the count
                count++;
            } while (count <= 10 * max); // Do that again

            // Once the list is built, return it
            return lstNumbers;
        }

        public static List<int> GetUniqueRandomList(int size)
        {
            List<int> list = new List<int>();
            Random randomGenerator = new Random();

            List<int> range = new List<int>();
            for (int i = 0; i < size; i++)
                range.Add(i);

            while (range.Count > 0)
            {
                int item = range[randomGenerator.Next(range.Count)];
                list.Add(item);
                range.Remove(item);
            }
            return list;
        }

        static void TestUniqueRandom1()
        {
            int size = 15;
            int[] myData = UniqueRandom1(size);
            Console.WriteLine("\n\nTesting unique random numbers version 1\n");
            for (int i = 0; i < myData.Length; i++)
            {
                Console.WriteLine("myData[{0}] = {1}", i, myData[i]);
            }
            Console.ReadLine();
        }

        static void TestUniqueRandom2()
        {
            const int Total = 15;
            ArrayList lstNumbers = UniqueRandom2(Total);

            Console.WriteLine("\nTesting unique random numbers version 2\n");
            for (int i = 0; i < lstNumbers.Count; i++)
                Console.WriteLine("x[{0}] = {1}", i, lstNumbers[i]);
            Console.ReadLine();
        }

        static void TestUniqueRandom3()
        {
            List<int> iList = new List<int>();
            iList = GetUniqueRandomList(15);

            Console.WriteLine("\nTesting unique random numbers version 3\n");
            foreach (int i in iList)
            {
                Console.WriteLine("x[{0}] = {1}", i, iList[i]);
            }
            Console.ReadLine();
        }

        #endregion
        
        static void Main(string[] args)
        {
            TestRNG();
            TestRandomPassword();
            TestBetterRandomString();
            TestRandomEncryptedPassword();
            TestUniqueRandom1();
            TestUniqueRandom2();
            TestUniqueRandom3();
            TestMersenneTwister();
        }
    }
}
