﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

using System.IO;
using System.Net;

namespace NaturalRandomNumbers
{
    class Program
    {
        //Returns an array of random integers between two numbers, both inclusive        
        public static int[] GetRandomInts(int min, int max, int trials)
        {
            //Build the url string to www.random.org
            string url = "http://www.random.org/integers/?num=" + trials.ToString();

            url += "&min=" + min.ToString();
            url += "&max=" + max.ToString();
            url += "&col=1&base=10&format=html&rnd=new";

            string data = DownloadData(url);

            if (data != string.Empty)
            {
                //Parse the data
                string startMarker = "<pre class=" + '"' + "data" + '"' + ">"; //<pre class="data">
                int j = data.IndexOf(startMarker);
                if (j != -1)
                {
                    int k = data.IndexOf("</pre>", j);
                    if (k != -1)
                    {
                        string intString = data.Substring(j + startMarker.Length, k - j - startMarker.Length);
                        intString = intString.Trim();

                        //Read each line
                        List<int> integers = new List<int>();
                        StringReader readLines = new StringReader(intString);

                        while (readLines.Peek() != -1)
                        {
                            integers.Add(int.Parse(readLines.ReadLine()));
                        }

                        return integers.ToArray();
                    }
                }
            }
            return new int[] { -1 };
        }

        //Connects to URL to download data
        private static string DownloadData(string url)
        {
            try
            {
                //Get a data stream from the url
                WebRequest req = WebRequest.Create(url);
                WebResponse response = req.GetResponse();
                Stream stream = response.GetResponseStream();

                //Download in chuncks
                byte[] buffer = new byte[1024];

                //Get Total Size
                int dataLength = (int)response.ContentLength;

                //Download to memory
                //Note: adjust the streams here to download directly to the hard drive
                MemoryStream memStream = new MemoryStream();
                while (true)
                {
                    //Try to read the data
                    int bytesRead = stream.Read(buffer, 0, buffer.Length);

                    if (bytesRead == 0)
                        break;
                    else
                        memStream.Write(buffer, 0, bytesRead);
                }

                //Convert the downloaded stream to a byte array
                string downloadedData = System.Text.ASCIIEncoding.ASCII.GetString(memStream.ToArray());

                //Clean up
                stream.Close();
                memStream.Close();

                return downloadedData;
            }
            catch (Exception)
            {
                return string.Empty;
            }
        }

        static void Main(string[] args)
        {
            //Sample driver program for testing purposes only
            //Sets up amount of random numbers to download
            //Change this to the number of random 
            //numbers you desire to download
            int NToDownload = 10;

            //Sets up random number range. Change this to 
            //the range of values you want.
            int min = 10;
            int max = 100;

            string numberArray = "";

            Console.WriteLine("\nAttempting to download true random numbers from RANDOM.ORG server. Please wait..");

            //Download numbers into integer array
            int[] randomNumbers = GetRandomInts(min, max, NToDownload);

            foreach (int rnd in randomNumbers)
            {
                numberArray += rnd.ToString() + Environment.NewLine;
            }
            Console.WriteLine(numberArray);
            Console.WriteLine("\nDone! Press ENTER key to terminate program.");
            Console.ReadLine();
        }
    }
}
