﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Collections;

namespace RandomDistributionExamples
{
    class Program
    {
        #region Special Functions from Chapter 14

        public static double GammaLn(double x)
        {
            if (x <= 0) throw
            new Exception("Input value must be > 0");

            double[] coef = new double[14]
            {57.1562356658629235,
            -59.5979603554754912,
            14.1360979747417471,
            -0.491913816097620199,
             0.339946499848118887E-4,
             0.465236289270485756E-4,
            -0.983744753048795646E-4,
             0.158088703224912494E-3,
            -0.210264441724104883E-3,
             0.217439618115212643E-3,
            -0.164318106536763890E-3,
             0.844182239838527433E-4,
            -0.261908384015814087E-4,
             0.368991826595316234E-5};

            double denominator = x;
            double series = 0.999999999999997092;
            double temp = x + 5.24218750000000000;
            temp = (x + 0.5) * Math.Log(temp) - temp;
            for (int j = 0; j < 14; j++)
                series += coef[j] / ++denominator;
            return (temp + Math.Log(2.5066282746310005 * series / x));
        }

        public static double Gamma(double x)
        {
            if (x <= 0) throw
                new Exception("Input value must be > 0");
            return Math.Exp(GammaLn(x));
        }

        public static double Beta(double x, double y)
        {
            if ((x <= 0) || (y <= 0)) throw
            new Exception("Input values must both be > 0");

            return Math.Exp(GammaLn(x) + GammaLn(y) - GammaLn(x + y));
        }

        public static double Factorial(double number)
        {
            double Result;
            if (number < 0)
            {
                throw new Exception("Input value must be > 0");
            }
            if (number == 0.0) Result = 1; //0! = 1
            else
                //Recursively call the Factorial function
                Result = (number * Factorial(number - 1));
            return Result;
        }

        private static readonly double[] factorialLookup = 
        { 
            1.0,
            1.0,
            2.0,
            6.0,
            24.0,
            120.0,
            720.0,
            5040.0, 
            40320.0,
            362880.0,
            3628800.0,
            39916800.0,
            479001600.0,
            6227020800.0,
            87178291200.0,
            1307674368000.0,
            20922789888000.0,
            355687428096000.0,
            6402373705728000.0,
            121645100408832000.0,
            2432902008176640000.0,
            51090942171709440000.0,
            1124000727777607680000.0,
            25852016738884976640000.0
        };

        //Calculates n! using a lookup table up to n < 24
        //For n > 24, it calculates n! = exp(Ln(n!))
        public static double Factorial(int n)
        {
            if (n < 0)
            {
                throw new Exception("Input value must be a > 0");
            }
            else if (n < 24)
            {
                return factorialLookup[n];
            }
            else
            {
                return Math.Floor(System.Math.Exp(FactorialLn(n)) + 0.5);
            }
        }

        public static double FactorialLn(int n)
        {
            if (n < 0)
            {
                throw new Exception("Input value must be > 0");
            }
            else
            {
                return GammaLn(n + 1.0);
            }
        }

        #endregion        

        #region Instantiation of random number generator object

        //Use the Visual C\# built-in random number
        //generator or one of your own choosing here.
        private static Random randObj = new Random();
        
        #endregion

        #region Utilities

        public static double[] makeHistogram(double[] data, double min, double max, int nBins)
        {
            double[] Histogram = new double[nBins];
            double BinWidth = (max - min) / nBins;
            for (int i = 0; i < nBins; i++)
            {
                int nCounts = 0;
                for (int j = 0; j < data.Length; j++)
                {
                    if (data[j] >= min + (i) * BinWidth && data[j] < min + (i + 1) * BinWidth)
                    {
                        nCounts++;
                    }
                }
                Histogram[i] = nCounts;
            }
            return Histogram;
        }

        public static double[] makeHistogram(double[] data, int nBins)
        {
            double[] Histogram = new double[nBins];
            for (int i = 0; i < nBins; i++)
            {
                int nCounts = 0;
                for (int j = 0; j < data.Length; j++)
                {
                    if (data[j] == i)
                    {
                        nCounts++;
                    }
                }
                Histogram[i] = nCounts;
            }
            return Histogram;
        }

        public static double dataMax(double[] data)
        {
            double max = data[0];
            for (int i = 1; i < data.Length; i++)
            {
                max = Math.Max(max, data[i]);
            }
            return max;
        }

        public static double dataMin(double[] data)
        {
            double min = data[0];
            for (int i = 1; i < data.Length; i++)
            {
                min = Math.Min(min, data[i]);
            }
            return min;
        }

        public static void stats(double[] data, out double mean, out double variance, out double sigma)
        {
            double sum = 0.0;
            mean = 0.0;
            variance = 0.0;
            sigma = 0.0;

            for (int i = 0; i < data.Length; i++)
            {
                sum += data[i];
            }
            mean = sum / data.Length;
            for (int i = 0; i < data.Length; i++)
            {
                variance += Math.Pow(data[i] - mean, 2);
            }
            variance /= data.Length;
            sigma = Math.Sqrt(variance);
            return;
        }

        static void TestBasicStats()
        {
            int nPoints = 10000;
            double mu = 2.0;
            double sigma = 0.5;

            double[] randObj = NextNormal(mu, sigma, nPoints);

            double mean = 0.0, variance = 0.0, CalcSigma = 0.0;
            stats(randObj, out mean, out variance, out CalcSigma);

            Console.WriteLine("\nTesting Basic Stats Using a Normal Distribution");
            Console.WriteLine("\nInput parameters: number of points = 10,000, mu = 2.0, sigma = 0.5");

            Console.WriteLine("\nResults (calculated):\n");
            Console.WriteLine("Mean = {0}", Math.Round(mean, 4));
            Console.WriteLine("Variance = {0}", Math.Round(variance, 4));
            Console.WriteLine("Standard Deviation = {0}", Math.Round(CalcSigma, 4));

            Console.WriteLine("\nResults (expected):\n");
            Console.WriteLine("Mean = {0}", Math.Round(mu, 4));
            Console.WriteLine("Variance = {0}", Math.Round(sigma * sigma, 4));
            Console.WriteLine("Standard Deviation = {0}", Math.Round(sigma, 4));

            Console.WriteLine("\nPress ENTER key to continue...");
            Console.ReadLine();
            Console.Clear();
        }
        
        #endregion

        //Discrete Distributions

        #region Bernoulli Distribution

        public static double BernoulliPMF(int x, double p)
        {
            if ((x == 0) || (x == 1))
                return (Math.Pow(p, x) * Math.Pow(1 - p, 1 - x));
            else
                return 0.0;
        }

        public static double[] BernoulliPMF(int[] x, double p)
        {
            double[] tempArr = new double[x.Length];
            for (int i = 0; i < x.Length; i++)
            {
                tempArr[i] = BernoulliPMF(x[i], p);
            }
            return tempArr;
        }

        public static double NextBernoulli(double p)
        {
            if (randObj.NextDouble() <= p)
                return 1.0;
            else
                return 0.0;
        }

        public static double[] NextBernoulli(int nLen, double p)
        {
            double[] tempArr = new double[nLen];
            for (int i = 0; i < nLen; i++)
            {
                tempArr[i] = NextBernoulli(p);
            }
            return tempArr;
        }

        static void TestBernoulli()
        {
            int nBins = 20;
            int nPoints = 10000;
            double probSuccess = 0.60;
            double displayScaleFactor = 500.0;

            double[] randObj = NextBernoulli(nPoints, probSuccess);
            double[] HistogramValues = makeHistogram(randObj, nBins);
            double[] xdata = new double[nBins];
            double[] ydata = new double[nBins];
            double[] yCalcDistribution = new double[nBins];
            for (int i = 0; i < nBins; i++)
            {
                xdata[i] = i;
                ydata[i] = (double)HistogramValues[i];
                yCalcDistribution[i] = BernoulliPMF(i, probSuccess);
            }

            double ydataMax = dataMax(ydata);
            double yCalcDistributionMax = dataMax(yCalcDistribution);

            Console.WriteLine("\nTesting the Bernoulli Probability Distribution");
            Console.WriteLine("\nKey: Bin number = *****... [random distrib.] [calculated distrib.]\n");
            for (int i = 0; i < nBins; i++)
            {
                Console.Write("Bin {0,2:n0} = ", i);
                for (int j = 0; j < (Math.Round(ydata[i] * displayScaleFactor / nBins / ydataMax, 0)); j++) Console.Write("*");
                Console.WriteLine("  [{0}] [{1}]", Math.Round(ydata[i] * 100.0 / ydataMax,0), 
                    Math.Round(yCalcDistribution[i] * 100.0 / yCalcDistributionMax, 0));
            }
            Console.WriteLine("\nPress ENTER key to continue...");
            Console.ReadLine();
            Console.Clear();
        }     

        #endregion

        #region Binomial Distribution

        public static double BinomialPMF(int x, int n, double p)
        {
            return Gamma(n + 1) * Math.Pow(p, x) * Math.Pow(1 - p, n - x) / Gamma(x + 1) / Gamma(n - x + 1);
        }

        public static double[] BinomialPMF(int[] x, int n, double p)
        {
            double[] tempArr = new double[x.Length];
            for (int i = 0; i < x.Length; i++)
            {
                tempArr[i] = BinomialPMF(x[i], n, p);
            }
            return tempArr;
        }

        public static double NextBinomial(int n, double p)
        {
            double total = 0.0;
            for (int i = 0; i < n; i++)
            {
                if (randObj.NextDouble() < p)
                {
                    total++;
                }
            }
            return total;
        }

        public static double[] NextBinomial(int n, double p, int nLen)
        {
            double[] tempArr = new double[nLen];
            for (int i = 0; i < nLen; i++)
            {
                tempArr[i] = NextBinomial(n, p);
            }
            return tempArr;
        }

        static void TestBinomial()
        {
            int nBins = 20;
            int nTrials = 10;
            int nPoints = 10000;            
            double probSuccess = 0.60;
            double displayScaleFactor = 500.0;

            double[] randObj = NextBinomial(nTrials, probSuccess, nPoints);
            double[] HistogramValues = makeHistogram(randObj, nBins);
            double[] xdata = new double[nBins];
            double[] ydata = new double[nBins];
            double[] yCalcDistribution = new double[nBins];
            for (int i = 0; i < nTrials; i++)
            {
                xdata[i] = i;
                ydata[i] = (double)HistogramValues[i];
                yCalcDistribution[i] = BinomialPMF(i, nTrials, probSuccess);
            }

            double ydataMax = dataMax(ydata);
            double yCalcDistributionMax = dataMax(yCalcDistribution);

            Console.WriteLine("\nTesting the Binomial Probability Distribution");
            Console.WriteLine("\nKey: Bin number = *****... [random distrib.] [calculated distrib.]\n");          
            for (int i = 0; i < nBins; i++)
            {
                Console.Write("Bin {0,2:n0} = ", i);
                for (int j = 0; j < (Math.Round(ydata[i] * displayScaleFactor / nBins / ydataMax, 0)); j++) Console.Write("*");
                Console.WriteLine("  [{0}] [{1}]", Math.Round(ydata[i] * 100.0 / ydataMax, 0),
                    Math.Round(yCalcDistribution[i] * 100.0 / yCalcDistributionMax, 0));
            }
            Console.WriteLine("\nPress ENTER key to continue...");
            Console.ReadLine();
            Console.Clear();
        }      

        #endregion        

        #region Geometric Distribution

        public static double GeometricPMF(int x, double p)
        {
            if (x >= 1)
                return (p * Math.Pow(1 - p, x - 1));
            else
                return 0.0;
        }

        public static double[] GeometricPMF(int[] x, double p)
        {
            double[] tempArr = new double[x.Length];
            for (int i = 0; i < x.Length; i++)
            {
                tempArr[i] = GeometricPMF(x[i], p);
            }
            return tempArr;
        }

        public static double NextGeometric(double p)
        {
            return Math.Ceiling(Math.Log(randObj.NextDouble()) / Math.Log(1.0 - p));
        }

        public static double NextGeometric2(double p)
        {
            //Alternate method for calculating geometric random variates
            double rn = 1.0;
            for (; randObj.NextDouble() >= p; rn++) { }
            return rn;
        }

        public static double[] NextGeometric(int nLen, double p)
        {
            double[] tempArr = new double[nLen];
            for (int i = 0; i < nLen; i++)
            {
                tempArr[i] = NextGeometric(p);
            }
            return tempArr;
        }

        static void TestGeometric()
        {
          int nBins = 20;
          int nPoints = 10000;
          double probSuccess = 0.60;
          double displayScaleFactor = 500.0;

          double[] randObj = NextGeometric(nPoints, probSuccess);
          double[] HistogramValues = makeHistogram(randObj, nBins);
          double[] xdata = new double[nBins];
          double[] ydata = new double[nBins];
          double[] yCalcDistribution = new double[nBins];
          for (int i = 0; i < nBins; i++)
          {
            xdata[i] = i;
            ydata[i] = (double)HistogramValues[i];
            yCalcDistribution[i] = GeometricPMF(i, probSuccess);
          }

          double ydataMax = dataMax(ydata);
          double yCalcDistributionMax = dataMax(yCalcDistribution);

          Console.WriteLine("\nTesting the Geometric Probability Distribution");
          Console.WriteLine("\nKey: Bin number = *****... [random distrib.] [calculated distrib.]\n");
          for (int i = 0; i < nBins; i++)
          {
                Console.Write("Bin {0,2:n0} = ", i);
                for (int j = 0; j < (Math.Round(ydata[i]*displayScaleFactor / 
                nBins / ydataMax, 0)); j++) Console.Write("*");
                Console.WriteLine("  [{0}] [{1}]",Math.Round(ydata[i]*100.0 / 
                    ydataMax,0),Math.Round(yCalcDistribution[i]*100.0 / 
                    yCalcDistributionMax,0));
          }
          Console.WriteLine("\nPress ENTER key to continue...");
          Console.ReadLine();
          Console.Clear();
        }  

        #endregion                

        #region Negative Binomial Distribution

        public static double NegativeBinomialPMF(int r, double p, int x)
        {
            return (Gamma(r + x) / Gamma(r) / Gamma(x + 1)) * Math.Pow(p, r) * Math.Pow(1 - p, x);
        }

        public static double[] NegativeBinomialPMF(int r, double p, int[] x)
        {
            double[] tempArr = new double[x.Length];
            for (int i = 0; i < x.Length; i++)
            {
                tempArr[i] = NegativeBinomialPMF(r, p, x[i]);
            }
            return tempArr;
        }

        public static double NextNegativeBinomial(int r, double p, int x)
        {
            double y = NextGamma(r,(p / (1.0 - p)));
            return NextPoisson(y);
        }

        public static double[] NextNegativeBinomial(int r, double p, int x, int nLen)
        {
            double[] tempArr = new double[nLen];
            for (int i = 0; i < nLen; i++)
            {
                tempArr[i] = NextNegativeBinomial(r, p, x);
            }
            return tempArr;
        }

        static void TestNegativeBinomial()
        {
            int nBins = 20;
            int nPoints = 10000;
            double pVal = 0.5;
            int rVal = 10;
            int xVal = 10;
            double displayScaleFactor = 500.0;

            double[] randObj = NextNegativeBinomial(rVal, pVal, xVal, nPoints);
            double[] HistogramValues = makeHistogram(randObj, nBins);
            double[] xdata = new double[nBins];
            double[] ydata = new double[nBins];
            double[] yCalcDistribution = new double[nBins];
            for (int i = 0; i < nBins; i++)
            {
                xdata[i] = i;
                ydata[i] = (double)HistogramValues[i];
                yCalcDistribution[i] = NegativeBinomialPMF(rVal, pVal, i);
            }

            double ydataMax = dataMax(ydata);
            double yCalcDistributionMax = dataMax(yCalcDistribution);

            Console.WriteLine("\nTesting the Negative Binomial Probability Distribution");
            Console.WriteLine("\nKey: Bin number = *****... [random distrib.] [calculated distrib.]\n");
            for (int i = 0; i < nBins; i++)
            {
                Console.Write("Bin {0,2:n0} = ", i);
                for (int j = 0; j < (Math.Round(ydata[i] * displayScaleFactor / nBins / ydataMax, 0)); j++) Console.Write("*");
                Console.WriteLine("  [{0}] [{1}]", Math.Round(ydata[i] * 100.0 / ydataMax, 0),
                    Math.Round(yCalcDistribution[i] * 100.0 / yCalcDistributionMax, 0));
            }
            Console.WriteLine("\nPress ENTER key to continue...");
            Console.ReadLine();
            Console.Clear();
        }

        #endregion

        #region Poisson Distribution

        public static double PoissonPMF(int x, double lambda)
        {
            return Math.Exp(-lambda) * Math.Pow(lambda, x) / Gamma(x + 1);
        }

        public static double[] PoissonPMF(int[] x, double lambda)
        {
            double[] tempArray = new double[x.Length];
            for (int i = 0; i < x.Length; i++)
            {
                tempArray[i] = PoissonPMF(x[i], lambda);
            }
            return tempArray;
        }

        public static double NextPoisson(double lambda)
        {
            //Using a recursive algorithm
            int i = 0;
            double w = Math.Exp(-lambda);
            double v = w;
            double u = randObj.NextDouble();
            while (w <= u)
            {
                v = v * (lambda / (double)(i + 1));
                w = w + v;
                i = i + 1;
            }
            return i;
        }

        public static double NextPoisson2(double lambda)
        {
            //Using algorithm proposed by Knuth
            //see http://en.wikipedia.org/wiki/Poisson_distribution
            int k = 0;
            double p = 1.0;
            double L = Math.Exp(-lambda);
            do
            {
                k++;
                p *= randObj.NextDouble();
            } while (p >= L);
            return k - 1;
        }

        public static double NextPoisson3(double lambda)
        {
            //Using algorithm proposed by Kemp (1981)
            int i = 0;
            double w = Math.Exp(-lambda);
            double u = randObj.NextDouble();
            while (u > w)
            {
                u = u - w;
                i = i + 1;
                w = w * lambda / i;
            }
            return i;
        }

        public static double[] NextPoisson(double lambda, int nLength)
        {
            double[] tempArr = new double[nLength];
            for (int i = 0; i < nLength; i++)
            {
                tempArr[i] = NextPoisson(lambda);
            }
            return tempArr;
        }

        static void TestPoisson()
        {
            int nBins = 20;
            int nPoints = 10000;
            double lambdaValue = 4.0; 
            double displayScaleFactor = 500.0;

            double[] randObj = NextPoisson(lambdaValue, nPoints);
            double[] HistogramValues = makeHistogram(randObj, nBins);
            double[] xdata = new double[nBins];
            double[] ydata = new double[nBins];
            double[] yCalcDistribution = new double[nBins];
            for (int i = 0; i < nBins; i++)
            {
                xdata[i] = i;
                ydata[i] = (double)HistogramValues[i];
                yCalcDistribution[i] = PoissonPMF(i, lambdaValue);
            }

            double ydataMax = dataMax(ydata);
            double yCalcDistributionMax = dataMax(yCalcDistribution);

            Console.WriteLine("\nTesting the Poisson Probability Distribution");
            Console.WriteLine("\nKey: Bin number = *****... [random distrib.] [calculated distrib.]\n");
            for (int i = 0; i < nBins; i++)
            {
                Console.Write("Bin {0,2:n0} = ", i);
                for (int j = 0; j < (Math.Round(ydata[i] * displayScaleFactor / nBins / ydataMax, 0)); j++) Console.Write("*");
                Console.WriteLine("  [{0}] [{1}]", Math.Round(ydata[i] * 100.0 / ydataMax, 0),
                    Math.Round(yCalcDistribution[i] * 100.0 / yCalcDistributionMax, 0));
            }
            Console.WriteLine("\nPress ENTER key to continue...");
            Console.ReadLine();
            Console.Clear();
        }

        #endregion

        #region Uniform Discrete Distribution

        public static double UniformDiscretePMF(double x, double a, double b)
        {
            if ((x >= a) && (x < b))
                return (1.0 / (b - a + 1));
            else
                return 0.0;
        }

        public static double[] UniformDiscretePMF(double[] x, double a, double b)
        {
            double[] tempArr = new double[x.Length];
            for (int i = 0; i < x.Length; i++)
            {
                tempArr[i] = UniformDiscretePMF(x[i], a, b);
            }
            return tempArr;
        }

        public static double NextUniformDiscrete(double a, double b)
        {
            return Math.Floor((a + (b - a + 1) * randObj.NextDouble()));
        }

        public static double[] NextUniformDiscrete(int nLen, double a, double b)
        {
            double[] tempArr = new double[nLen];
            for (int i = 0; i < nLen; i++)
            {
                tempArr[i] = NextUniformDiscrete(a, b);
            }
            return tempArr;
        }

        static void TestUniformDiscrete()
        {
            int nBins = 20;
            int nPoints = 10000;
            double xmin = 0.0;
            double xmax = 20.0;
            double a = 5.0;
            double b = 15.0;
            double displayScaleFactor = 500.0;

            double[] randObj = NextUniformDiscrete(nPoints, a, b);
            double[] HistogramValues = makeHistogram(randObj, xmin, xmax, nBins);
            double[] xdata = new double[nBins];
            double[] ydata = new double[nBins];
            double[] yCalcDistribution = new double[nBins];

            for (int i = 0; i < nBins; i++)
            {
                xdata[i] = i;
                ydata[i] = (double)HistogramValues[i];
                yCalcDistribution[i] = UniformDiscretePMF(i, a, b);
            }

            double ydataMax = dataMax(ydata);
            double yCalcDistributionMax = dataMax(yCalcDistribution);

            Console.WriteLine("\nTesting the Uniform (Discrete) Probability Distribution");
            Console.WriteLine("\nKey: Bin number = *****... [random distrib.] [calculated distrib.]\n");
            for (int i = 0; i < nBins; i++)
            {
                Console.Write("Bin {0,2:n0} = ", i);
                for (int j = 0; j < (Math.Round(ydata[i] * displayScaleFactor / nBins / ydataMax, 0)); j++) Console.Write("*");
                Console.WriteLine("  [{0}] [{1}]", Math.Round(ydata[i] * 100.0 / ydataMax, 0),
                    Math.Round(yCalcDistribution[i] * 100.0 / yCalcDistributionMax, 0));
            }
            Console.WriteLine("\nPress ENTER key to continue...");
            Console.ReadLine();
            Console.Clear();
        }

        #endregion

        //Continuous Distributions

        #region Beta Distribution

        public static double BetaPDF(double x, double a, double b)
        {
            double bTemp = Beta(a, b);
            return Math.Pow(x, a - 1) * Math.Pow(1 - x, b - 1) / bTemp;
        }

        public static double[] BetaPDF(double[] x, double a, double b)
        {
            double[] tempArray = new double[x.Length];
            for (int i = 0; i < x.Length; i++)
            {
                tempArray[i] = BetaPDF(x[i], a, b);
            }
            return tempArray;
        }

        public static double NextBeta(int a, int b)
        {
            double gamma1 = NextGamma(a, 1);
            double gamma2 = NextGamma(b, 1);
            return gamma1 / (gamma1 + gamma2);
        }

        public static double[] NextBeta(int a, int b, int nLength)
        {
            double[] tempArr = new double[nLength];
            for (int i = 0; i < nLength; i++)
            {
                tempArr[i] = NextBeta(a, b);
            }
            return tempArr;
        }

        static void TestBeta()
        {
            int nBins = 20;
            int nPoints = 10000;
            double xmin = 0.0;
            double xmax = 1.0;
            double aVal = 2.0;
            double bVal = 5.0;
            double displayScaleFactor = 500.0;

            double[] randObj = NextBeta((int)aVal, (int)bVal, nPoints);
            double[] HistogramValues = makeHistogram(randObj, xmin, xmax, nBins);
            double[] xdata = new double[nBins];
            double[] ydata = new double[nBins];
            double[] yCalcDistribution = new double[nBins];
            for (int i = 0; i < nBins; i++)
            {
                xdata[i] = xmin + (i + 0.5) * (xmax - xmin) / nBins;
                ydata[i] = (double)HistogramValues[i];
                yCalcDistribution[i] = BetaPDF(xdata[i], aVal, bVal);
            }

            double ydataMax = dataMax(ydata);
            double yCalcDistributionMax = dataMax(yCalcDistribution);

            Console.WriteLine("\nTesting the Beta Probability Distribution");
            Console.WriteLine("\nKey: Bin number = *****... [random distrib.] [calculated distrib.]\n");
            for (int i = 0; i < nBins; i++)
            {
                Console.Write("Bin {0,2:n0} = ", i);
                for (int j = 0; j < (Math.Round(ydata[i] * displayScaleFactor / nBins / ydataMax, 0)); j++) Console.Write("*");
                Console.WriteLine("  [{0}] [{1}]", Math.Round(ydata[i] * 100.0 / ydataMax, 0),
                    Math.Round(yCalcDistribution[i] * 100.0 / yCalcDistributionMax, 0));
            }
            Console.WriteLine("\nPress ENTER key to continue...");
            Console.ReadLine();
            Console.Clear();
        }

        #endregion

        #region Beta Prime Distribution

        public static double BetaPrimePDF(double x, double a, double b)
        {
            return Math.Pow(x, a - 1) * Math.Pow(1 + x, -a - b) / Beta(a, b);
        }

        public static double[] BetaPrimePDF(double[] x, double a, double b)
        {
            double[] tempArray = new double[x.Length];
            for (int i = 0; i < x.Length; i++)
            {
                tempArray[i] = BetaPDF(x[i], a, b);
            }
            return tempArray;
        }

        public static double NextBetaPrime(int a, int b)
        {
           double betaVariate = NextBeta(a, b);
           return betaVariate / (1.0 - betaVariate);
        }

        public static double[] NextBetaPrime(int a, int b, int nLen)
        {
            double[] tempArr = new double[nLen];
            for (int i = 0; i < nLen; i++)
            {
                tempArr[i] = NextBetaPrime(a, b);
            }
            return tempArr;
        }

        static void TestBetaPrime()
        {
            int nBins = 20;
            int nPoints = 10000;
            double xmin = 0.0;
            double xmax = 1.0;
            int aVal = 3;
            int bVal = 6;
            double displayScaleFactor = 500.0;

            double[] randObj = NextBetaPrime(aVal, bVal, nPoints);
            double[] HistogramValues = makeHistogram(randObj, xmin, xmax, nBins);
            double[] xdata = new double[nBins];
            double[] ydata = new double[nBins];
            double[] yCalcDistribution = new double[nBins];
            for (int i = 0; i < nBins; i++)
            {
                xdata[i] = xmin + (i + 0.5) * (xmax - xmin) / nBins;
                ydata[i] = (double)HistogramValues[i];
                yCalcDistribution[i] = BetaPrimePDF(xdata[i], aVal, bVal);
            }

            double ydataMax = dataMax(ydata);
            double yCalcDistributionMax = dataMax(yCalcDistribution);

            Console.WriteLine("\nTesting the Beta Prime Probability Distribution");
            Console.WriteLine("\nKey: Bin number = *****... [random distrib.] [calculated distrib.]\n");
            for (int i = 0; i < nBins; i++)
            {
                Console.Write("Bin {0,2:n0} = ", i);
                for (int j = 0; j < (Math.Round(ydata[i] * displayScaleFactor / nBins / ydataMax, 0)); j++) Console.Write("*");
                Console.WriteLine("  [{0}] [{1}]", Math.Round(ydata[i] * 100.0 / ydataMax, 0),
                    Math.Round(yCalcDistribution[i] * 100.0 / yCalcDistributionMax, 0));
            }
            Console.WriteLine("\nPress ENTER key to continue...");
            Console.ReadLine();
            Console.Clear();
        }

        #endregion

        #region Cauchy Distribution

        public static double CauchyPDF(double x, double a, double b)
        {
            return b / (Math.PI * (b * b + (x - a) * (x - a)));
        }

        public static double[] CauchyPDF(double[] x, double a, double b)
        {
            double[] tempArr = new double[x.Length];
            for (int i = 0; i < x.Length; i++)
            {
                tempArr[i] = CauchyPDF(x[i], a, b);
            }
            return tempArr;
        }

        public static double NextCauchy(double a, double b)
        {
            return a + b * (Math.Tan(Math.PI * randObj.NextDouble() - 0.5));
        }

        public static double[] NextCauchy(double a, double b, int nLength)
        {
            double[] tempArr = new double[nLength];
            for (int i = 0; i < nLength; i++)
            {
                tempArr[i] = NextCauchy(a, b);
            }
            return tempArr;
        }

        static void TestCauchy()
        {
            int nBins = 20;
            int nPoints = 10000;
            double xmin = -5;
            double xmax = 5;
            double aVal = 0.0;
            double bVal = 0.5;
            double displayScaleFactor = 500.0;

            double[] randObj = NextCauchy(aVal, bVal, nPoints);
            double[] HistogramValues = makeHistogram(randObj, xmin, xmax, nBins);
            double[] xdata = new double[nBins];
            double[] ydata = new double[nBins];
            double[] yCalcDistribution = new double[nBins];
            for (int i = 0; i < nBins; i++)
            {
                xdata[i] = xmin + (i + 0.5) * (xmax - xmin) / nBins;
                ydata[i] = HistogramValues[i];
                yCalcDistribution[i] = CauchyPDF(xdata[i], aVal, bVal);
            }

            double ydataMax = dataMax(ydata);
            double yCalcDistributionMax = dataMax(yCalcDistribution);

            Console.WriteLine("\nTesting the Cauchy Probability Distribution");
            Console.WriteLine("\nKey: Bin number = *****... [random distrib.] [calculated distrib.]\n");
            for (int i = 0; i < nBins; i++)
            {
                Console.Write("Bin {0,2:n0} = ", i);
                for (int j = 0; j < (Math.Round(ydata[i] * displayScaleFactor / nBins / ydataMax, 0)); j++) Console.Write("*");
                Console.WriteLine("  [{0}] [{1}]", Math.Round(ydata[i] * 100.0 / ydataMax, 0),
                    Math.Round(yCalcDistribution[i] * 100.0 / yCalcDistributionMax, 0));
            }
            Console.WriteLine("\nPress ENTER key to continue...");
            Console.ReadLine();
            Console.Clear();
        }

        #endregion

        #region Chi and ChiSquare Distribution

        public static double ChiPDF(double x, int n)
        {
            double gamma = Gamma(n / 2.0);
            double exp = Math.Exp(-n * x * x / 2.0);
            return 2.0 * Math.Pow(n / 2.0, n / 2) * Math.Pow(x, n - 1) * exp / Math.Pow(2, n / 2) / gamma;
        }

        public static double[] ChiPDF(double[] x, int n)
        {
            double[] tempArr = new double[x.Length];
            for (int i = 0; i < x.Length; i++)
            {
                tempArr[i] = ChiPDF(x[i], n);
            }
            return tempArr;
        }

        public static double ChiPDF(double x, int n, double sigma)
        {
            double gamma = Gamma(n / 2.0);
            double exp = Math.Exp(-n * x * x / 2.0 / sigma / sigma);
            return 2.0 * Math.Pow(n / 2.0, n / 2) * Math.Pow(x, n - 1) * exp / Math.Pow(2, n / 2) / gamma / Math.Pow(sigma, n);
        }

        public static double[] ChiPDF(double[] x, int n, double sigma)
        {
            double[] tempArr = new double[x.Length];
            for (int i = 0; i < x.Length; i++)
            {
                tempArr[i] = ChiPDF(x[i], n, sigma);
            }
            return tempArr;
        }

        public static double NextChi(int n)
        {
            return Math.Sqrt(NextChiSquare(n) / n);
        }

        public static double[] NextChi(int n, int nLen)
        {
            double[] tempArr = new double[nLen];
            for (int i = 0; i < nLen; i++)
            {
                tempArr[i] = NextChi(n);
            }
            return tempArr;
        }

        public static double NextChi(int n, double sigma)
        {
            return Math.Sqrt(NextChiSquare(n, sigma) / n);
        }

        public static double[] NextChi(int n, double sigma, int nLen)
        {
            double[] tempArr = new double[nLen];
            for (int i = 0; i < nLen; i++)
            {
                tempArr[i] = NextChi(n, sigma);
            }
            return tempArr;
        }

        public static double ChiSquarePDF(double x, int n)
        {
            double gamma = Gamma(n / 2.0);
            double exp = Math.Exp(-x / 2.0);
            return Math.Pow(x, n / 2 - 1) * exp / Math.Pow(2, n / 2) / gamma;
        }

        public static double[] ChiSquarePDF(double[] x, int n)
        {
            double[] tempArr = new double[x.Length];
            for (int i = 0; i < x.Length; i++)
            {
                tempArr[i] = ChiSquarePDF(x[i], n);
            }
            return tempArr;
        }

        public static double ChiSquarePDF(double x, int n, double sigma)
        {
            double gamma = Gamma(n / 2.0);
            double exp = Math.Exp(-x / 2.0 / sigma / sigma);
            return Math.Pow(x, n / 2 - 1) * exp / Math.Pow(2, n / 2) / gamma / Math.Pow(sigma, n);
        }

        public static double[] ChiSquarePDF(double[] x, int n, double sigma)
        {
            double[] tempArr = new double[x.Length];
            for (int i = 0; i < x.Length; i++)
            {
                tempArr[i] = ChiSquarePDF(x[i], n, sigma);
            }
            return tempArr;
        }

        public static double NextChiSquare(int n)
        {
            double sum = 0.0;
            for (int i = 0; i < n; i++)
            {
                sum += Math.Pow(NextNormal(0, 1), 2);
            }
            return sum;
        }

        public static double[] NextChiSquare(int n, int nLen)
        {
            double[] tempArr = new double[nLen];
            for (int i = 0; i < nLen; i++)
            {
                tempArr[i] = NextChiSquare(n);
            }
            return tempArr;
        }

        public static double NextChiSquare(int n, double sigma)
        {
            double sum = 0.0;
            for (int i = 0; i < n; i++)
            {
                sum += Math.Pow(NextNormal(0, sigma * sigma), 2);
            }
            return sum;
        }

        public static double[] NextChiSquare(int n, double sigma, int nLen)
        {
            double[] tempArr = new double[nLen];
            for (int i = 0; i < nLen; i++)
            {
                tempArr[i] = NextChiSquare(n, sigma);
            }
            return tempArr;
        }

        static void TestChiChiSquare()
        {
            int nBins = 20;
            int nPoints = 10000;
            double xmin = 0;
            double xmax = 5;
            double sigma = 1.0;
            int nVal = 2;
            double displayScaleFactor = 500.0;

            double[] randChi = NextChi(nVal, sigma, nPoints);
            double[] randChiSquare = NextChiSquare(nVal, sigma, nPoints);
            double[] RandomDistributionValuesChi = makeHistogram(randChi, xmin, xmax, nBins);
            double[] RandomDistributionValuesChiSquare = makeHistogram(randChiSquare, xmin, xmax, nBins);
            double[] xdata = new double[nBins];
            double[] ydataChi = new double[nBins];
            double[] ydataChiSquare = new double[nBins];
            double[] yChiCalcDistribution = new double[nBins];
            double[] yChiSquareCalcDistribution = new double[nBins];

            for (int i = 0; i < nBins; i++)
            {
                xdata[i] = xmin + (i + 0.5) * (xmax - xmin) / nBins;
                ydataChi[i] = (double)RandomDistributionValuesChi[i];
                ydataChiSquare[i] = (double)RandomDistributionValuesChiSquare[i];
                yChiCalcDistribution[i] = ChiPDF(xdata[i], nVal, sigma);
                yChiSquareCalcDistribution[i] = ChiSquarePDF(xdata[i], nVal, sigma);
            }

            double ydataChiMax = dataMax(ydataChi);
            double yChiCalcDistributionMax = dataMax(yChiCalcDistribution);

            double ydataChiSquareMax = dataMax(ydataChiSquare);
            double yChiSquareCalcDistributionMax = dataMax(yChiSquareCalcDistribution);

            Console.WriteLine("\nTesting the Chi Probability Distribution");
            Console.WriteLine("\nKey: Bin number = *****... [random distrib.] [calculated distrib.]\n");
            for (int i = 0; i < nBins; i++)
            {
                Console.Write("Bin {0,2:n0} = ", i);
                for (int j = 0; j < (Math.Round(ydataChi[i] * displayScaleFactor / nBins / ydataChiMax, 0)); j++) Console.Write("*");
                Console.WriteLine("  [{0}] [{1}]", Math.Round(ydataChi[i] * 100.0 / ydataChiMax, 0),
                    Math.Round(yChiCalcDistribution[i] * 100.0 / yChiCalcDistributionMax, 0));
            }

            Console.WriteLine("\nTesting the Chi Square Probability Distribution");
            Console.WriteLine("\nKey: Bin number = *****... [random distrib.] [calculated distrib.]\n");
            for (int i = 0; i < nBins; i++)
            {
                Console.Write("Bin {0,2:n0} = ", i);
                for (int j = 0; j < (Math.Round(ydataChiSquare[i] * displayScaleFactor / nBins / ydataChiSquareMax, 0)); j++) Console.Write("*");
                Console.WriteLine("  [{0}] [{1}]", Math.Round(ydataChiSquare[i] * 100.0 / ydataChiSquareMax, 0),
                    Math.Round(yChiSquareCalcDistribution[i] * 100.0 / yChiSquareCalcDistributionMax, 0));
            }
            Console.WriteLine("\nPress ENTER key to continue...");
            Console.ReadLine();
            Console.Clear();
        }

        #endregion

        #region Erlang Distribution

        public static double ErlangPDF(double x, double a, int b)
        {
            if (x >= 0.0)
                return (Math.Pow(x,b-1)*Math.Exp(-x/a))/(Gamma(b)*Math.Pow(a,b));
            else
                return 0.0;
        }

        public static double[] ErlangPDF(double[] x, double a, int b)
        {
            double[] tempArr = new double[x.Length];
            for (int i = 0; i < x.Length; i++)
            {
                tempArr[i] = ErlangPDF(x[i], a, b);
            }
            return tempArr;
        }

        public static double NextErlang(double a, int b)
        {
            double product = 1.0;
            for (int i = 0; i < b; i++)
            {
                product *= randObj.NextDouble();
            }
            return (-a) * Math.Log(product);
        }

        public static double[] NextErlang(double a, int b, int nLen)
        {
            double[] tempArr = new double[nLen];
            for (int i = 0; i < nLen; i++)
            {
                tempArr[i] = NextErlang(a,b);
            }
            return tempArr;
        }

        static void TestErlang()
        {
            int nBins = 20;
            int nPoints = 10000;
            double xmin = 0;
            double xmax = 4;
            double aVal = 1.0;
            int bVal = 1;
            double displayScaleFactor = 500.0;

            double[] randObj = NextErlang(aVal, bVal, nPoints);
            double[] HistogramValues = makeHistogram(randObj, xmin, xmax, nBins);
            double[] xdata = new double[nBins];
            double[] ydata = new double[nBins];

            double[] yCalcDistribution = new double[nBins];
            for (int i = 0; i < nBins; i++)
            {
                xdata[i] = xmin + (i + 0.5) * (xmax - xmin) / nBins;
                ydata[i] = (double)HistogramValues[i];
                yCalcDistribution[i] = ErlangPDF(xdata[i], aVal, bVal);
            }

            double ydataMax = dataMax(ydata);
            double yCalcDistributionMax = dataMax(yCalcDistribution);

            Console.WriteLine("\nTesting the Erlang Probability Distribution");
            Console.WriteLine("\nKey: Bin number = *****... [random distrib.] [calculated distrib.]\n");
            for (int i = 0; i < nBins; i++)
            {
                Console.Write("Bin {0,2:n0} = ", i);
                for (int j = 0; j < (Math.Round(ydata[i] * displayScaleFactor / nBins / ydataMax, 0)); j++) Console.Write("*");
                Console.WriteLine("  [{0}] [{1}]", Math.Round(ydata[i] * 100.0 / ydataMax, 0),
                    Math.Round(yCalcDistribution[i] * 100.0 / yCalcDistributionMax, 0));
            }
            Console.WriteLine("\nPress ENTER key to continue...");
            Console.ReadLine();
            Console.Clear();
        }

        #endregion        

        #region Exponential Distribution

        public static double ExponentialPDF(double x, double a)
        {
            return a * Math.Exp(-a * x);
        }

        public static double[] ExponentialPDF(double[] x, double a)
        {
            double[] tempArr = new double[x.Length];
            for (int i = 0; i < x.Length; i++)
            {
                tempArr[i] = ExponentialPDF(x[i], a);
            }
            return tempArr;
        }

        public static double NextExponential(double a)
        {
            return -Math.Log(randObj.NextDouble()) / a;
        }

        public static double[] NextExponential(double a, int nLen)
        {
            double[] tempArr = new double[nLen];
            for (int i = 0; i < nLen; i++)
            {
                tempArr[i] = NextExponential(a);
            }
            return tempArr;
        }

        static void TestExponential()
        {
            int nBins = 20;
            int nPoints = 10000;
            double xmin = 0.0;
            double xmax = 4.0;
            double alphaVal = 1.5;
            double displayScaleFactor = 500.0;

            double[] randObj = NextExponential(alphaVal, nPoints);
            double[] HistogramValues = makeHistogram(randObj, xmin, xmax, nBins);
            double[] xdata = new double[nBins];
            double[] ydata = new double[nBins];

            double[] yCalcDistribution = new double[nBins];
            for (int i = 0; i < nBins; i++)
            {
                xdata[i] = xmin + (i + 0.5) * (xmax - xmin) / nBins;
                ydata[i] = (double)HistogramValues[i];
                yCalcDistribution[i] = ExponentialPDF(xdata[i], alphaVal);
            }

            double ydataMax = dataMax(ydata);
            double yCalcDistributionMax = dataMax(yCalcDistribution);

            Console.WriteLine("\nTesting the Exponential Probability Distribution");
            Console.WriteLine("\nKey: Bin number = *****... [random distrib.] [calculated distrib.]\n");
            for (int i = 0; i < nBins; i++)
            {
                Console.Write("Bin {0,2:n0} = ", i);
                for (int j = 0; j < (Math.Round(ydata[i] * displayScaleFactor / nBins / ydataMax, 0)); j++) Console.Write("*");
                Console.WriteLine("  [{0}] [{1}]", Math.Round(ydata[i] * 100.0 / ydataMax, 0),
                    Math.Round(yCalcDistribution[i] * 100.0 / yCalcDistributionMax, 0));
            }
            Console.WriteLine("\nPress ENTER key to continue...");
            Console.ReadLine();
            Console.Clear();
        }

        #endregion

        #region Extreme Value Distribution

        public static double ExtremeValuePDF(double x, double mu, double sigma)
        {
            double w = (x-mu)/sigma;
            return (1.0 / sigma) * Math.Exp(-w) * Math.Exp(-Math.Exp(-w));
        }

        public static double[] ExtremeValuePDF(double[] x, double mu, double sigma)
        {
            double[] tempArr = new double[x.Length];
            for (int i = 0; i < x.Length; i++)
            {
                tempArr[i] = ExtremeValuePDF(x[i], mu, sigma);
            }
            return tempArr;
        }

        public static double NextExtremeValue(double mu, double sigma)
        {
            return mu - sigma * Math.Log(-Math.Log(randObj.NextDouble()));
        }

        public static double[] NextExtremeValue(double mu, double sigma, int nLen)
        {
            double[] tempArr = new double[nLen];
            for (int i = 0; i < nLen; i++)
            {
                tempArr[i] = NextExtremeValue(mu, sigma);
            }
            return tempArr;
        }

        static void TestExtremeValue()
        {
            int nBins = 20;
            int nPoints = 10000;
            double xmin =-4;
            double xmax = 8;
            double muVal = 1.0;
            double sigmaVal = 0.750;
            double displayScaleFactor = 500.0;

            double[] randObj = NextExtremeValue(muVal,sigmaVal, nPoints);
            double[] HistogramValues = makeHistogram(randObj, xmin, xmax, nBins);
            double[] xdata = new double[nBins];
            double[] ydata = new double[nBins];

            double[] yCalcDistribution = new double[nBins];
            for (int i = 0; i < nBins; i++)
            {
                xdata[i] = xmin + (i + 0.5) * (xmax - xmin) / nBins;
                ydata[i] = (double)HistogramValues[i];
                yCalcDistribution[i] = ExtremeValuePDF(xdata[i], muVal,sigmaVal);
            }

            double ydataMax = dataMax(ydata);
            double yCalcDistributionMax = dataMax(yCalcDistribution);

            Console.WriteLine("\nTesting the Extreme Value Probability Distribution");
            Console.WriteLine("\nKey: Bin number = *****... [random distrib.] [calculated distrib.]\n");
            for (int i = 0; i < nBins; i++)
            {
                Console.Write("Bin {0,2:n0} = ", i);
                for (int j = 0; j < (Math.Round(ydata[i] * displayScaleFactor / nBins / ydataMax, 0)); j++) Console.Write("*");
                Console.WriteLine("  [{0}] [{1}]", Math.Round(ydata[i] * 100.0 / ydataMax, 0),
                    Math.Round(yCalcDistribution[i] * 100.0 / yCalcDistributionMax, 0));
            }
            Console.WriteLine("\nPress ENTER key to continue...");
            Console.ReadLine();
            Console.Clear();
        }

        #endregion        

        #region Gamma Distribution

        public static double GammaPDF(double x, int b, double a)
        {
            return Math.Pow(a, b) * Math.Pow(x, b - 1) * Math.Exp(-a * x) / Gamma(b);
        }

        public static double[] GammaPDF(double[] x, int b, double a)
        {
            double[] tempArray = new double[x.Length];
            for (int i = 0; i < x.Length; i++)
            {
                tempArray[i] = GammaPDF(x[i], b, a);
            }
            return tempArray;
        }

        public static double NextGamma(int b, double a)
        {
            double temp = 0.0;
            for (int i = 0; i < b; i++)
            {
                temp += -Math.Log(randObj.NextDouble()) / a;
            }
            return temp;
        }

        public static double[] NextGamma(int b, double a, int nLen)
        {
            double[] tempArr = new double[nLen];
            for (int i = 0; i < nLen; i++)
            {
                tempArr[i] = NextGamma(b, a);
            }
            return tempArr;
        }

        static void TestGamma()
        {
            int nBins = 20;
            int nPoints = 10000;
            double xmin = 0.0;
            double xmax = 15.0;
            double aVal = 0.6;
            int bVal = 2;
            double displayScaleFactor = 500.0;

            double[] randObj = NextGamma(bVal, aVal, nPoints);
            double[] HistogramValues = makeHistogram(randObj, xmin, xmax, nBins);
            double[] xdata = new double[nBins];
            double[] ydata = new double[nBins];
            double[] yCalcDistribution = new double[nBins];
            for (int i = 0; i < nBins; i++)
            {
                xdata[i] = xmin + (i + 0.5) * (xmax - xmin) / nBins;
                ydata[i] = (double)HistogramValues[i];
                yCalcDistribution[i] = GammaPDF(xdata[i], bVal, aVal);
            }

            double ydataMax = dataMax(ydata);
            double yCalcDistributionMax = dataMax(yCalcDistribution);

            Console.WriteLine("\nTesting the Gamma Probability Distribution");
            Console.WriteLine("\nKey: Bin number = *****... [random distrib.] [calculated distrib.]\n");
            for (int i = 0; i < nBins; i++)
            {
                Console.Write("Bin {0,2:n0} = ", i);
                for (int j = 0; j < (Math.Round(ydata[i] * displayScaleFactor / nBins / ydataMax, 0)); j++) Console.Write("*");
                Console.WriteLine("  [{0}] [{1}]", Math.Round(ydata[i] * 100.0 / ydataMax, 0),
                    Math.Round(yCalcDistribution[i] * 100.0 / yCalcDistributionMax, 0));
            }
            Console.WriteLine("\nPress ENTER key to continue...");
            Console.ReadLine();
            Console.Clear();
        }

        #endregion

        #region Laplace Distribution

        public static double LaplacePDF(double x, double sigma)
        {
            double coef = Math.Sqrt(2.0) / sigma;
            return 0.5 * coef * Math.Exp(-coef * Math.Abs(x));
        }

        public static double[] LaplacePDF(double[] x, double sigma)
        {
            double[] tempArr = new double[x.Length];
            for (int i = 0; i < x.Length; i++)
            {
                tempArr[i] = LaplacePDF(x[i], sigma);
            }
            return tempArr;
        }

        public static double NextLaplace(double sigma)
        {
            double coef = sigma / Math.Sqrt(2.0);
            double u = randObj.NextDouble();
            if ((u > 0.0) && (u < 0.5))
                return coef * Math.Log(2.0 * u);
            else if ((u >= 0.5) && (u < 1.0))
                return coef * Math.Log(1.0 / (2.0 * (1.0 - u)));
            else return 0.0;
        }

        public static double[] NextLaplace(double sigma, int nLen)
        {
            double[] tempArr = new double[nLen];
            for (int i = 0; i < nLen; i++)
            {
                tempArr[i] = NextLaplace(sigma);
            }
            return tempArr;
        }

        static void TestLaplace()
        {
            int nBins = 20;
            int nPoints = 10000;
            double xmin = -5.0;
            double xmax = 5.0;
            double sigmaVal = 2.0;
            double displayScaleFactor = 500.0;

            double[] randObj = NextLaplace(sigmaVal, nPoints);
            double[] HistogramValues = makeHistogram(randObj, xmin, xmax, nBins);
            double[] xdata = new double[nBins];
            double[] ydata = new double[nBins];

            double[] yCalcDistribution = new double[nBins];
            for (int i = 0; i < nBins; i++)
            {
                xdata[i] = xmin + (i + 0.5) * (xmax - xmin) / nBins;
                ydata[i] = (double)HistogramValues[i];
                yCalcDistribution[i] = LaplacePDF(xdata[i], sigmaVal);
            }

            double ydataMax = dataMax(ydata);
            double yCalcDistributionMax = dataMax(yCalcDistribution);

            Console.WriteLine("\nTesting the Laplace Probability Distribution");
            Console.WriteLine("\nKey: Bin number = *****... [random distrib.] [calculated distrib.]\n");
            for (int i = 0; i < nBins; i++)
            {
                Console.Write("Bin {0,2:n0} = ", i);
                for (int j = 0; j < (Math.Round(ydata[i] * displayScaleFactor / nBins / ydataMax, 0)); j++) Console.Write("*");
                Console.WriteLine("  [{0}] [{1}]", Math.Round(ydata[i] * 100.0 / ydataMax, 0),
                    Math.Round(yCalcDistribution[i] * 100.0 / yCalcDistributionMax, 0));
            }
            Console.WriteLine("\nPress ENTER key to continue...");
            Console.ReadLine();
            Console.Clear();
        }

        #endregion

        #region Logistic Disbribution

        public static double LogisticPDF(double x, double b, double a)
        {
            return Math.Exp(-(x - a) / b) / (b * Math.Pow((1.0 + Math.Exp(-(x - a) / b)), 2.0));
        }

        public static double[] LogisticPDF(double[] x, double b, double a)
        {
            double[] tempArr = new double[x.Length];
            for (int i = 0; i < x.Length; i++)
            {
                tempArr[i] = LogisticPDF(x[i], b, a);
            }
            return tempArr;
        }

        public static double NextLogistic(double b, double a)
        {
            double u = randObj.NextDouble();
            return a + b * Math.Log(u / (1 - u));
        }

        public static double[] NextLogistic(double b, double a, int nLen)
        {
            double[] tempArr = new double[nLen];
            for (int i = 0; i < nLen; i++)
            {
                tempArr[i] = NextLogistic(b, a);
            }
            return tempArr;
        }

        static void TestLogistic()
        {
            int nBins = 20;
            int nPoints = 10000;
            double xmin = -5;
            double xmax = 15;
            double aVal = 2.0;
            double bVal = 1.0;
            double displayScaleFactor = 500.0;

            double[] randObj = NextLogistic(bVal, aVal, nPoints);
            double[] HistogramValues = makeHistogram(randObj, xmin, xmax, nBins);
            double[] xdata = new double[nBins];
            double[] ydata = new double[nBins];

            double[] yCalcDistribution = new double[nBins];
            for (int i = 0; i < nBins; i++)
            {
                xdata[i] = xmin + (i + 0.5) * (xmax - xmin) / nBins;
                ydata[i] = (double)HistogramValues[i];
                yCalcDistribution[i] = LogisticPDF(xdata[i], bVal, aVal);
            }

            double ydataMax = dataMax(ydata);
            double yCalcDistributionMax = dataMax(yCalcDistribution);

            Console.WriteLine("\nTesting the Logistic Probability Distribution");
            Console.WriteLine("\nKey: Bin number = *****... [random distrib.] [calculated distrib.]\n");
            for (int i = 0; i < nBins; i++)
            {
                Console.Write("Bin {0,2:n0} = ", i);
                for (int j = 0; j < (Math.Round(ydata[i] * displayScaleFactor / nBins / ydataMax, 0)); j++) Console.Write("*");
                Console.WriteLine("  [{0}] [{1}]", Math.Round(ydata[i] * 100.0 / ydataMax, 0),
                    Math.Round(yCalcDistribution[i] * 100.0 / yCalcDistributionMax, 0));
            }
            Console.WriteLine("\nPress ENTER key to continue...");
            Console.ReadLine();
            Console.Clear();
        }

        #endregion        

        #region Lognormal Distribution

        public static double LognormalPDF(double x, double mu, double sigma)
        {
            double x1, x2;
            if (x > 0.0)
            {
                x1 = 1.0 / (sigma * x * Math.Sqrt(2.0 * Math.PI));
                x2 = (Math.Log(x) - mu) * (Math.Log(x) - mu) / (2.0 * sigma * sigma);
                return x1 * Math.Exp(-x2);
            }
            else
                return 0.0;
        }

        public static double[] LognormalPDF(double[] x, double mu, double sigma)
        {
            double[] tempArr = new double[x.Length];
            for (int i = 0; i < x.Length; i++)
            {
                tempArr[i] = LognormalPDF(x[i], mu, sigma);
            }
            return tempArr;
        }

        public static double NextLognormal(double mu, double sigma)
        {
            return Math.Exp(mu + sigma * NextNormal(0.0, 1.0));
        }

        public static double[] NextLognormal(double mu, double sigma, int nLen)
        {
            double[] tempArr = new double[nLen];
            for (int i = 0; i < nLen; i++)
            {
                tempArr[i] = NextLognormal(mu, sigma);
            }
            return tempArr;
        }

        static void TestLognormal()
        {
            int nBins = 20;
            int nPoints = 10000;
            double xmin = 0.0;
            double xmax = 5.0;
            double mu = 0.0;
            double sigma =0.5;
            double displayScaleFactor = 500.0;

            double[] randObj = NextLognormal(mu, sigma, nPoints);
            double[] HistogramValues = makeHistogram(randObj, xmin, xmax, nBins);
            double[] xdata = new double[nBins];
            double[] ydata = new double[nBins];

            double[] yCalcDistribution = new double[nBins];
            for (int i = 0; i < nBins; i++)
            {
                xdata[i] = xmin + (i + 0.5) * (xmax - xmin) / nBins;
                ydata[i] = (double)HistogramValues[i];
                yCalcDistribution[i] = LognormalPDF(xdata[i], mu, sigma);
            }

            double ydataMax = dataMax(ydata);
            double yCalcDistributionMax = dataMax(yCalcDistribution);

            Console.WriteLine("\nTesting the Lognormal Probability Distribution");
            Console.WriteLine("\nKey: Bin number = *****... [random distrib.] [calculated distrib.]\n");

            for (int i = 0; i < nBins; i++)
            {
                Console.Write("Bin {0,2:n0} = ", i);
                for (int j = 0; j < (Math.Round(ydata[i] * displayScaleFactor / nBins / ydataMax, 0)); j++) Console.Write("*");
                Console.WriteLine("  [{0}] [{1}]", Math.Round(ydata[i] * 100.0 / ydataMax, 0),
                    Math.Round(yCalcDistribution[i] * 100.0 / yCalcDistributionMax, 0));
            }
            Console.WriteLine("\nPress ENTER key to continue...");
            Console.ReadLine();
            Console.Clear();
        }

        #endregion

        #region Normal Distribution

        public static double NormalPDF(double x, double mu, double sigma)
        {
            double x1, x2;
            x1 = 1 / sigma / Math.Sqrt(2 * Math.PI);
            x2 = (x - mu) * (x - mu) / (2 * sigma * sigma);
            return x1 * Math.Exp(-x2);
        }

        public static double[] NormalPDF(double[] x, double mu, double sigma)
        {
            double[] tempArr = new double[x.Length];
            for (int i = 0; i < x.Length; i++)
            {
                tempArr[i] = NormalPDF(x[i], mu, sigma);
            }
            return tempArr;
        }

        public static double NextNormal(double mu, double sigma)
        {
            double x = 2.0 * randObj.NextDouble() - 1.0;
            double y = 2.0 * randObj.NextDouble() - 1.0;
            double s = x * x + y * y;

            while (s > 1.0)
            {
                x = 2.0 * randObj.NextDouble() - 1.0;
                y = 2.0 * randObj.NextDouble() - 1.0;
                s = x * x + y * y;
            }

            double xGaussian = Math.Sqrt(-2.0 * Math.Log(s) / s) * x * sigma + mu;
            double yGaussian = Math.Sqrt(-2.0 * Math.Log(s) / s) * y * sigma + mu;

            return xGaussian;
        }

        public static double[] NextNormal(double mu, double sigma, int nLen)
        {
            double[] tempArr = new double[nLen];
            for (int i = 0; i < nLen; i++)
            {
                tempArr[i] = NextNormal(mu, sigma);
            }
            return tempArr;
        }

        static void TestNormal()
        {
            int nBins = 20;
            int nPoints = 10000;
            double xmin = 0.0;
            double xmax = 4.0;
            double mu = 2.0;
            double sigma = 0.5;
            double displayScaleFactor = 500.0;

            double[] randObj = NextNormal(mu, sigma, nPoints);
            double[] HistogramValues = makeHistogram(randObj, xmin, xmax, nBins);
            double[] xdata = new double[nBins];
            double[] ydata = new double[nBins];

            double[] yCalcDistribution = new double[nBins];
            for (int i = 0; i < nBins; i++)
            {
                xdata[i] = xmin + (i + 0.5) * (xmax - xmin) / nBins;
                ydata[i] = (double)HistogramValues[i];
                yCalcDistribution[i] = NormalPDF(xdata[i], mu, sigma);
            }

            double ydataMax = dataMax(ydata);
            double yCalcDistributionMax = dataMax(yCalcDistribution);

            Console.WriteLine("\nTesting the Normal (Gaussian) Probability Distribution");
            Console.WriteLine("\nKey: Bin number = *****... [random distrib.] [calculated distrib.]\n");

            for (int i = 0; i < nBins; i++)
            {
                Console.Write("Bin {0,2:n0} = ", i);
                for (int j = 0; j < (Math.Round(ydata[i] * displayScaleFactor / nBins / ydataMax, 0)); j++) Console.Write("*");
                Console.WriteLine("  [{0}] [{1}]", Math.Round(ydata[i] * 100.0 / ydataMax, 0),
                    Math.Round(yCalcDistribution[i] * 100.0 / yCalcDistributionMax, 0));
            }
            Console.WriteLine("\nPress ENTER key to continue...");
            Console.ReadLine();
            Console.Clear();
        }

        #endregion
        
        #region Pareto Distribution

        public static double ParetoPDF(double x, double a, double b)
        {
            if (x > a)
                return b * Math.Pow(a, b) * Math.Pow(x, -b - 1.0);
            else
                return 0.0;
        }

        public static double[] ParetoPDF(double[] x, double a, double b)
        {
            double[] tempArr = new double[x.Length];
            for (int i = 0; i < x.Length; i++)
            {
                tempArr[i] = ParetoPDF(x[i], a, b);
            }
            return tempArr;
        }

        public static double NextPareto(double a, double b)
        {
            return a * Math.Pow(randObj.NextDouble(), -1.0 / b);
        }

        public static double[] NextPareto(double a, double b, int nLen)
        {
            double[] tempArr = new double[nLen];
            for (int i = 0; i < nLen; i++)
            {
                tempArr[i] = NextPareto(a,b);
            }
            return tempArr;
        }

        static void TestPareto()
        {
            int nBins = 20;
            int nPoints = 10000;
            double xmin = 0.0;
            double xmax = 5.0;
            double bVal = 1.0;
            double aVal = 1.0;
            double displayScaleFactor = 500.0;

            double[] randObj = NextPareto(aVal, bVal, nPoints);
            double[] HistogramValues = makeHistogram(randObj, xmin, xmax, nBins);
            double[] xdata = new double[nBins];
            double[] ydata = new double[nBins];

            double[] yCalcDistribution = new double[nBins];
            for (int i = 0; i < nBins; i++)
            {
                xdata[i] = xmin + (i + 0.5) * (xmax - xmin) / nBins;
                ydata[i] = (double)HistogramValues[i];
                yCalcDistribution[i] = ParetoPDF(xdata[i], aVal, bVal);
            }

            double ydataMax = dataMax(ydata);
            double yCalcDistributionMax = dataMax(yCalcDistribution);

            Console.WriteLine("\nTesting the Pareto Probability Distribution");
            Console.WriteLine("\nKey: Bin number = *****... [random distrib.] [calculated distrib.]\n");
            for (int i = 0; i < nBins; i++)
            {
                Console.Write("Bin {0,2:n0} = ", i);
                for (int j = 0; j < (Math.Round(ydata[i] * displayScaleFactor / nBins / ydataMax, 0)); j++) Console.Write("*");
                Console.WriteLine("  [{0}] [{1}]", Math.Round(ydata[i] * 100.0 / ydataMax, 0),
                    Math.Round(yCalcDistribution[i] * 100.0 / yCalcDistributionMax, 0));
            }
            Console.WriteLine("\nPress ENTER key to continue...");
            Console.ReadLine();
            Console.Clear();
        }

        #endregion

        #region Rayleigh Distribution

        public static double RayleighPDF(double x, double sigma)
        {
            if (x >= 0.0)
                return (x / sigma / sigma) * Math.Exp(-x * x / 2.0 / sigma / sigma);
            else
                return 0.0;
        }

        public static double[] RayleighPDF(double[] x, double sigma)
        {
            double[] tempArr = new double[x.Length];
            for (int i = 0; i < x.Length; i++)
            {
                tempArr[i] = RayleighPDF(x[i], sigma);
            }
            return tempArr;
        }

        public static double NextRayleigh(double sigma)
        {
            double u = randObj.NextDouble();
            return sigma * Math.Sqrt(-2.0 * Math.Log(u));
        }

        public static double[] NextRayleigh(double sigma, int nLen)
        {
            double[] tempArr = new double[nLen];
            for (int i = 0; i < nLen; i++)
            {
                tempArr[i] = NextRayleigh(sigma);
            }
            return tempArr;
        }

        static void TestRayleigh()
        {
            int nBins = 20;
            int nPoints = 10000;
            double xmin = 0;
            double xmax = 4;
            double sigma = 0.5;
            double displayScaleFactor = 500.0;

            double[] randObj = NextRayleigh(sigma, nPoints);
            double[] HistogramValues = makeHistogram(randObj, xmin, xmax, nBins);
            double[] xdata = new double[nBins];
            double[] ydata = new double[nBins];

            double[] yCalcDistribution = new double[nBins];
            for (int i = 0; i < nBins; i++)
            {
                xdata[i] = xmin + (i + 0.5) * (xmax - xmin) / nBins;
                ydata[i] = (double)HistogramValues[i];
                yCalcDistribution[i] = RayleighPDF(xdata[i], sigma);
            }

            double ydataMax = dataMax(ydata);
            double yCalcDistributionMax = dataMax(yCalcDistribution);

            Console.WriteLine("\nTesting the Rayleigh Probability Distribution");
            Console.WriteLine("\nKey: Bin number = *****... [random distrib.] [calculated distrib.]\n");
            for (int i = 0; i < nBins; i++)
            {
                Console.Write("Bin {0,2:n0} = ", i);
                for (int j = 0; j < (Math.Round(ydata[i] * displayScaleFactor / nBins / ydataMax, 0)); j++) Console.Write("*");
                Console.WriteLine("  [{0}] [{1}]", Math.Round(ydata[i] * 100.0 / ydataMax, 0),
                    Math.Round(yCalcDistribution[i] * 100.0 / yCalcDistributionMax, 0));
            }
            Console.WriteLine("\nPress ENTER key to continue...");
            Console.ReadLine();
            Console.Clear();
        }

        #endregion

        #region Student-T Distribution

        public static double StudentTPDF(double x, int n)
        {
            double gamma1 = Gamma((n + 1.0) / 2.0);
            double gamma2 = Gamma(1.0 / 2.0);
            double gamma3 = Gamma(n / 2.0);
            return Math.Pow(n, -0.5) * Math.Pow(1 + x * x / n, -(n + 1) / 2) * gamma1 / gamma2 / gamma3;
        }

        public static double[] StudentTPDF(double[] x, int n)
        {
            double[] tempArr = new double[x.Length];
            for (int i = 0; i < x.Length; i++)
            {
                tempArr[i] = StudentTPDF(x[i], n);
            }
            return tempArr;
        }

        public static double NextStudentT(int n)
        {
            if (n <= 0) throw new Exception("n must be positive!");
            return NextNormal(0, 1) / NextChi(n);
        }

        public static double[] NextStudentT(int n, int nLen)
        {
            double[] tempArr = new double[nLen];
            for (int i = 0; i < nLen; i++)
            {
                tempArr[i] = NextStudentT(n);
            }
            return tempArr;
        }

        static void TestStudentT()
        {
            int nBins = 20;
            int nPoints = 10000;
            double xmin = -7.0;
            double xmax = 7.0;
            int nVal = 5;
            double displayScaleFactor = 500.0;

            double[] randObj = NextStudentT(nVal, nPoints);
            double[] HistogramValues = makeHistogram(randObj, xmin, xmax, nBins);
            double[] xdata = new double[nBins];
            double[] ydata = new double[nBins];
            double[] yCalcDistribution = new double[nBins];
            for (int i = 0; i < nBins; i++)
            {
                xdata[i] = xmin + (i + 0.5) * (xmax - xmin) / nBins;
                ydata[i] = HistogramValues[i];
                yCalcDistribution[i] = StudentTPDF(xdata[i], nVal);
            }

            double ydataMax = dataMax(ydata);
            double yCalcDistributionMax = dataMax(yCalcDistribution);

            Console.WriteLine("\nTesting the Student-T Probability Distribution");
            Console.WriteLine("\nKey: Bin number = *****... [random distrib.] [calculated distrib.]\n");
            for (int i = 0; i < nBins; i++)
            {
                Console.Write("Bin {0,2:n0} = ", i);
                for (int j = 0; j < (Math.Round(ydata[i] * displayScaleFactor / nBins / ydataMax, 0)); j++) Console.Write("*");
                Console.WriteLine("  [{0}] [{1}]", Math.Round(ydata[i] * 100.0 / ydataMax, 0),
                    Math.Round(yCalcDistribution[i] * 100.0 / yCalcDistributionMax, 0));
            }
            Console.WriteLine("\nPress ENTER key to continue...");
            Console.ReadLine();
            Console.Clear();
        }

        #endregion

        #region Triangular Distribution

        public static double TriangularPMF(double a, double b, double c, double x)
        {
            if ((a <= x) && (x <= c))
            { return 2.0 * (x - a) / (b - a) / (c - a); }
            else if ((c <= x) && (x <= b))
            { return 2.0 * (b - x) / (b - a) / (b - c); }
            else return 0.0;
        }

        public static double[] TriangularPMF(double a, double b, double c, int[] x)
        {
            double[] tempArr = new double[x.Length];
            for (int i = 0; i < x.Length; i++)
            {
                tempArr[i] = TriangularPMF(a, b, c, x[i]);
            }
            return tempArr;
        }

        public static double NextTriangular(double a, double b, double c)
        {
            double u = randObj.NextDouble();
            if (u <= ((c - a) / (b - a)))
            {
                return a + Math.Sqrt(u) * Math.Sqrt((c - a) * (b - a));
            }
            else
            {
                return b - Math.Sqrt(u * (b - a) - (c - a)) * (b - c);
            }
        }

        public static double[] NextTriangular(double a, double b, double c, int nLen)
        {
            double[] tempArr = new double[nLen];
            for (int i = 0; i < nLen; i++)
            {
                tempArr[i] = NextTriangular(a, b, c);
            }
            return tempArr;
        }

        static void TestTriangular()
        {
            int nBins = 20;
            int nPoints = 10000;
            double aVal = 0.0;
            double bVal = 50;
            double cVal = 100;
            double xmin = 5.0;
            double xmax = 15.0;
            double displayScaleFactor = 500.0;

            double[] randObj = NextTriangular(aVal, bVal, cVal, nPoints);
            double[] HistogramValues = makeHistogram(randObj,dataMin(randObj), dataMax(randObj), nBins);
            double[] xdata = new double[nBins];
            double[] ydata = new double[nBins];
            double[] yCalcDistribution = new double[nBins];
            for (int i = 0; i < nBins; i++)
            {
                xdata[i] = xmin + (i + 0.5) * (xmax - xmin) / nBins; 
                ydata[i] = (double)HistogramValues[i];
                yCalcDistribution[i] = TriangularPMF(aVal, bVal, cVal, i);
            }

            double ydataMax = dataMax(ydata);
            double yCalcDistributionMax = dataMax(yCalcDistribution);

            Console.WriteLine("\nTesting the Triangular Probability Distribution");
            Console.WriteLine("\nKey: Bin number = *****... [random distrib.] [calculated distrib.]\n");
            for (int i = 0; i < nBins; i++)
            {
                Console.Write("Bin {0,2:n0} = ", i);
                for (int j = 0; j < (Math.Round(ydata[i] * displayScaleFactor / nBins / ydataMax, 0)); j++) Console.Write("*");
                Console.WriteLine("  [{0}] [{1}]", Math.Round(ydata[i] * 100.0 / ydataMax, 0), 
                    Math.Round(yCalcDistribution[i] * 100.0 / yCalcDistributionMax, 0));
            }
            Console.WriteLine("\nPress ENTER key to continue...");
            Console.ReadLine();
            Console.Clear();
        }

        #endregion

        #region Uniform Continuous Distribution

        public static double UniformContinuousPDF(double x, double a, double b)
        {
            if ((x >= a) && (x < b))
                return (1.0 / (b - a));
            else
                return 0.0;
        }

        public static double[] UniformContinuousPDF(double[] x, double a, double b)
        {
            double[] tempArr = new double[x.Length];
            for (int i = 0; i < x.Length; i++)
            { tempArr[i] = UniformContinuousPDF(x[i], a, b); }
            return tempArr;
        }

        public static double NextUniformContinuous(double a, double b)
        {
            return (a + (b - a) * randObj.NextDouble());
        }

        public static double[] NextUniformContinuous(int nLen, double a, double b)
        {
            double[] tempArr = new double[nLen];
            for (int i = 0; i < nLen; i++)
            { tempArr[i] = NextUniformContinuous(a, b); }
            return tempArr;
        }

        static void TestUniformContinuous()
        {
            int nBins = 20;
            int nPoints = 10000;
            double xmin = 0.0;
            double xmax = 20.0;
            double a = 5.0;
            double b = 15.0;
            double displayScaleFactor = 500.0;
            double[] randObj = NextUniformContinuous(nPoints, a, b);
            double[] HistogramValues = makeHistogram(randObj, xmin, xmax, nBins);
            double[] xdata = new double[nBins];
            double[] ydata = new double[nBins];
            double[] yCalcDistribution = new double[nBins];
            for (int i = 0; i < nBins; i++)
            {
                xdata[i] = xmin + (i + 0.5) * (xmax - xmin) / nBins;
                ydata[i] = (double)HistogramValues[i];
                yCalcDistribution[i] = UniformContinuousPDF(i, a, b);
            }
            double ydataMax = dataMax(ydata);
            double yCalcDistributionMax = dataMax(yCalcDistribution);
            Console.WriteLine("\nTesting the Uniform (Continuous) Probability Distribution");
            Console.WriteLine("\nKey: Bin number = *****... [random distrib.] [calculated distrib.]\n");
            for (int i = 0; i < nBins; i++)
            {
                Console.Write("Bin {0,2:n0} = ", i);
                for (int j = 0; j < (Math.Round(ydata[i] * displayScaleFactor / nBins / ydataMax, 0)); j++) Console.Write("*");
                Console.WriteLine("  [{0}] [{1}]", Math.Round(ydata[i] * 100.0 / ydataMax, 0), 
                    Math.Round(yCalcDistribution[i] * 100.0 / yCalcDistributionMax, 0));
            }
            Console.WriteLine("\nPress ENTER key to continue...");
            Console.ReadLine(); Console.Clear();
        }

        #endregion

        #region Weibull Distribution

        public static double WeibullPDF(double x, double a, double b)
        {
            if (x > 0.0)
                return (a / b) * Math.Pow(x / b, a - 1) * Math.Exp(-Math.Pow(x / b, a));
            else
                return 0.0;
        }

        public static double[] WeibullPDF(double[] x, double a, double b)
        {
            double[] tempArr = new double[x.Length];
            for (int i = 0; i < x.Length; i++)
            { tempArr[i] = WeibullPDF(x[i], a, b); }
            return tempArr;
        }

        public static double NextWeibull(double a, double b)
        {
            return b * Math.Pow(-Math.Log(randObj.NextDouble()), 1.0 / a);
        }

        public static double[] NextWeibull(double a, double b, int nLen)
        {
            double[] tempArr = new double[nLen];
            for (int i = 0; i < nLen; i++)
            { tempArr[i] = NextWeibull(a, b); }
            return tempArr;
        }

        static void TestWeibull()
        {
            int nBins = 20; int nPoints = 10000; double xmin = 0.0;
            double xmax = 4.0; double aVal = 2.0; double bVal = 1.0;
            double displayScaleFactor = 500.0;
            double[] randObj = NextWeibull(aVal, bVal, nPoints);
            double[] HistogramValues = makeHistogram(randObj, xmin, xmax, nBins);
            double[] xdata = new double[nBins];
            double[] ydata = new double[nBins];
            double[] yCalcDistribution = new double[nBins];
            for (int i = 0; i < nBins; i++)
            {
                xdata[i] = xmin + (i + 0.5) * (xmax - xmin) / nBins;
                ydata[i] = (double)HistogramValues[i];
                yCalcDistribution[i] = WeibullPDF(xdata[i], aVal, bVal);
            }
            double ydataMax = dataMax(ydata);
            double yCalcDistributionMax = dataMax(yCalcDistribution);
            Console.WriteLine("\nTesting the Weibull Distribution");
            Console.WriteLine("\nKey: Bin number = *****... [random distrib.] [calculated distrib.]\n");
            for (int i = 0; i < nBins; i++)
            {
                Console.Write("Bin {0,2:n0} = ", i);
                for (int j = 0; j < (Math.Round(ydata[i] * displayScaleFactor / nBins / ydataMax, 0)); j++) Console.Write("*");
                Console.WriteLine("  [{0}] [{1}]", Math.Round(ydata[i] * 100.0 / ydataMax, 0), 
                    Math.Round(yCalcDistribution[i] * 100.0 / yCalcDistributionMax, 0));
            }
            Console.WriteLine("\nPress ENTER key to continue...");
            Console.ReadLine(); Console.Clear();
        }

        #endregion        

        //Random permutations and perturbations

        #region Random Permutations

        //Does a random permutation on a one-dimensional input array
        //with input array and number of permutations chosen by the user.
        public static int[] RandomPermutation(int[] numbers, int n)
        {
            //Exchange each entry of the one-dimensional array with  
            //another entry located at a random position in the array.
            for (int i = numbers.Length - 1; i > 0; i--)
            {
                int randomPosition = randObj.Next(i + 1);
                int temp = numbers[i];
                numbers[i] = numbers[randomPosition];
                numbers[randomPosition] = temp;
            }
            return numbers;
        }

        public static void TestRandomPermutation()
        {
            //Setup a simple one-dimensional array for testing permutations
            int arraySize = 6;

            Console.WriteLine("Testing random permutation of array with {0} elements\n", arraySize);
            int[] arrayTest = new int[arraySize];
            for (int i = 0; i < arraySize; i++)
            {
                arrayTest[i] = i;
            }
            //and display this one-dimensional array on the screen
            Console.WriteLine("Original input array configuration\n");
            foreach (int j in arrayTest)
                Console.Write("{0}    ", j);
            Console.WriteLine("\n\n");

            //Do some arbitrary number of random permutations on the input array
            int nPermutations = 4;
            Console.WriteLine("Results after performing {0} random permutations of original input array\n", nPermutations);
            int k = 0;
            do
            {
                int[] x = RandomPermutation(arrayTest, arraySize);
                foreach (int d in x)
                    Console.Write("{0}    ", d);
                Console.WriteLine("\n");
                k++;
            } while (k < nPermutations);

            Console.WriteLine("\nPress ENTER key to continue...");
            Console.ReadLine();
            Console.Clear();
        }

        #endregion

        #region Random Perturbations

        //Does a random permutation on a one-dimensional input array
        //with a default initialized array and an abitrary number of 
        //permutations as chosen by the user.
        public static int[] RandomPermutation(int n)
        {
            //Create an array of size specified by nSize
            int[] numbers = new int[n];
            //to hold the numbers 0,1,2,...,n
            for (int i = 0; i < numbers.Length; i++)
            {
                numbers[i] = i;
            }
            //Exchange each entry of the array with another 
            //entry located at a random position in the array.
            for (int i = numbers.Length - 1; i > 0; i--)
            {
                int randomPosition = randObj.Next(i + 1);
                int temp = numbers[i];
                numbers[i] = numbers[randomPosition];
                numbers[randomPosition] = temp;
            }
            return numbers;
        }

        public static double[] RandomPerturbation(double[] data)
        {
            //Randomize the positions of all input data array indices
            int[] randomizedIndicesArray = RandomPermutation(data.Length);
            //then just pick one arbitrary position and add a little 
            //amount of perturbation to it. The normal distribution was
            //chosen arbitrarily. You can substitute another probablity
            //distribution in its place if so desired.
            data[randomizedIndicesArray[0]] += NextNormal(0, 1) / 100.0;
            return data;
        }

        public static void TestRandomPertubation()
        {
            //Testing perturbations of entries in a simple one-dimensional array
            int arraySize = 6;
            int nPerturbations = 5;
            Console.WriteLine("Testing {0} perturbation(s) on array with {0} elements\n", nPerturbations, arraySize);

            double[] arrayTest = new double[arraySize];
            for (int i = 0; i < arraySize; i++)
            {
                arrayTest[i] = i;
            }
            //and display this one-dimensional array on the screen
            Console.WriteLine("Original input array configuration\n");
            foreach (double j in arrayTest)
                Console.Write("{0}    ", j);
            Console.WriteLine("\n\n");

            Console.WriteLine("Results after performing {0} random perturbations of original input array\n", nPerturbations);
            //Then perturb the input data array a number of times
            int k = 0;
            do
            {
                //Initialize input data array
                for (int i = 0; i < arraySize; i++)
                {
                    arrayTest[i] = i;
                }
                //and then perturb it
                double[] y = RandomPerturbation(arrayTest);
                //and print the output results
                foreach (double c in y)
                    Console.Write("{0}    ", c);
                Console.WriteLine("\n");
                k++;
            } while (k < nPerturbations);
            Console.WriteLine("\n\nPress ENTER key to continue...");
            Console.ReadLine();
            Console.Clear();
        }

        #endregion

        //Adding random noise to data

        #region Add Random Noise to Data

        //Purpose: Add random noise to data

        //General function routine to process a single data point by
        //some user supplied function. In this example, f(x) = 2x.
        public static double f(double x)
        {
            return 2 * x;
        }

        //General routine to process an array of data points provided
        //by some arbitrary function.
        public static double[] f(double[] x)
        {
            double[] tempArr = new double[x.Length];
            for (int i = 0; i < x.Length; i++)
            {
                tempArr[i] = f(x[i]);
            }
            return tempArr;
        }

        //General routine for calculating a grid of n points (n > 1)
        //between a set of minimum and maximum values. It the evaluates
        //the user supplied function f(x) at each of those points. 
        public static void f(int n, double xLeft, double xRight,
                    out double[] x, out double[] y)
        {
            double dx = (xRight - xLeft) / (n - 1);

            double[] xtemp = new double[n];
            double[] ytemp = new double[n];

            for (int i = 0; i < n; i++)
            {
                xtemp[i] = xLeft + i * dx;
                ytemp[i] = f(xtemp[i]);
            }

            x = xtemp;
            y = ytemp;
        }

        //Adds random noise between -magtitude and +magnitude to the data
        //in array data[]. Random data source can be easily changed to
        //whatever distribution desired.
        public static double[] AddNoise(double[] data, double magnitude)
        {
            for (int i = 0; i < data.Length; i++)
            {
                //Get a random number between -1 and 1
                double r = 1.0 - 2.0 * randObj.NextDouble();
                data[i] = data[i] + magnitude * r;
            }
            return data;
        }

        //Driver routine to test adding random noise to data
        public static void TestAddRandomNoiseToData()
        {
            Console.WriteLine("Results: Add Random Noise Algorithm\n");

            int n = 20; //Number of grid points
            double xLeft = -5.0; //Leftmost starting point
            double xRight = 5.0; //Rightmost ending point                
            
            double[] x = new double[n];
            double[] y = new double[n];
            
            //Create the data arrays
            f(n, xLeft, xRight, out x, out y);
            
            Console.WriteLine("Data BEFORE adding noise\n");
            Console.WriteLine("x values\t y values\n");
            
            for (int i = 0; i < n; i++)
                Console.WriteLine("{0} \t\t {1}", Math.Round(x[i], 3), Math.Round(y[i], 3));
            
            double[] y_noisy = new double[n];
            
            //Magnitude was purposely made very large for this example
            double noise_magnitude = 10.0;
            
            //Add some random noise to the data array just created
            y_noisy = AddNoise(y, noise_magnitude);
            
            Console.WriteLine("\n\nData AFTER adding noise\n");
            Console.WriteLine("x values\t y values\n");
            
            for (int i = 0; i < n; i++)
                Console.WriteLine("{0} \t\t {1}", Math.Round(x[i], 3), Math.Round(y_noisy[i], 3));

            Console.WriteLine("\nPress ENTER key to continue...");
            Console.ReadLine(); Console.Clear();
        }

        #endregion

        //Removing random noise from data

        #region Remove Random Noise From Data

        //Purpose: Filter random noise from data

        //Returns the index <= i at which signal values can start to
        //contribute to the filtered value of data element i.
        public static int FirstIndex(int i, int width)
        {
            if (i >= width / 2)
                return (i - width / 2);
            else
                return 0;
        }

        //Returns the index >= i beyond which signal values cease to 
        //contribute to the filtered value of data element i.
        public static int LastIndex(int i, int width, int size)
        {
            if (i + width / 2 < size)
                return (i + width / 2);
            else
                return (size - 1);
        }

        //Filters the input signal using a moving average algorithm 
        //of specified width. The width must be odd.
        public static double[] MovingAverage(double[] RoughData, int width)
        {
            double[] y = new double[RoughData.Length];
            if (width % 2 == 0)
            {
                Console.WriteLine("Error: The width must be odd.");
                return y;
            }

            for (int i = 0; i < RoughData.Length; i++)
            {
                y[i] = 0.0;
                int start = FirstIndex(i, width);
                int stop = LastIndex(i, width, RoughData.Length);

                for (int j = start; j <= stop; j++)
                {
                    y[i] = y[i] + RoughData[j];
                }
                y[i] = y[i] / (stop - start + 1);
            }
            return y;
        }

        //Driver routine to test filtering random noise from data
        //using a moving average algorithm.
        private static void TestRemoveRandomNoiseFromData()
        {
            double[] RoughData =
              new double[] {  75.475, 75.600, 75.100, 73.525, 73.275, 
                73.325, 73.225, 74.350, 73.600, 74.475, 75.975, 76.850, 
                77.525, 78.200, 79.225, 78.850, 76.225, 76.850, 76.725, 76.100};

            Console.WriteLine("Results: Data Filtering Algorithm\n");

            Console.WriteLine("Rough Data\n");
            for (int i = 0; i < RoughData.Length; i++)
                Console.WriteLine("{0}", Math.Round(RoughData[i], 3));

            //Width of filter
            int width = 5;

            //Array to hold the filtered data
            double[] SmoothData = new double[RoughData.Length];

            //Calculate the filtered data values using a moving average
            //algorithm of the specified rough data and filter width.
            SmoothData = MovingAverage(RoughData, width);

            //Display the results on the screen or save the results
            //to a file right here.
            Console.WriteLine("\n\nSmoothed Data\n");
            for (int i = 0; i < RoughData.Length; i++)
                Console.WriteLine("{0}", Math.Round(SmoothData[i], 3));

            Console.WriteLine("\nPress ENTER key to continue...");
            Console.ReadLine(); Console.Clear();
        }
        #endregion

        static void Main(string[] args)
        {
            TestBasicStats();

            //Discrete Distributions            
            TestBernoulli();
            TestBinomial();
            TestGeometric();
            TestNegativeBinomial();
            TestPoisson();
            TestUniformDiscrete();            

            //Continuous Distributions
            TestBeta();
            TestBetaPrime();
            TestCauchy();
            TestChiChiSquare();           
            TestErlang();
            TestExponential();
            TestExtremeValue();
            TestGamma(); 
            TestLaplace();
            TestLogistic();
            TestLognormal();
            TestNormal();                        
            TestPareto();
            TestRayleigh();
            TestStudentT();
            TestTriangular();
            TestUniformContinuous();
            TestWeibull();            

            //Random Permutations
            TestRandomPermutation();
            TestRandomPertubation();

            //Random Noise
            TestAddRandomNoiseToData();
            TestRemoveRandomNoiseFromData();
        }
    }
}




