﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace Chapter11_CodeExamples
{
    class Program
    {
        public delegate double Function(double x);

        #region Forward Derivatives

        public static double DerivativeForward1(Function f, double x, double h)
        {
            h = (h == 0) ? 0.01 : h;
            return (-3 * f(x) + 4 * f(x + h) - f(x + 2 * h)) / 2 / h;
        }

        public static double DerivativeForward1(double[] y, int i, double h)
        {
            if (y == null || y.Length < 3 || i < 0 || i > y.Length - 3 || h == 0)
                return double.NaN;
            return (-3 * y[i] + 4 * y[i + 1] - y[i + 2]) / 2 / h;
        }

        public static double DerivativeForward2(Function f, double x, double h)
        {
            h = (h == 0) ? 0.01 : h;
            return (2 * f(x) - 5 * f(x + h) + 4 * f(x + 2 * h) - f(x + 3 * h)) / h / h;
        }

        public static double DerivativeForward2(double[] y, int i, double h)
        {
            if (y == null || y.Length < 4 || i < 0 || i > y.Length - 4 || h == 0)
                return double.NaN;
            return (2 * y[i] - 5 * y[i + 1] + 4 * y[i + 2] - y[i + 3]) / h / h;
        }

        public static double DerivativeForward3(Function f, double x, double h)
        {
            h = (h == 0) ? 0.01 : h;
            return (-5 * f(x) + 18 * f(x + h) - 24 * f(x + 2 * h) + 14 * f(x + 3 * h) - 3 * f(x + 4 * h)) / 2 / h / h / h;
        }

        public static double DerivativeForward3(double[] y, int i, double h)
        {
            if (y == null || y.Length < 5 || i < 0 || i > y.Length - 5 || h == 0)
                return double.NaN;
            return (-5 * y[i] + 18 * y[i + 1] - 24 * y[i + 2] + 14 * y[i + 3] -
                     3 * y[i + 4]) / 2 / h / h / h;
        }

        public static double DerivativeForward4(Function f, double x, double h)
        {
            h = (h == 0) ? 0.01 : h;
            return (3 * f(x) - 14 * f(x + h) + 26 * f(x + 2 * h) - 24 * f(x + 3 * h) + 11 * f(x + 4 * h) - 2 * f(x + 5 * h)) / h / h / h / h;
        }

        public static double DerivativeForward4(double[] y, int i, double h)
        {
            if (y == null || y.Length < 6 || i < 0 || i > y.Length - 6 || h == 0)
                return double.NaN;
            return (3 * y[i] - 14 * y[i + 1] + 26 * y[i + 2] - 24 * y[i + 3] +
                    11 * y[i + 4] - 2 * y[i + 5]) / h / h / h / h;
        }

        #endregion

        #region Backward Derivatives

        public static double DerivativeBackward1(Function f, double x, double h)
        {
            h = (h == 0) ? 0.01 : h;
            return (3 * f(x) - 4 * f(x - h) + f(x - 2 * h)) / 2 / h;
        }

        public static double DerivativeBackward1(double[] y, int i, double h)
        {
            if (y == null || y.Length < 3 || i < 0 || i < 3 || h == 0)
                return double.NaN;
            return (3 * y[i] - 4 * y[i - 1] + y[i - 2]) / 2 / h;
        }

        public static double DerivativeBackward2(Function f, double x, double h)
        {
            h = (h == 0) ? 0.01 : h;
            return (2 * f(x) - 5 * f(x - h) + 4 * f(x - 2 * h) - f(x - 3 * h)) / h / h;
        }

        public static double DerivativeBackward2(double[] y, int i, double h)
        {
            if (y == null || y.Length < 4 || i < 0 || i < 4 || h == 0)
                return double.NaN;
            return (2 * y[i] - 5 * y[i - 1] + 4 * y[i - 2] - y[i - 3]) / h / h;
        }

        public static double DerivativeBackward3(Function f, double x, double h)
        {
            h = (h == 0) ? 0.01 : h;
            return (5 * f(x) - 18 * f(x - h) + 24 * f(x - 2 * h) - 14 * f(x - 3 * h) + 3 * f(x - 4 * h)) / 2 / h / h / h;
        }

        public static double DerivativeBackward3(double[] y, int i, double h)
        {
            if (y == null || y.Length < 5 || i < 0 || i < 5 || h == 0)
                return double.NaN;
            return (5 * y[i] - 18 * y[i - 1] + 24 * y[i - 2] - 14 * y[i - 3] +
                     3 * y[i - 4]) / 2 / h / h / h;
        }

        public static double DerivativeBackward4(Function f, double x, double h)
        {
            h = (h == 0) ? 0.01 : h;
            return (3 * f(x) - 14 * f(x - h) + 26 * f(x - 2 * h) - 24 * f(x - 3 * h) + 11 * f(x - 4 * h) - 2 * f(x - 5 * h)) / h / h / h / h;
        }

        public static double DerivativeBackward4(double[] y, int i, double h)
        {
            if (y == null || y.Length < 6 || i < 0 || i < 6 || h == 0)
                return double.NaN;
            return (3 * y[i] - 14 * y[i - 1] + 26 * y[i - 2] - 24 * y[i - 3] +
                    11 * y[i - 4] - 2 * y[i - 5]) / h / h / h / h;
        }

        #endregion

        #region Central Derivatives

        public static double DerivativeCentral1(Function f, double x, double h)
        {
            h = (h == 0) ? 0.01 : h;
            return (f(x + h) - f(x - h)) / 2 / h;
        }

        public static double DerivativeCentral1(double[] y, int i, double h)
        {
            if (y == null || y.Length < 3 || i < 1 || i > y.Length - 2 || h == 0)
                return double.NaN;
            return (y[i + 1] - y[i - 1]) / 2 / h;
        }

        public static double DerivativeCentral2(Function f, double x, double h)
        {
            h = (h == 0) ? 0.01 : h;
            return (f(x - h) - 2 * f(x) + f(x + h)) / h / h;
        }

        public static double DerivativeCentral2(double[] y, int i, double h)
        {
            if (y == null || y.Length < 3 || i < 1 || i > y.Length - 2 || h == 0)
                return double.NaN;
            return (y[i - 1] - 2 * y[i] + y[i + 1]) / h / h;
        }

        public static double DerivativeCentral3(Function f, double x, double h)
        {
            h = (h == 0) ? 0.01 : h;
            return (-f(x - 2 * h) + 2 * f(x - h) - 2 * f(x + h) + f(x + 2 * h)) / 2 / h / h / h;
        }

        public static double DerivativeCentral3(double[] y, int i, double h)
        {
            if (y == null || y.Length < 5 || i < 2 || i > y.Length - 3 || h == 0)
                return double.NaN;
            return (-y[i - 2] + 2 * y[i - 1] - 2 * y[i + 1] + y[i + 2]) / 2 / h / h / h;
        }

        public static double DerivativeCentral4(Function f, double x, double h)
        {
            h = (h == 0) ? 0.01 : h;
            return (f(x - 2 * h) - 4 * f(x - h) + 6 * f(x) - 4 * f(x + h) + f(x + 2 * h)) / h / h / h / h;
        }

        public static double DerivativeCentral4(double[] y, int i, double h)
        {
            if (y == null || y.Length < 5 || i < 2 || i > y.Length - 3 || h == 0)
                return double.NaN;
            return (y[i - 2] - 4 * y[i - 1] + 6 * y[i] -
                    4 * y[i + 1] + y[i + 2]) / h / h / h / h;
        }

        #endregion

        #region Central Oh4 Derivatives

        public static double DerivativeOh4Central1(Function f, double x, double h)
        {
            h = (h == 0) ? 0.01 : h;
            return (f(x - 2 * h) - 8 * f(x - h) + 8 * f(x + h) - f(x + 2 * h)) / 12 / h;
        }

        public static double DerivativeOh4Central1(double[] y, int i, double h)
        {
            if (y == null || y.Length < 5 || i < 2 || i > y.Length - 3 || h == 0)
                return double.NaN;
            return (y[i - 2] - 8 * y[i - 1] + 8 * y[i + 1] - y[i + 2]) / 12 / h;
        }

        public static double DerivativeOh4Central2(Function f, double x, double h)
        {
            h = (h == 0) ? 0.01 : h;
            return (-f(x - 2 * h) + 16 * f(x - h) - 30 * f(x) + 16 * f(x + h) - f(x + 2 * h)) / 12 / h / h;
        }

        public static double DerivativeOh4Central2(double[] y, int i, double h)
        {
            if (y == null || y.Length < 5 || i < 2 || i > y.Length - 3 || h == 0)
                return double.NaN;
            return (-y[i - 2] + 16 * y[i - 1] - 30 * y[i] + 16 * y[i + 1] - 
                     y[i + 2]) / 12 / h / h;
        }

        public static double DerivativeOh4Central3(Function f, double x, double h)
        {
            h = (h == 0) ? 0.01 : h;
            return (f(x - 3 * h) - 8 * f(x - 2 * h) + 13 * f(x - h) - 13 * f(x + h) + 
                    8 * f(x + 2 * h) - f(x + 3 * h)) / 8 / h / h / h;
        }

        public static double DerivativeOh4Central3(double[] y, int i, double h)
        {
            if (y == null || y.Length < 7 || i < 3 || i > y.Length - 4 || h == 0)
                return double.NaN;
            return (y[i - 3] - 8 * y[i - 2] + 13 * y[i - 1] - 13 * y[i + 1] + 
                    8 * y[i + 2] - y[i + 3]) / 8 / h / h / h;
        }

        public static double DerivativeOh4Central4(Function f, double x, double h)
        {
            h = (h == 0) ? 0.01 : h;
            return (-f(x - 3 * h) + 12 * f(x - 2 * h) - 39 * f(x - h) + 56 * f(x) - 
                    39 * f(x + h) + 12 * f(x + 2 * h) - f(x + 3 * h)) / 6 / h / h / h / h;
        }

        public static double DerivativeOh4Central4(double[] y, int i, double h)
        {
            if (y == null || y.Length < 7 || i < 3 || i > y.Length - 4 || h == 0)
                return double.NaN;
            return (-y[i - 3] + 12 * y[i - 2] - 39 * y[i - 1] + 56 * y[i] - 
                    39 * y[i + 1] + 12 * y[i + 2] - y[i + 3]) / 6 / h / h / h / h;
        }

        #endregion

        #region Richardson Extrapolation
        public static double DerivativeRichardson1(Function f, double x, double h, string flag)
        {
            double result = double.NaN;
            switch (flag)
            {
                case "Backward": 
                    result = (4 * DerivativeBackward1(f, x, h / 2) - DerivativeBackward1(f, x, h)) / 3;
                break;
                case "Forward": 
                    result = (4 * DerivativeForward1(f, x, h / 2) - DerivativeForward1(f, x, h)) / 3;
                break;
                case "Central": 
                    result = (4 * DerivativeCentral1(f, x, h / 2) - DerivativeCentral1(f, x, h)) / 3;
                break;
                default: 
                    result = double.NaN;
                break;                
            }      
            return result;
        }

        public static double DerivativeRichardson1(double[] y, int i, double h, string flag)
        {
            double result = double.NaN;
            switch (flag)
            {
                case "Backward":
                    result = (4 * DerivativeBackward1(y, i, h / 2) - DerivativeBackward1(y, i, h)) / 3;
                    break;
                case "Forward":
                    result = (4 * DerivativeForward1(y, i, h / 2) - DerivativeForward1(y, i, h)) / 3;
                    break;
                case "Central":
                    result = (4 * DerivativeCentral1(y, i, h / 2) - DerivativeCentral1(y, i, h)) / 3;
                    break;
                default:
                    result = double.NaN;
                    break;
            }
            return result;
        }

        public static double DerivativeRichardson2(Function f, double x, double h, string flag)
        {
            double result = double.NaN;
            switch (flag)
            {
                case "Backward":
                    result = (4 * DerivativeBackward2(f, x, h / 2) - DerivativeBackward2(f, x, h)) / 3;
                    break;
                case "Forward":
                    result = (4 * DerivativeForward2(f, x, h / 2) - DerivativeForward2(f, x, h)) / 3;
                    break;
                case "Central":
                    result = (4 * DerivativeCentral2(f, x, h / 2) - DerivativeCentral2(f, x, h)) / 3;
                    break;
                default:
                    result = double.NaN;
                    break;
            }
            return result;
        }

        public static double DerivativeRichardson2(double[] y, int i, double h, string flag)
        {
            double result = double.NaN;
            switch (flag)
            {
                case "Backward":
                    result = (4 * DerivativeBackward2(y, i, h / 2) - DerivativeBackward2(y, i, h)) / 3;
                    break;
                case "Forward":
                    result = (4 * DerivativeForward2(y, i, h / 2) - DerivativeForward2(y, i, h)) / 3;
                    break;
                case "Central":
                    result = (4 * DerivativeCentral2(y, i, h / 2) - DerivativeCentral2(y, i, h)) / 3;
                    break;
                default:
                    result = double.NaN;
                    break;
            }
            return result;
        }

        public static double DerivativeRichardson3(Function f, double x, double h, string flag)
        {
            double result = double.NaN;
            switch (flag)
            {
                case "Backward":
                    result = (4 * DerivativeBackward3(f, x, h / 2) - DerivativeBackward3(f, x, h)) / 3;
                    break;
                case "Forward":
                    result = (4 * DerivativeForward3(f, x, h / 2) - DerivativeForward3(f, x, h)) / 3;
                    break;
                case "Central":
                    result = (4 * DerivativeCentral3(f, x, h / 2) - DerivativeCentral3(f, x, h)) / 3;
                    break;
                default:
                    result = double.NaN;
                    break;
            }
            return result;
        }

        public static double DerivativeRichardson3(double[] y, int i, double h, string flag)
        {
            double result = double.NaN;
            switch (flag)
            {
                case "Backward":
                    result = (4 * DerivativeBackward3(y, i, h / 2) - DerivativeBackward3(y, i, h)) / 3;
                    break;
                case "Forward":
                    result = (4 * DerivativeForward3(y, i, h / 2) - DerivativeForward3(y, i, h)) / 3;
                    break;
                case "Central":
                    result = (4 * DerivativeCentral3(y, i, h / 2) - DerivativeCentral3(y, i, h)) / 3;
                    break;
                default:
                    result = double.NaN;
                    break;
            }
            return result;
        }

        public static double DerivativeRichardson4(Function f, double x, double h, string flag)
        {
            double result = double.NaN;
            switch (flag)
            {
                case "Backward":
                    result = (4 * DerivativeBackward4(f, x, h / 2) - DerivativeBackward4(f, x, h)) / 3;
                    break;
                case "Forward":
                    result = (4 * DerivativeForward4(f, x, h / 2) - DerivativeForward4(f, x, h)) / 3;
                    break;
                case "Central":
                    result = (4 * DerivativeCentral4(f, x, h / 2) - DerivativeCentral4(f, x, h)) / 3;
                    break;
                default:
                    result = double.NaN;
                    break;
            }
            return result;
        }

        public static double DerivativeRichardson4(double[] y, int i, double h, string flag)
        {
            double result = double.NaN;
            switch (flag)
            {
                case "Backward":
                    result = (4 * DerivativeBackward4(y, i, h / 2) - DerivativeBackward4(y, i, h)) / 3;
                    break;
                case "Forward":
                    result = (4 * DerivativeForward4(y, i, h / 2) - DerivativeForward4(y, i, h)) / 3;
                    break;
                case "Central":
                    result = (4 * DerivativeCentral4(y, i, h / 2) - DerivativeCentral4(y, i, h)) / 3;
                    break;
                default:
                    result = double.NaN;
                    break;
            }
            return result;
        }

        #endregion

        #region Derivatives by Interpolation

        public static double DerivativeByInterpolation0(double[] x, double[] y, double xval)
        {
            double result = double.NaN;
            int n = y.Length;
            for (int i = 1; i < n - 1; i++)
            {
                if (xval > x[i - 1] && xval < x[i + 1])
                {
                    result = y[i - 1] * (xval - x[i]) * (xval - x[i + 1]) / ((x[i - 1] - x[i]) * (x[i - 1] - x[i + 1])) +
                             y[i] * (xval - x[i - 1]) * (xval - x[i + 1]) / ((x[i] - x[i - 1]) * (x[i] - x[i + 1])) +
                             y[i + 1] * (xval - x[i - 1]) * (xval - x[i]) / ((x[i + 1] - x[i - 1]) * (x[i + 1] - x[i]));
                }
            }
            return result;
        }

        public static double DerivativeByInterpolation1(double[] x, double[] y, double xval)
        {
            double result = double.NaN;
            int n = y.Length;
            for (int i = 1; i < n - 1; i++)
            {
                if (xval > x[i - 1] && xval < x[i + 1])
                {
                    result = y[i - 1] * (2 * xval - x[i] - x[i + 1]) / ((x[i - 1] - x[i]) * (x[i - 1] - x[i + 1])) +
                             y[i] * (2 * xval - x[i - 1] - x[i + 1]) / ((x[i] - x[i - 1]) * (x[i] - x[i + 1])) +
                             y[i + 1] * (2 * xval - x[i - 1] - x[i]) / ((x[i + 1] - x[i - 1]) * (x[i + 1] - x[i]));
                }
            }
            return result;
        }

        public static double DerivativeByInterpolation2(double[] x, double[] y, double xval)
        {
            double result = double.NaN;
            int n = y.Length;
            for (int i = 1; i < n - 1; i++)
            {
                if (xval > x[i - 1] && xval < x[i + 1])
                {
                    result = y[i - 1] * 2 / ((x[i - 1] - x[i]) * (x[i - 1] - x[i + 1])) +
                             y[i] * 2 / ((x[i] - x[i - 1]) * (x[i] - x[i + 1])) +
                             y[i + 1] * 2 / ((x[i + 1] - x[i - 1]) * (x[i + 1] - x[i]));
                }
            }
            return result;
        }

        #endregion

        #region Test Functions

        //Test function: f(x) = cos(x)
        static double f1(double x)
        {
            return Math.Cos(x); 
        }

        //Test function: f(x) = x^3 + e^x
        static double f2(double x)
        {
            return x * x * x + Math.Exp(x);
        }

        static void TestForwardDifferenceMethod()
        {
            double h = 0.1;

            // Calculate derivatives using f(x) = cos(x) at x=0.3:
            Console.WriteLine("Using the FORWARD difference method\n");

            double dy = DerivativeForward1(f1, 0.3, h);
            Console.WriteLine(" f'(x) = " + dy.ToString());
            dy = DerivativeForward2(f1, 0.3, h);
            Console.WriteLine(" f''(x) = " + dy.ToString());
            dy = DerivativeForward3(f1, 0.3, h);
            Console.WriteLine(" f'''(x) = " + dy.ToString());
            dy = DerivativeForward4(f1, 0.3, h);
            Console.WriteLine(" f''''(x) = " + dy.ToString());

            // Calculate derivatives using just array values at x=0.3 or i=3
            Console.WriteLine("\nDerivatives using just array values:\n");
            double[] y=new double[10];
            for (int i = 0; i < 10; i++)
                y[i] = f1(i * h);
            dy = DerivativeForward1(y, 3, h);
            Console.WriteLine(" y' = " + dy.ToString());
            dy = DerivativeForward2(y, 3, h);
            Console.WriteLine(" y'' = " + dy.ToString());
            dy = DerivativeForward3(y, 3, h);
            Console.WriteLine(" y''' = " + dy.ToString());
            dy = DerivativeForward4(y, 3, h);
            Console.WriteLine(" y'''' = " + dy.ToString());
            Console.WriteLine("------------------------------------");
        }

        static void TestBackwardDifferenceMethod()
        {
            double h = 0.1;

            // Calculate derivatives using f(x) = cos(x) at x=0.7:
            Console.WriteLine("\nUsing the BACKWARD difference method\n");

            double dy = DerivativeBackward1(f1, 0.7, h);
            Console.WriteLine(" f'(x) = " + dy.ToString());
            dy = DerivativeBackward2(f1, 0.7, h);
            Console.WriteLine(" f''(x) = " + dy.ToString());
            dy = DerivativeBackward3(f1, 0.7, h);
            Console.WriteLine(" f'''(x) = " + dy.ToString());
            dy = DerivativeBackward4(f1, 0.7, h);
            Console.WriteLine(" f''''(x) = " + dy.ToString());

            // Calculate derivatives for array values at x=0.7 or i=7:
            Console.WriteLine("\nDerivatives for array values:\n");
            double[] y = new double[10];
            for (int i = 0; i < 10; i++)
                y[i] = f1(i * h);
            dy = DerivativeBackward1(y, 7, h);
            Console.WriteLine(" y' = " + dy.ToString());
            dy = DerivativeBackward2(y, 7, h);
            Console.WriteLine(" y'' = " + dy.ToString());
            dy = DerivativeBackward3(y, 7, h);
            Console.WriteLine(" y''' = " + dy.ToString());
            dy = DerivativeBackward4(y, 7, h);
            Console.WriteLine(" y'''' = " + dy.ToString());
            Console.WriteLine("---------------------------------------");
        }

        static void TestCentralDifferenceMethod()
        {
            double h = 0.1;

            // Calculate derivatives using f(x) = cos(x) at x=0.5:
            Console.WriteLine("\nUsing the CENTRAL difference method\n");

            double dy = DerivativeCentral1(f1, 0.5, h);
            Console.WriteLine(" f'(x) = " + dy.ToString());
            dy = DerivativeCentral2(f1, 0.5, h);
            Console.WriteLine(" f''(x) = " + dy.ToString());
            dy = DerivativeCentral3(f1, 0.5, h);
            Console.WriteLine(" f'''(x) = " + dy.ToString());
            dy = DerivativeCentral4(f1, 0.5, h);
            Console.WriteLine(" f''''(x) = " + dy.ToString());

            // Calculate derivatives for array values at x=0.5 or i=5
            Console.WriteLine("\nDerivatives for array values:\n");
            double[] y = new double[10];
            for (int i = 0; i < 10; i++)
                y[i] = f1(i * h);
            dy = DerivativeCentral1(y, 5, h);
            Console.WriteLine(" y' = " + dy.ToString());
            dy = DerivativeCentral2(y, 5, h);
            Console.WriteLine(" y'' = " + dy.ToString());
            dy = DerivativeCentral3(y, 5, h);
            Console.WriteLine(" y''' = " + dy.ToString());
            dy = DerivativeCentral4(y, 5, h);
            Console.WriteLine(" y'''' = " + dy.ToString());
            Console.WriteLine("---------------------------------------");
        }

        static void TestOh4CentralDifferenceMethod()
        {
            double h = 0.1;

            // Calculate derivatives using f(x) = cos(x) at x=0.5:
            Console.WriteLine("\nUsing the O(h^4) CENTRAL difference method\n");

            double dy = DerivativeOh4Central1(f1, 0.5, h);
            Console.WriteLine(" f'(x) = " + dy.ToString());
            dy = DerivativeOh4Central2(f1, 0.5, h);
            Console.WriteLine(" f''(x) = " + dy.ToString());
            dy = DerivativeOh4Central3(f1, 0.5, h);
            Console.WriteLine(" f'''(x) = " + dy.ToString());
            dy = DerivativeOh4Central4(f1, 0.5, h);
            Console.WriteLine(" f''''(x) = " + dy.ToString());

            // Calculate derivatives for array values at x=0.5 or i = 5:
            Console.WriteLine("\nDerivatives for array values:\n");
            double[] y = new double[10];
            for (int i = 0; i < 10; i++)
                y[i] = f1(i * h);
            dy = DerivativeOh4Central1(y, 5, h);
            Console.WriteLine(" y' = " + dy.ToString());
            dy = DerivativeOh4Central2(y, 5, h);
            Console.WriteLine(" y'' = " + dy.ToString());
            dy = DerivativeOh4Central3(y, 5, h);
            Console.WriteLine(" y''' = " + dy.ToString());
            dy = DerivativeOh4Central4(y, 5, h);
            Console.WriteLine(" y'''' = " + dy.ToString());
            Console.WriteLine("------------------------------------");
            // Analytic results:
            Console.WriteLine("\nAnalytic results:\n");
            dy = -Math.Sin(0.5);
            Console.WriteLine(" y' = " + dy.ToString());
            dy = -Math.Cos(0.5);
            Console.WriteLine(" y'' = " + dy.ToString());
            dy = Math.Sin(0.5);
            Console.WriteLine(" y''' = " + dy.ToString());
            dy = Math.Cos(0.5);
            Console.WriteLine(" y'''' = " + dy.ToString());
            Console.WriteLine("---------------------------------------");
        }

        static void TestRichardson()
        {
            double h = 0.1;
            double x = 0.5;
            Console.WriteLine("Testing Richardson's Method\n");
            Console.WriteLine("Function used: f(x) = x^3 + e^x\n");
            Console.WriteLine("KEY:\n");
            Console.WriteLine("f[n]exact = f'exact, f''exact, ... f''''exact\n");
            Console.WriteLine("central[n] = 1st derivative using central method, etc\n");
            Console.WriteLine("richardson[n] = 1st derivative using Richardson method, etc\n");

            Console.WriteLine("Results:\n");
            double f1exact = 3 * x * x + Math.Exp(x); //f'(x)
            double central1 = DerivativeCentral1(f2, x, h);
            double richardson1 = DerivativeRichardson1(f2, x, h, "Central");
            Console.WriteLine(" f1exact = {0,11:n8}, central1 = {1,11:n8}, richardson1 = {2,11:n8}", f1exact, central1, richardson1);

            double f2exact = 6 * x + Math.Exp(x); //f''(x)
            double central2 = DerivativeCentral2(f2, x, h);
            double richardson2 = DerivativeRichardson2(f2, x, h, "Central");
            Console.WriteLine(" f2exact = {0,11:n8}, central2 = {1,11:n8}, richardson2 = {2,11:n8}", f2exact, central2, richardson2);

            double f3exact = 6 + Math.Exp(x); //f'''(x)
            double central3 = DerivativeCentral3(f2, x, h);
            double richardson3 = DerivativeRichardson3(f2, x, h, "Central");
            Console.WriteLine(" f3exact = {0,11:n8}, central3 = {1,11:n8}, richardson3 = {2,11:n8}", f3exact, central3, richardson3);

            double f4exact = Math.Exp(x); //f''''(x)
            double central4 = DerivativeCentral4(f2, x, h);
            double richardson4 = DerivativeRichardson4(f2, x, h, "Central");
            Console.WriteLine(" f4exact = {0,11:n8}, central4 = {1,11:n8}, richardson4 = {2,11:n8}", f4exact, central4, richardson4); 
        }

        static void TestInterpolation()
        {
            Console.WriteLine("\n\nTesting Derivatives by Polynomial Interpolation Method\n");
            double x = 3.8;
            double[] xa = new double[] { 2.0, 2.1, 2.3, 2.6, 2.7, 4.0 };
            double[] ya = new double[] { 3.2332, 3.1634, 2.7220, 2.1635, -2.0829, -2.6365 };
            double y0 = DerivativeByInterpolation0(xa, ya, x);
            double y1 = DerivativeByInterpolation1(xa, ya, x);
            double y2 = DerivativeByInterpolation2(xa, ya, x);
            Console.WriteLine("y = {0:n6}\ny' = {1:n6}\ny'' = {2:n6}\n", y0, y1, y2);
        }

        #endregion
      
        static void Main(string[] args)
        {
            Console.WriteLine("Testing various numerical derivative methods...\n");
            Console.WriteLine("Using Finite Differences Method and function: f(x) = cos(x)\n");
            TestForwardDifferenceMethod();
            TestBackwardDifferenceMethod();
            TestCentralDifferenceMethod();
            TestOh4CentralDifferenceMethod();
            TestRichardson();
            TestInterpolation();
            Console.ReadLine();
        }
    }
}

