﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace Chapter12_CodeExamples
{
    class Program
    {

        #region Polynomials

        public static double GammaLn(double x)
        {
            if (x <= 0) throw
            new Exception("Input value must be > 0");

            double[] coef = new double[14]
            {57.1562356658629235,
            -59.5979603554754912,
            14.1360979747417471,
            -0.491913816097620199,
             0.339946499848118887E-4,
             0.465236289270485756E-4,
            -0.983744753048795646E-4,
             0.158088703224912494E-3,
            -0.210264441724104883E-3,
             0.217439618115212643E-3,
            -0.164318106536763890E-3,
             0.844182239838527433E-4,
            -0.261908384015814087E-4,
             0.368991826595316234E-5};

            double denominator = x;
            double series = 0.999999999999997092;
            double temp = x + 5.24218750000000000;
            temp = (x + 0.5) * Math.Log(temp) - temp;
            for (int j = 0; j < 14; j++)
                series += coef[j] / ++denominator;
            return (temp + Math.Log(2.5066282746310005 * series / x));
        }

        public static double Gamma(double x)
        {
            if (x <= 0) throw
                new Exception("Input value must be > 0");
            return Math.Exp(GammaLn(x));
        }

        public static double Laguerre(double x, int deg)
        {
            double L0 = 1.0;
            double L1 = 1.0 - x;
            double L2 = (x * x - 4 * x + 2.0) / 2.0;
            int n = 1;
            if (deg < 0)
                throw new Exception("Bad Laguerre polynomial: deg < 0");
            if (deg == 0)
                return L0;
            else if (deg == 1)
                return L1;
            else
            {
                while (n < deg)
                {
                    L2 = ((2.0 * n + 1.0 - x) * L1 - n * L0) / (n + 1);
                    L0 = L1;
                    L1 = L2;
                    n++;
                }
                return L2;
            }
        }

        public static double Hermite(double x, int deg)
        {
            double H0 = 1.0;
            double H1 = 2 * x;
            double H2 = 4 * x * x - 2;
            int n = 1;
            if (deg < 0)
                throw new Exception("Bad Hermite polynomial: deg < 0");
            if (deg == 0)
                return H0;
            else if (deg == 1)
                return H1;
            else
            {
                while (n < deg)
                {
                    H2 = 2.0 * x * H1 - 2.0 * n * H0;
                    H0 = H1;
                    H1 = H2;
                    n++;
                }
                return H2;
            }
        }

        public static double ChebyshevT(double x, int deg)
        {
            double T0 = 1.0;
            double T1 = x;
            double T2 = 2.0 * x * x - 1.0;
            int n = 1;
            if (deg < 0)
                throw new Exception("Bad Hermite polynomial: deg < 0");
            if (deg == 0)
                return T0;
            else if (deg == 1)
                return T1;
            else
            {
                while (n < deg)
                {
                    T2 = 2.0 * x * T1 - T0;
                    T0 = T1;
                    T1 = T2;
                    n++;
                }
                return T2;
            }
        }

        public static double ChebyshevU(double x, int deg)
        {
            double U0 = 1.0;
            double U1 = 2.0 * x;
            double U2 = 4 * x * x - 1.0;
            int n = 1;
            if (deg < 0)
                throw new Exception("Bad Hermite polynomial: deg < 0");
            if (deg == 0)
                return U0;
            else if (deg == 1)
                return U1;
            else
            {
                while (n < deg)
                {
                    U2 = 2.0 * x * U1 - U0;
                    U0 = U1;
                    U1 = U2;
                    n++;
                }
                return U2;
            }
        }

        public static double Legendre(double x, int deg)
        {
            double P0 = 1.0;
            double P1 = x;
            double P2 = (3.0 * x * x - 1) / 2.0;
            int n = 1;
            if (deg < 0)
                throw new Exception("Bad Hermite polynomial: deg < 0");
            if (deg == 0)
                return P0;
            else if (deg == 1)
                return P1;
            else
            {
                while (n < deg)
                {
                    P2 = 2.0 * x * P1 - P0 - (x * P1 - P0) / (deg + 1);
                    P0 = P1;
                    P1 = P2;
                    n++;
                }
                return P2;
            }
        }

        #endregion        

        #region Rectangular Method
        public static double Rectangular(Function f, double a, double b, int n)
        {
            double sum = 0.0;
            double h = (b - a) / n ;
            for (int i = 0; i < n ; i++)
            {
                sum += h * f(a + i * h);
            }
            return sum;
        }

        public static double Rectangular(double[] y, double a, double b, int n)
        {
            double sum = 0.0;
            double h = (b - a) / n;
            for (int i = 0; i < n; i++)
            {
                sum +=  h * y[i];
            }
            return sum;
        }
        #endregion

        #region Midpoint Method
        public static double MidpointMethod(Function f, double a, double b, int n)
        {
            double sum = 0.0;
            double h = (b - a) / n;
            for (int i = 0; i < n; i++)
            {
                sum += h * f(a + (i + 0.5) * h);
            }
            return sum;
        }

        public static double MidpointMethod(double[] y, double a, double b, int n)
        {
            double sum = 0.0;
            double h = (b - a) / (n - 1);
            for (int i = 0; i < (n - 1); i++)
            {
                sum += h * 0.5 * (y[i] + y[i + 1]);
            }
            return sum;
        }
        #endregion

        #region Trapezoidal Method

        public static double Trapezoidal(Function f, double a, double b, int n)
        {
            double sum = 0.0;
            double h = (b - a) / n;
            for (int i = 0; i < n; i++)
            {
                sum += 0.5 * h * (f(a + i * h) + f(a + (i + 1) * h));
            }
            return sum;
        }

        public static double Trapezoidal(double[] y, double a, double b, int n)
        {
            double sum = 0.0;
            double h = (b - a) / (n-1);
            for (int i = 0; i < (n-1); i++)
            {
                sum += 0.5 * h * (y[i] + y[i + 1]);
            }
            return sum;
        }

        #endregion

        #region Simpson's Method

        public static double Simpson(Function f, double a, double b, int n)
        {
            if (n < 3) return double.NaN; //Need at least 3 points
            double sum = 0.0;
            double h = (b - a) / n;
            if (n % 2 != 0)
            {
                for (int i = 0; i < n - 1; i += 2)
                {
                    sum += h * (f(a + i * h) + 4 * f(a + (i + 1) * h) + f(a + (i + 2) * h)) / 3;
                }
            }
            else
            {
                sum = 3 * h * (f(a) + 3 * f(a + h) + 3 * f(a + 2 * h) + f(a + 3 * h)) / 8;
                for (int i = 3; i < n - 1; i += 2)
                {
                    sum += h * (f(a + i * h) + 4 * f(a + (i + 1) * h) + f(a + (i + 2) * h)) / 3;
                }
            }
            return sum;
        }

        public static double Simpson(double[] y, double a, double b, int n)
        {
            double h = (b - a) / n;
            //Need at least 3 points
            if (n < 3 || h == 0) return double.NaN;

            double sum = 0.0;            
            if (n % 2 != 0)
            {
                for (int i = 0; i < n - 1; i += 2)
                {
                    sum += h * (y[i] + 4 * y[i + 1] + y[i + 2]) / 3;
                }
            }
            else
            {
                sum = 3 * h * (y[0] + 3 * y[1] + 3 * y[2] + y[3]) / 8;
                for (int i = 3; i < n - 1; i += 2)
                {
                    sum += h * (y[i] + 4 * y[i + 1] + y[i + 2]) / 3;
                }
            }
            return sum;
        }

        #endregion

        #region Romberg's Method

        public static double RombergIntegration(Function f, double a, double b, int maxIterations, double tolerance)
        {
            int n = 2;
            double h = b - a;
            double sum = 0.0;
            int j = 0;
            double[,] R = new double[maxIterations, maxIterations];

            R[1, 1] = h * (f(a) + f(b)) / 2.0;
            h = h / 2;
            R[2, 1] = R[1, 1] / 2 + h * f(a + h);
            R[2, 2] = (4 * R[2, 1] - R[1, 1]) / 3;

            for (j = 3; j <= maxIterations; j++)
            {
                n = 2 * n;
                h = h / 2;
                sum = 0.0;
                for (int k = 1; k <= n; k += 2)
                {
                    sum += f(a + k * h);
                }
                R[j, 1] = R[j - 1, 1] / 2 + h * sum;
                double factor = 4.0;
                for (int k = 2; k <= j; k++)
                {
                    R[j, k] = (factor * R[j, k - 1] - R[j - 1, k - 1]) / (factor - 1);
                    factor = factor * 4.0;
                }
                if (Math.Abs(R[j, j] - R[j, j - 1]) < tolerance * Math.Abs(R[j, j]))
                {
                    sum = R[j, j];
                    return sum;
                }
            }
            sum = R[n, n];
            return sum;
        }

        #endregion

        #region Gauss-Legendre Quadrature Method

        public static double GaussLegendre(Function f, double a, double b, int n)
        {
            double[] x, w;
            LegendreNodesWeights(n, out x, out w);

            double sum = 0.0;
            for (int i = 0; i < n; i++)
            {
                sum += 0.5 * (b - a) * w[i] * f(0.5 * (a + b) + 0.5 * (b - a) * x[i]);
            }
            return sum;
        }

        public static void LegendreNodesWeights(int n, out double[] x, out double[] w)
        {
            double c, d, p1, p2, p3, dp;

            x = new double[n];
            w = new double[n];

            for (int i = 0; i < (n + 1) / 2; i++)
            {
                c = Math.Cos(Math.PI * (4 * i + 3) / (4 * n + 2));
                do
                {
                    p2 = 0;
                    p3 = 1;
                    for (int j = 0; j < n; j++)
                    {
                        p1 = p2;
                        p2 = p3;
                        p3 = ((2 * j + 1) * c * p2 - j * p1) / (j + 1);
                    }
                    dp = n * (c * p3 - p2) / (c * c - 1);
                    d = c;
                    c -= p3 / dp;
                }
                while (Math.Abs(c - d) > 1e-12);
                x[i] = c;
                x[n - 1 - i] = -c;
                w[i] = 2 * (1 - x[i] * x[i]) / (n + 1) / (n + 1) / Legendre(x[i], n + 1) / Legendre(x[i], n + 1);
                w[n - 1 - i] = w[i];
            }
        }

        #endregion

        #region Gauss-Laguerre Quadrature Method

        public static double GaussLaguerre(Function f, int n)
        {
            double[] x, w;
            LaguerreNodesWeights(n, out x, out w);

            double sum = 0.0;
            for (int i = 0; i < n; i++)
            {
                sum += w[i] * f(x[i]);
            }
            return sum;
        }

        public static void LaguerreNodesWeights(int n, out double[] x, out double[] w)
        {
            double c = 0.0;
            double d, p1, p2, p3, dp;

            x = new double[n];
            w = new double[n];
            for (int i = 0; i < n; i++)
            {
                if (i == 0)
                {
                    c = 3 / (1 + 2.4 * n);
                }
                else
                {
                    if (i == 1)
                    {
                        c += 15 / (1 + 2.5 * n);
                    }
                    else
                    {
                        c += (1 + 2.55 * (i - 1)) / (1.9 * (i - 1)) * (c - x[i - 2]);
                    }
                }
                do
                {
                    p2 = 0;
                    p3 = 1;
                    for (int j = 0; j < n; j++)
                    {
                        p1 = p2;
                        p2 = p3;
                        p3 = ((-c + 2 * j + 1) * p2 - j * p1) / (j + 1);
                    }
                    dp = (n * p3 - n * p2) / c;
                    d = c;
                    c = c - p3 / dp;
                }
                while (Math.Abs(c - d) > 1e-12);
                x[i] = c;
                w[i] = x[i] / (n + 1) / (n + 1) / Laguerre(x[i], n + 1) / Laguerre(x[i], n + 1);
            }
        }

        #endregion

        #region Gauss-Hermite Quadrature Method

        public static double GaussHermite(Function f, int n)
        {
            double[] x, w;
            HermiteNodesWeights(n, out x, out w);

            double sum = 0.0;
            for (int i = 0; i < n; i++)
            {
                sum += w[i] * f(x[i]);
            }
            return sum;
        }

        public static void HermiteNodesWeights(int n, out double[] x, out double[] w)
        {
            double c = 0.0;
            double d, p1, p2, p3, dp;

            x = new double[n];
            w = new double[n];
            for (int i = 0; i < (n + 1) / 2; i++)
            {
                if (i == 0)
                {
                    c = Math.Sqrt(2 * n + 1) - 1.85575 * Math.Pow(2 * n + 1, -((double)(1) / (double)(6)));
                }
                else
                {
                    if (i == 1)
                    {
                        c = c - 1.14 * Math.Pow(n, 0.426) / c;
                    }
                    else
                    {
                        if (i == 2)
                        {
                            c = 1.86 * c - 0.86 * x[0];
                        }
                        else
                        {
                            if (i == 3)
                            {
                                c = 1.91 * c - 0.91 * x[1];
                            }
                            else
                            {
                                c = 2 * c - x[i - 2];
                            }
                        }
                    }
                }
                do
                {
                    p2 = 0;
                    p3 = Math.Pow(Math.PI, -0.25);
                    for (int j = 0; j < n; j++)
                    {
                        p1 = p2;
                        p2 = p3;
                        p3 = p2 * c * Math.Sqrt((double)(2) / ((double)(j + 1))) - p1 * Math.Sqrt((double)(j) / ((double)(j + 1)));
                    }
                    dp = Math.Sqrt(2 * n) * p2;
                    d = c;
                    c -= p3 / dp;
                }
                while (Math.Abs(c - d) > 1e-12);
                x[i] = c;
                w[i] = Math.Pow(2, n + 1) * Gamma(n + 1) * Math.Sqrt(Math.PI) /
                       Hermite(x[i], n + 1) / Hermite(x[i], n + 1);
                x[n - 1 - i] = -x[i];
                w[n - 1 - i] = w[i];
            }
        }

        #endregion

        #region Gauss-Chebyshev Quadrature Method
        
        public static double GaussChebyshev(Function f, int n)
        {
            double x;
            double w = Math.PI / n;

            double sum = 0.0;
            for (int i = 0; i < n; i++)
            {
                x = Math.Cos(Math.PI * (i + 0.5) / n);
                sum += f(x);
            }
            return sum * w;
        }

        #endregion

        public delegate double Function(double x);

        #region Material for running the integration tests

        static void TestRectangular()
        {
            int n = 1000;
            double result;
            double a = 0.0;
            double b = 1.0;

            Console.WriteLine("\nTesting Rectangular method for integral(df(x)) = f(x) where");
            Console.WriteLine("f(x) = sin(x), df(x) = cos(x) and a=0 to b=1.\n");

            result = f(b) - f(a);
            Console.WriteLine("Analytic result         = " + result.ToString());

            result = Rectangular(df, a, b, n);
            Console.WriteLine("Result using function   = " + result.ToString());

            double[] y = new double[n];
            double h = (b - a) / (n - 1);
            for (int i = 0; i < n; i++)
            {
                double x = i * h;
                y[i] = df(x);
            }
            result = Rectangular(y, a, b, n);
            Console.WriteLine("Result using data array = " + result.ToString());
        }

        static void TestMidpointMethod()
        {
            int n = 1000;
            double result;
            double a = 0.0;
            double b = 1.0;

            Console.WriteLine("\nTesting Midpoint method for integral(df(x)) = f(x) where");
            Console.WriteLine("f(x) = sin(x), df(x) = cos(x) and a=0 to b=1.\n");

            result = f(b) - f(a);
            Console.WriteLine("Analytic result         = " + result.ToString());

            result = MidpointMethod(df, a, b, n);
            Console.WriteLine("Result using function   = " + result.ToString());

            double[] y = new double[n];
            double h = (b - a) / (n - 1);
            for (int i = 0; i < n; i++)
            {
                double x = i * h;
                y[i] = df(x);
            }
            result = MidpointMethod(y, a, b, n);
            Console.WriteLine("Result using data array = " + result.ToString());
        }

        static void TestTrapezoidal()
        {
            int n = 1000;
            double result;
            double a = 0.0;
            double b = 1.0;

            Console.WriteLine("\nTesting Trapezoidal method for integral(df(x)) = f(x) where");
            Console.WriteLine("f(x) = sin(x), df(x) = cos(x) and a=0 to b=1.\n");

            result = f(b) - f(a);
            Console.WriteLine("Analytic result         = " + result.ToString());

            result = Trapezoidal(df, a, b, n);
            Console.WriteLine("Result using function   = " + result.ToString());

            double[] y = new double[n];
            double h = (b-a) / (n-1);
            for (int i = 0; i < n; i++)
            {
                double x = i * h;
                y[i] = df(x);
            }
            result = Trapezoidal(y, a, b, n);
            Console.WriteLine("Result using data array = " + result.ToString());
        }

        static void TestSimpson()
        {
            int n = 1000;
            double result;
            double a = 0.0;
            double b = 1.0;

            Console.WriteLine("\nTesting Simpson method for integral(df(x)) = f(x) where");
            Console.WriteLine("f(x) = sin(x), df(x) = cos(x) and a=0 to b=1.\n");

            result = f(b) - f(a);
            Console.WriteLine("Analytic result         = " + result.ToString());

            result = Simpson(df, a, b, n);
            Console.WriteLine("Result using function   = " + result.ToString());


            double[] ya = new double[n];
            double h = (b - a) / n;
            for (int i = 0; i < n; i++)
            {
                double x = i * h;
                ya[i] = df(x);
            }
            result = Simpson(ya, a, b, n);

            Console.WriteLine("Result using data array = " + result.ToString());
        }

        static void TestRomberg()
        {
            Console.WriteLine("\nTesting Romberg's Method for integral(df(x)) = f(x) where");
            Console.WriteLine("f(x) = sin(x), df(x) = cos(x) and a=0 to b=1.\n");

            double result = f(1) - f(0);
            Console.WriteLine("Analytic result              = " + result.ToString());

            result = RombergIntegration(df, 0, 1, 10, 1e-10);
            Console.WriteLine("Result from Romberg's method = {0}", result);
        }

        static void TestGaussLegendre()
        {
            Console.WriteLine("\nResult from Gauss-Legendre method:\n");
            double result;
            for (int n = 1; n < 9; n++)
            {
                result = GaussLegendre(f2, 1, 2, n);
                Console.WriteLine(" n = {0}, result = {1}", n, result);
            }
        }

        static void TestGaussLaguerre()
        {
            Console.WriteLine("\nResult from Gauss-Laguerre method:\n");
            double result;
            for (int n = 1; n < 9; n++)
            {
                result = GaussLaguerre(f3, n);
                Console.WriteLine(" n = {0}, result = {1}", n, result);
            }
        }

        static void TestGaussHermite()
        {
            Console.WriteLine("\nResult from Gauss-Hermit method:\n");
            double result;
            for (int n = 1; n < 9; n++)
            {
                result = GaussHermite(f4, n);
                Console.WriteLine(" n = {0}, result = {1}", n, result);
            }
        }

        static void TestChebyshev()
        {
            Console.WriteLine("\nResult from Gauss-Chebyshev method:\n");
            double result;
            for (int n = 1; n < 9; n++)
            {
                result = GaussChebyshev(f5, n);
                Console.WriteLine(" n = {0}, result = {1}", n, result);
            }
        }

        static double df(double x)
        {
            return Math.Cos(x);
        }

        static double f(double x)
        {
            return Math.Sin(x);
        }

        static double f2(double x)
        {
            return Math.Exp(-x * x);
        }

        static double f3(double x)
        {
            return Math.Sin(x);
        }

        static double f4(double x)
        {
            return (1 - x * x) * Math.Exp(x);
        }

        static double f5(double x)
        {
            return (1 - x * x) * Math.Exp(x);
        }

        #endregion

        #region Monte Carlo Integration

        public static void MonteCarloIntegrationDemo()
        {
            Console.WriteLine("\n\nMonte Carlo Integration Example");
            Console.WriteLine("Integrating function f(x) = x * x from 0 to 1\n");            
            Random rand = new Random();
            int nPts = 1000000;
            double s0 = 0;
            double sigma = 0;
            double x = 0.0;
            double fcn;
            for (int i = 0; i < nPts; ++i)
            {
                x = rand.NextDouble();
                fcn = x * x;
                s0 += fcn;
                sigma += fcn * fcn;
            }
            s0 /= nPts;
            sigma /= nPts;
            sigma = Math.Sqrt(Math.Abs(sigma - s0 * s0) / nPts);
            Console.WriteLine("Analytical result           = 0.3333333....333");
            Console.WriteLine("Monte Carlo Integral result = " + s0 + " +- " + sigma);
        }

        #endregion

        #region Metropolis Method

        public static void MetropolisAlgorithmDemo()
        {
            Console.WriteLine("\n\nMetropolis Integration Algorithm Example");
            Console.WriteLine("Integrating function f(x) = x * x from 0 to 1\n");  
            int nsize = 100000;
            int nskip = 15;
            int ntotal = nsize * nskip;
            int neq = 10000;
            int iaccept = 0;
            double x, w, h = 0.4, z = 0.46265167;
            Random r = new Random();

            x = r.NextDouble();
            w = weight(x);
            for (int i = 0; i < neq; ++i)
            {
                double rand = r.NextDouble();
                Metropolis(ref x, ref  h, ref r, ref iaccept, ref w);
            }

            double s0 = 0;
            double ds = 0;
            iaccept = 0;
            for (int i = 0; i < ntotal; ++i)
            {
                double rand = r.NextDouble();
                Metropolis(ref  x, ref h, ref r, ref iaccept, ref w);

                if (i % nskip == 0)
                {
                    double f = g(x);
                    s0 += f;
                    ds += f * f;
                }
            }
            s0 /= nsize;
            ds /= nsize;
            ds = Math.Sqrt(Math.Abs(ds - s0 * s0) / nsize);
            s0 *= z;
            ds *= z;
            double accept = 100.0 * iaccept / ntotal;
            Console.WriteLine("Analytical result  = 0.3333333....333");
            Console.WriteLine("Integral result    = " + s0 + " +- " + ds);
            Console.WriteLine("Acceptance rate    = " + accept + "%");
        }

        public static void Metropolis(ref double x, ref double h, ref Random rand, ref int iaccept, ref double w)
        {
            double xold = x;
            x = x + 2 * h * (rand.NextDouble() - 0.5);
            if ((x < 0) || (x > 1)) x = xold;
            else
            {
                double wnew = weight(x);
                if (wnew > w * rand.NextDouble())
                {
                    w = wnew;
                    ++iaccept;
                }
                else x = xold;
            }
        }

        public static double weight(double x)
        {
            return Math.Exp(x * x) - 1;
        }

        public static double g(double x)
        {
            return x * x / (Math.Exp(x * x) - 1);
        }

        #endregion

        #region Convolution Methods

        public static double[] Convolution1(double[] x, double[] h)
        {
            //Uses the output side algorithm
            double[] y = new double[x.Length + h.Length - 1];

            for (int i = 0; i < y.Length; i++)
            {
                y[i] = 0.0;
                for (int j = 0; j < h.Length; j++)
                {
                    if ((i - j) < 0) continue;
                    if ((i - j) > (x.Length - 1)) continue;
                    y[i] = y[i] + h[j] * x[i - j];
                }
            }
            return y;
        }

        public static double[] Convolution2(double[] x, double[] h)
        {
            //Uses the input side algorithm
            double[] y = new double[x.Length + h.Length - 1];

            for (int i = 0; i < y.Length; i++)
            {
                y[i] = 0.0;
            }

            for (int i = 0; i < x.Length; i++)
            {
                for (int j = 0; j < h.Length; j++)
                {
                    y[i + j] = y[i + j] + x[i] * h[j];
                }
            }
            return y;
        }

        static void ConvolutionTestDiscrete()
        {
            Console.WriteLine("Testing Convolution Methods\n");

            double[] x = { 2, 2, 3, 3, 4 };
            double[] h = { 1, 1, 2 };

            double[] y = new double[x.Length + h.Length - 1];

            Console.WriteLine("Running Convolution1()");
            y = Convolution1(x, h);
            for (int i = 0; i < y.Length; i++)
            {
                Console.WriteLine("y[{0}] = {1}", i, y[i]);
            }
            Console.WriteLine("\n\nRunning Convolution2()");
            y = Convolution2(x, h);
            for (int i = 0; i < y.Length; i++)
            {
                Console.WriteLine("y[{0}] = {1}", i, y[i]);
            }
            Console.WriteLine("\nExpected Results = 2 4 9 10 13 10 8");
        }

        #endregion

        static void Main(string[] args)
        {
            Console.WriteLine("Testing Numerical Integration Methods\n");
            TestRectangular();
            TestMidpointMethod();
            TestTrapezoidal();
            TestSimpson();
            TestRomberg();
            TestGaussLegendre();
            TestGaussLaguerre();
            TestGaussHermite();
            TestChebyshev();
            MonteCarloIntegrationDemo();
            MetropolisAlgorithmDemo();
            Console.WriteLine("Testing Convolution Methods\n");
            ConvolutionTestDiscrete();
            Console.ReadLine();
        }
    }
}
