﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace Chapter13_CodeExamples
{
    class Program
    {        
        #region Instantiation of random number generator object

        //Use the Visual C\# built-in random number 
        //generator or one of your own choosing here.
        private static Random randObj = new Random();

        #endregion

        #region Statistical Methods

        public static bool NaNTest(double[] x)
        {
            for (int i = 0; i < x.Length; i++)
            {
                if (Double.IsNaN(x[i]))
                {
                    return true;
                }
            }
            return false;
        }

        //Removes all NaN values that may be found in a data set
        public static double[] NaNDelete(double[] x)
        {
            int n = 0;
            for (int i = 0; i < x.Length; i++)
            {
                if (Double.IsNaN(x[i]))
                {
                    n++;
                }
            }
            double[] removed = new double[x.Length - n];
            int index = 0;
            for (int i = 0; i < x.Length; i++)
            {
                if (!Double.IsNaN(x[i]))
                {
                    removed[index] = x[i];
                    index++;
                }
            }
            return removed;
        }

        enum SortOrder { Ascending, Descending };
        static double[] Sort(double[] x, SortOrder order)
        {
            return SortNoClone((double[])x.Clone(), order);
        }

        static double[] Sort(double[] x)
        {
            return SortNoClone((double[])x.Clone(), SortOrder.Ascending);
        }

        static double[] SortNoClone(double[] x, SortOrder order)
        {
            Array.Sort(x);
            if (order == SortOrder.Descending)
            {
                Array.Reverse(x);
            }
            return x;
        }

        public static double Mean(double[] x)
        {
            double total = 0.0d;
            for (int i = 0; i < x.Length; i++)
            {
                total += x[i];
            }
            return (total / x.Length);
        }

        public static double WeightedMean(double[] x, double[] weights)
        {
            if (x.Length != weights.Length)
            {
                throw new Exception("Data and Weight arrays are not the same size");
            }
            double numerator = 0.0d;
            double denominator = 0.0d;
            for (int i = 0; i < x.Length; i++)
            {
                numerator += x[i] * weights[i];
                denominator += weights[i];
            }
            return numerator / denominator;
        }

        public static double GeometricMean(double[] x)
        {
            if (x.Length == 0)
            {
                throw new Exception("Geometric Mean requires at least one value in the data.");
            }
            double product = 1.0;
            for (int i = 0; i < x.Length; i++)
            {
                product *= x[i];
            }
            return Math.Pow(product, (1.0 / x.Length));
        }

        public static double WeighedGeometricMean(double[] x, double[] weights)
        {
            if (x.Length != weights.Length)
            {
                throw new Exception("Data and Weight arrays are not the same size");
            }
            double product = 1.0;
            double wtotal = 0.0;
            for (int i = 0; i < x.Length; i++)
            {
                product += weights[i] * Math.Log(x[i]);
                wtotal += weights[i];
            }
            return Math.Exp(product / wtotal);
        }

        public static double HarmonicMean(double[] x)
        {
            double total = 0.0d;
            for (int i = 0; i < x.Length; i++)
            {
                total += 1.0d / x[i];
            }
            return x.Length / total;
        }

        public static double WeighedHarmonicMean(double[] x, double[] weights)
        {
            if (x.Length != weights.Length)
            {
                throw new Exception("Data and Weight arrays are not the same size");
            }
            double xtotal = 0.0d;
            double ytotal = 0.0d;
            for (int i = 0; i < x.Length; i++)
            {
                xtotal += (weights[i] / x[i]);
                ytotal += weights[i];
            }
            return ytotal / xtotal;
        }

        public static double TruncatedMean(double[] x, double trcValue)
        {
            int remove = (int)Math.Floor((trcValue * x.Length) / 2.0);
            if (remove == 0)
            {
                return Mean(x);
            }
            if (remove == x.Length)
            {
                return Median(x);
            }
            double[] sorted = Sort(x, SortOrder.Ascending);
            double sum = 0.0d;
            int count = 0;
            for (int i = remove; i < (sorted.Length - remove); i++)
            {
                sum += sorted[i];
                count++;
            }
            return sum / count;
        }

        public static double RMS(double[] x)
        {
            double sum = 0.0d;
            for (int i = 0; i < x.Length; i++)
            {
                sum += (x[i] * x[i]);
            }
            return Math.Sqrt(sum / x.Length);
        }

        public static double Range(double[] x)
        {
            double[] sorted = Sort(x);
            return (sorted[sorted.Length - 1] - sorted[0]);
        }

        public static double MidRange(double[] x)
        {
            double[] sorted = Sort(x);
            return 0.5 * (sorted[sorted.Length - 1] + sorted[0]);
        }

        public static double Median(double[] x)
        {
            double[] sorted = Sort(x);
            double position = 0.5 * (sorted.Length - 1);
            int lower = (int)System.Math.Floor(position);
            int upper = (int)System.Math.Ceiling(position);
            if (lower == upper)
            {
                return sorted[lower];
            }
            return sorted[lower] + ((position - lower) * (sorted[upper] - sorted[lower]));
        }

        public static double Mode(double[] x)
        {
            int[] counts = new int[x.Length];
            double[] temp = new double[x.Length];
            int index = 0;
            for (int i = 0; i < x.Length; i++)
            {
                int found = Array.IndexOf(temp, x[i], 0, index);
                if (found >= 0)
                {
                    // already seen
                    counts[found]++;
                }
                else
                {
                    // new value
                    temp[index] = x[i];
                    counts[index] = 1;
                    index++;
                }
            }
            int maxValue = 0;
            int maxIndex = 0;
            for (int i = 0; i < index; i++)
            {
                if (counts[i] > maxValue)
                {
                    maxValue = counts[i];
                    maxIndex = i;
                }
            }
            return temp[maxIndex];
        }

        public static double MeanDeviation(double[] x)
        {
            double sum = 0.0d;
            for (int i = 0; i < x.Length; i++)
            {
                sum += x[i];
            }
            double mean = sum / x.Length;
            sum = 0.0d;
            for (int i = 0; i < x.Length; i++)
            {
                sum += Math.Abs(x[i] - mean);
            }
            return (sum / x.Length);
        }

        public static double MedianDeviationOfTheMean(double[] x)
        {
            double sum = 0.0d;
            for (int i = 0; i < x.Length; i++)
            {
                sum += x[i];
            }
            double mean = sum / x.Length;
            double[] means = new double[x.Length];
            for (int i = 0; i < means.Length; i++)
            {
                means[i] = Math.Abs(x[i] - mean);
            }
            return Median(means);
        }

        public static double MedianDeviationOfTheMedian(double[] x)
        {
            double median = Median(x);
            double[] medians = new double[x.Length];
            for (int i = 0; i < medians.Length; i++)
            {
                medians[i] = Math.Abs(x[i] - median);
            }
            return Median(medians);
        }

        enum BiasType { Biased, Unbiased };
        static double Variance(double[] data, BiasType type)
        {
            double sum = 0.0d;
            for (int i = 0; i < data.Length; i++)
            {
                sum += data[i];
            }
            double mean = sum / data.Length;
            sum = 0.0d;
            for (int i = 0; i < data.Length; i++)
            {
                sum += Math.Pow((data[i] - mean), 2);
            }
            double variance;
            if (type == BiasType.Biased)
            {
                variance = sum / data.Length;
            }
            else
            {
                variance = sum / (data.Length - 1);
            }
            return variance;
        }

        public static double StandardDeviation(double[] data)
        {
            return Math.Sqrt(Variance(data, BiasType.Unbiased));
        }

        static double StandardDeviation(double[] data, BiasType btype)
        {
            return Math.Sqrt(Variance(data, btype));
        }

        public static double CentralMoment(double[] data, int k)
        {
            double sum = 0.0d;
            for (int i = 0; i < data.Length; i++)
            {
                sum += data[i];
            }
            double mean = sum / data.Length;
            sum = 0.0d;
            for (int i = 0; i < data.Length; i++)
            {
                sum += Math.Pow((data[i] - mean), k);
            }
            return sum / data.Length; ;
        }

        static double Skewness(double[] data, BiasType type)
        {
            if (data.Length < 3)
            {
                throw new Exception("Skewness requires at least three elements.");
            }
            // Calculate Mean
            double sum = 0.0d;
            for (int i = 0; i < data.Length; i++)
            {
                sum += data[i];
            }
            double mean = sum / data.Length;
            // Third and second moments
            double second = 0.0d;
            double third = 0.0d;
            double deviation;
            for (int i = 0; i < data.Length; i++)
            {
                deviation = data[i] - mean;
                second += (deviation * deviation);
                third += Math.Pow(deviation, 3);
            }
            // Standardized moment
            double variance;
            if (type == BiasType.Biased)
            {
                variance = second / data.Length;
            }
            else
            {
                variance = second / (data.Length - 1);
            }
            double standardizedMoment = third / Math.Pow(Math.Sqrt(variance), 3);
            // Skewness
            double d = (double)data.Length;
            double skewness;
            if (type == BiasType.Unbiased)
            {
                skewness = (d / ((d - 1) * (d - 2))) * standardizedMoment;
            }
            else
            {
                skewness = standardizedMoment / d;
            }
            return skewness;
        }

        static double Kurtosis(double[] data, BiasType type)
        {
            if (data.Length < 4)
            {
                throw new Exception("Kurtosis requires at least four elements.");
            }
            // Mean
            double sum = 0.0d;
            for (int i = 0; i < data.Length; i++)
            {
                sum += data[i];
            }
            double mean = sum / data.Length;

            // Fourth and second moments
            double second = 0.0d;
            double fourth = 0.0d;
            double deviation;
            for (int i = 0; i < data.Length; i++)
            {
                deviation = data[i] - mean;
                second += (deviation * deviation);
                fourth += Math.Pow(deviation, 4);
            }

            // Standardized moment
            double variance;
            if (type == BiasType.Biased)
            {
                variance = second / data.Length;
            }
            else
            {
                variance = second / (data.Length - 1);
            }
            double standardizedMoment = fourth / (variance * variance);

            // Kurtosis
            double d = (double)data.Length;
            double kurtosis;
            if (type == BiasType.Unbiased)
            {
                double term1 = d * (d + 1);
                double term2 = (d - 1) * (d - 2) * (d - 3);
                double term3 = (3 * (d - 1) * (d - 1)) / ((d - 2) * (d - 3));
                kurtosis = ((term1 / term2) * standardizedMoment) - term3;
            }
            else
            {
                kurtosis = (standardizedMoment / d) - 3;
            }
            return kurtosis;
        }

        static double Covariance(double[] data1, double[] data2, BiasType type)
        {
            if (data1.Length != data2.Length)
            {
                throw new Exception("The two data sets being tested are of different sizes");
            }
            if (data1.Length == 0)
            {
                throw new Exception("Data in covariance method has zero length.");
            }

            // Sum
            double sum1 = 0.0d;
            double sum2 = 0.0d;
            for (int i = 0; i < data1.Length; i++)
            {
                sum1 += data1[i];
                sum2 += data2[i];
            }

            // Mean
            double mean1 = sum1 / data1.Length;
            double mean2 = sum2 / data2.Length;

            // Covariance
            double total = 0.0d;
            for (int i = 0; i < data1.Length; i++)
            {
                total += ((data1[i] - mean1) * (data2[i] - mean2));
            }
            if (type == BiasType.Biased)
            {
                return total / data1.Length;
            }
            else
            {
                return total / (data1.Length - 1);
            }
        }

        static double Correlation(double[] data1, double[] data2)
        {
            if (data1.Length != data2.Length)
            {
                throw new Exception("The two data sets being tested are of different sizes");
            }
            if (data1.Length == 0)
            {
                throw new Exception("Data in correlation method has zero length.");
            }
            if (data1.Length == 1)
            {
                throw new Exception("Single data point is insufficient for correlation.");
            }
            // Sum
            double sum1 = 0.0d;
            double sum2 = 0.0d;
            for (int i = 0; i < data1.Length; i++)
            {
                sum1 += data1[i];
                sum2 += data2[i];
            }

            // Mean
            double mean1 = sum1 / data1.Length;
            double mean2 = sum2 / data2.Length;

            // Covariance
            double total = 0.0d;
            for (int i = 0; i < data1.Length; i++)
            {
                total += ((data1[i] - mean1) * (data2[i] - mean2));
            }
            double covariance = total / data1.Length;

            // Standard deviation
            sum1 = 0.0d;
            sum2 = 0.0d;
            for (int i = 0; i < data1.Length; i++)
            {
                sum1 += ((data1[i] - mean1) * (data1[i] - mean1));
                sum2 += ((data2[i] - mean2) * (data2[i] - mean2));
            }
            double stdev1 = Math.Sqrt(sum1 / data1.Length);
            double stdev2 = Math.Sqrt(sum2 / data2.Length);

            if ((stdev1 * stdev2) == 0)
            {
                throw new Exception("One of the standard deviations is zero.");
            }
            return (covariance / (stdev1 * stdev2));
        }

        //Calculates the minimum value in the given data set.
        public static double MinValue(double[] data)
        {
            double minimum = data[0];
            double d;

            for (int i = 1; i < data.Length; i++)
            {
                d = data[i];
                if (d < minimum)
                {
                    minimum = d;
                }
            }
            return minimum;
        }

        //Calculates the minimum absolute value in the given data set.
        public static double MinAbsValue(double[] data)
        {
            double minimum = Math.Abs(data[0]);
            double d;

            for (int i = 1; i < data.Length; i++)
            {
                d = Math.Abs(data[i]);
                if (d < minimum)
                {
                    minimum = d;
                }
            }
            return minimum;
        }

        //Calculates the index of the minimum value in the given data set.
        public static int MinIndex(double[] data)
        {
            double minimum = Double.MaxValue;
            int index = -1;
            double d;

            for (int i = 0; i < data.Length; i++)
            {
                d = data[i];
                if (d < minimum)
                {
                    index = i;
                    minimum = d;
                }
            }
            return index;
        }

        //Calculates the index of the minimum absolute value in the given data set.
        public static int MinAbsIndex(double[] data)
        {
            double minimum = Double.MaxValue;
            int index = -1;
            double d;

            for (int i = 0; i < data.Length; i++)
            {
                d = Math.Abs(data[i]);
                if (d < minimum)
                {
                    index = i;
                    minimum = d;
                }
            }
            return index;
        }

        //Calculates the maximum value in the given data set.
        public static double MaxValue(double[] data)
        {
            double maximum = data[0];
            double d;

            for (int i = 1; i < data.Length; i++)
            {
                d = data[i];
                if (d > maximum)
                {
                    maximum = d;
                }
            }
            return maximum;
        }

        //Calculates the maximum absolute value in the given data set.
        public static double MaxAbsValue(double[] data)
        {
            double maximum = Math.Abs(data[0]);
            double d;

            for (int i = 1; i < data.Length; i++)
            {
                d = Math.Abs(data[i]);
                if (d > maximum)
                {
                    maximum = d;
                }
            }
            return maximum;
        }

        //Calculates the index of the maximum value in the given data set.
        public static int MaxIndex(double[] data)
        {
            double maximum = Double.MinValue;
            int index = -1;
            double d;

            for (int i = 0; i < data.Length; i++)
            {
                d = data[i];
                if (d > maximum)
                {
                    index = i;
                    maximum = d;
                }
            }
            return index;
        }

        //Calculates the index of the maximum absolute value in the given data set.
        public static int MaxAbsIndex(double[] data)
        {
            double maximum = Double.MinValue;
            int index = -1;
            double d;

            for (int i = 0; i < data.Length; i++)
            {
                d = Math.Abs(data[i]);
                if (d > maximum)
                {
                    index = i;
                    maximum = d;
                }
            }
            return index;
        }

        //Returns the number of elements contained in a data set.
        public static int Count(double[] data)
        {
            return data.Length;
        }

        //Calculates the sum of the elements in the given data set.
        public static double Sum(double[] data)
        {
            double sum = 0.0d;
            for (int i = 0; i < data.Length; i++)
            {
                sum += data[i];
            }
            return sum;
        }

        public static double Percentile(double[] data, double n)
        {
            if ((n < 0) || (n > 1))
            {
                throw new Exception("Percentile must be between zero and one");
            }
            if (n == 0)
            {
                return MinValue(data);
            }
            if (n == 1.0)
            {
                return MaxValue(data);
            }
            double[] sorted = Sort(data);
            double position = n * (sorted.Length - 1);
            int lower = (int)Math.Floor(position);
            int upper = (int)Math.Ceiling(position);
            if (lower == upper)
            {
                return sorted[lower];
            }
            return sorted[lower] + ((position - lower) * (sorted[upper] - sorted[lower]));
        }

        public static double PercentileRank(double[] data, double val)
        {
            double[] sorted = Sort(data);
            if (val < sorted[0])
            {
                return 0.0;
            }
            if (val > sorted[sorted.Length - 1])
            {
                return 1.0;
            }
            int[] bounds = Bounds(sorted, 0, sorted.Length - 1, val);
            int lower = bounds[0];
            int upper = bounds[1];
            if (val == sorted[lower])
            {
                return ((double)lower) / (sorted.Length - 1);
            }
            else if (val == sorted[upper])
            {
                return (lower + 1.0d) / (sorted.Length - 1);
            }
            //Interpolate between bound percentiles.
            return (lower + ((val - sorted[lower]) / (sorted[upper] - sorted[lower]))) / (sorted.Length - 1);
        }

        // Assumes sorted ascending.  
        private static int[] Bounds(double[] data, int lower, int upper, double val)
        {
            if (upper == (lower + 1))
            {
                int[] bounds = new int[2];
                bounds[0] = lower;
                bounds[1] = upper;
                return bounds;
            }
            int mid = (lower + upper) / 2;
            if (val <= data[mid])
            {
                return Bounds(data, lower, mid, val);
            }
            else
            {
                return Bounds(data, mid, upper, val);
            }
        }

        #endregion
       
        static void Main(string[] args)
        {
            Console.WriteLine("\n\nStatistical Tests for Chapter 13");
            Console.WriteLine("\nInput Data: 20 random numbers between 0 and 10\n");

            int nPoints = 20;
            double magnitude = 10.0;
            double[] ydata = new double[nPoints];
            double[] xdata = new double[nPoints];

            for (int i = 0; i < ydata.Length; i++)
            {
                ydata[i] = magnitude * randObj.NextDouble();
                xdata[i] = ydata[i];
                Console.WriteLine("{0}", Math.Round(ydata[i],2));
            }

            double[] weights = new double[] { 1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 8.0, 9.0, 10.0, 
                                 11.0, 12.0, 13.0, 14.0, 15.0, 16.0, 17.0, 18.0, 19.0, 20.0 };

            Console.WriteLine("\n\nStatistical Tests Summary\n");

            Console.WriteLine("Mean = {0}", Mean(ydata));
            Console.WriteLine("Weighted Mean = {0}", WeightedMean(ydata, weights));
            Console.WriteLine("Geometric Mean = {0}", GeometricMean(ydata));
            Console.WriteLine("Weighed Geometric Mean = {0}", WeighedGeometricMean(ydata, weights));
            Console.WriteLine("Harmonic Mean = {0}", HarmonicMean(ydata));
            Console.WriteLine("Weighed Harmonic Mean = {0}", WeighedHarmonicMean(ydata, weights));
            Console.WriteLine("Truncated Mean = {0}", TruncatedMean(ydata, 2.0));
            Console.WriteLine("RMS = {0}", RMS(ydata));
            Console.WriteLine("Range = {0}", Range(ydata));
            Console.WriteLine("Mid Range = {0}", MidRange(ydata));
            Console.WriteLine("Median = {0}", Median(ydata));
            Console.WriteLine("Mode = {0}", Mode(ydata));
            Console.WriteLine("Mean Deviation = {0}", MeanDeviation(ydata));
            Console.WriteLine("Median Deviation Of The Mean = {0}", MedianDeviationOfTheMean(ydata));
            Console.WriteLine("Median Deviation Of The Median = {0}", MedianDeviationOfTheMedian(ydata));
            Console.WriteLine("Variance (biased) = {0}", Variance(ydata, BiasType.Biased));
            Console.WriteLine("Variance (unbiased) = {0}", Variance(ydata, BiasType.Unbiased));
            Console.WriteLine("Standard Deviation = {0}", StandardDeviation(ydata));
            Console.WriteLine("Central Moment = {0}", CentralMoment(ydata, 2));
            Console.WriteLine("Skewness (biased) = {0}", Skewness(ydata, BiasType.Biased));
            Console.WriteLine("Skewness (unbiased) = {0}", Skewness(ydata, BiasType.Unbiased));
            Console.WriteLine("Kurtosis (biased) = {0}", Kurtosis(ydata, BiasType.Biased));
            Console.WriteLine("Kurtosis (unbiased) = {0}", Kurtosis(ydata, BiasType.Unbiased));
            Console.WriteLine("Covariance (biased) = {0}", Covariance(ydata, xdata, BiasType.Biased));
            Console.WriteLine("Covariance (unbiased) = {0}", Covariance(ydata, xdata, BiasType.Unbiased));
            Console.WriteLine("Correlation = {0}", Correlation(ydata, ydata));
            Console.WriteLine("MinValue = {0}", MinValue(ydata));
            Console.WriteLine("MinAbsValue = {0}", MinAbsValue(ydata));
            Console.WriteLine("MinIndex = {0}", MinIndex(ydata));
            Console.WriteLine("MinAbsIndex = {0}", MinAbsIndex(ydata));
            Console.WriteLine("MaxValue = {0}", MaxValue(ydata));
            Console.WriteLine("MaxAbsValue = {0}", MaxAbsValue(ydata));
            Console.WriteLine("MaxIndex = {0}", MaxIndex(ydata));
            Console.WriteLine("MaxAbsIndex = {0}", MaxAbsIndex(ydata));
            Console.WriteLine("Count = {0}", Count(ydata));
            Console.WriteLine("Sum = {0}", Sum(ydata));
            Console.WriteLine("Percentile = {0}", Percentile(ydata, 0.25));
            Console.WriteLine("Percentile Rank = {0}", PercentileRank(ydata, 10.0));
            Console.WriteLine("\nPress ENTER key to continue...");
            Console.ReadLine();
        }
    }
}
