﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace Chapter14_CodeExamples
{
    class Program
    {
        #region Combination Methods

        public class combinationlex
        {
            private long n = 0;
            private long k = 0;
            private long[] combarray = null;

            //Accessors to private variables
            public long N
            {
                get { return n; }
                set { n = value; }
            }

            public long K
            {
                get { return k; }
                set { k = value; }
            }

            public long[] combArray
            {
                get { return combarray; }
                set { combarray = value; }
            }

            //Constructor
            public combinationlex(long n, long k)
            {
                // Both n and k must be positive
                if (n < 0 || k < 0)
                    throw new Exception("Negative argument in constructor");
                // Assign n, k and combarray their initial default values
                this.n = n;
                this.k = k;
                this.combArray = new long[k];
                for (long i = 0; i < k; ++i)
                    this.combarray[i] = i;
            }

            public double Chose(int n, int k)
            {
                if ((n < 0 || k < 0) || (n < k))
                    throw new Exception("Input value must be > 0");

                //Cleans up round off error for small values of n and k
                return Math.Floor(0.5 + (Math.Exp(FactorialLn(n) -
                           FactorialLn(k) - FactorialLn(n - k))));
            }

            public combinationlex next()
            {
                //Check to see if you are at the last combination element
                //and if so, return a null
                if (this.combArray[0] == this.n - this.k) return null;

                //Create a combination object to hold the result
                combinationlex result = new combinationlex(this.n, this.k);

                //Copy the combination array data to the combination
                //object that holds the result
                for (long i = 0; i < this.k; ++i)
                    result.combArray[i] = this.combArray[i];

                //Find the rightmost atom to be incremented
                //Use an index to start at the last rightmost position within 
                //the array and work towards the left (decrementing) until 
                //you find a false result to the condition 
                //result.combArray[rIndex] == this.n - this.k + rIndex
                //or hit the start of the array
                long rIndex;
                for (rIndex = this.k - 1; rIndex > 0 &&
                    result.combArray[rIndex] == this.n - this.k + rIndex; --rIndex) ;

                //Increment atom at that position
                ++result.combArray[rIndex];

                //Then increment every atom to the right of that atom as well
                for (long j = rIndex; j < this.k - 1; ++j)
                    result.combArray[j + 1] = result.combArray[j] + 1;

                //and return the result
                return result;
            }
        }

        public static double CombinationChose(int n, int k)
        {
            if ((n < 0 || k < 0) || (n < k))
                throw new Exception("Input value must be > 0");

            //Floor function cleans up round off error for small values of n and k
            return Math.Floor(0.5 + (Math.Exp(FactorialLn(n) - FactorialLn(k) - FactorialLn(n - k))));
        }

        #endregion

        #region Permutation Methods

        public static void Permutation(int N)
        {
            int[] p = new int[N];      // permutation
            int[] pi = new int[N];     // inverse permutation
            int[] dir = new int[N];    // direction = +1 or -1
            for (int i = 0; i < N; i++)
            {
                dir[i] = -1;
                p[i] = i;
                pi[i] = i;
            }
            Permutation(0, p, pi, dir);
        }

        public static void Permutation(int n, int[] p, int[] pi, int[] dir)
        {
            // base case - print out permutation
            if (n >= p.Length)
            {
                for (int i = 0; i < p.Length; i++)
                    Console.Write(p[i]);
                Console.Write("\n");
                return;
            }

            // swap 
            Permutation(n + 1, p, pi, dir);
            for (int i = 0; i <= n - 1; i++)
            {
                int z = p[pi[n] + dir[n]];
                p[pi[n]] = z;
                p[pi[n] + dir[n]] = n;
                pi[z] = pi[n];
                pi[n] = pi[n] + dir[n];
                Permutation(n + 1, p, pi, dir);
            }
            dir[n] = -dir[n];
        }

        public static double Permutation(int n, int r)
        {
            if ((n < 0 || r < 0) || (n < r))
                throw new Exception("Input value must be > 0");

            //Floor function cleans up round off error for small values of n and k
            return Math.Floor(0.5 + (Math.Exp(FactorialLn(n) - FactorialLn(n - r))));
        }

        #endregion

        #region Factorial Methods

        public static double Factorial(double number)
        {
            double Result;
            if (number < 0)
            {
                throw new Exception("Input value must be > 0");
            }
            if (number == 0.0) Result = 1; //0! = 1
            else
                //Recursively call the Factorial function
                Result = (number * Factorial(number - 1));
            return Result;
        }

        private static readonly double[] factorialLookup = 
        { 
            1.0,
            1.0,
            2.0,
            6.0,
            24.0,
            120.0,
            720.0,
            5040.0, 
            40320.0,
            362880.0,
            3628800.0,
            39916800.0,
            479001600.0,
            6227020800.0,
            87178291200.0,
            1307674368000.0,
            20922789888000.0,
            355687428096000.0,
            6402373705728000.0,
            121645100408832000.0,
            2432902008176640000.0,
            51090942171709440000.0,
            1124000727777607680000.0,
            25852016738884976640000.0
        };

        //Calculates n! using a lookup table up to n < 24
        //For n > 24, it calculates n! = exp(Ln(n!))
        public static double Factorial(int n)
        {
            if (n < 0)
            {
                throw new Exception("Input value must be a > 0");
            }
            else if (n < 24)
            {
                return factorialLookup[n];
            }
            else
            {
                return Math.Floor(System.Math.Exp(FactorialLn(n)) + 0.5);
            }
        }

        public static double FactorialLn(int n)
        {
            if (n < 0)
            {
                throw new Exception("Input value must be > 0");
            }
            else
            {
                return GammaLn(n + 1.0);
            }
        }

	    #endregion

        #region Gamma Methods

	    public static double GammaLn(double x)
        {
            if (x <= 0) throw
            new Exception("Input value must be > 0");

            double[] coef = new double[14]
            {57.1562356658629235,
            -59.5979603554754912,
            14.1360979747417471,
            -0.491913816097620199,
             0.339946499848118887E-4,
             0.465236289270485756E-4,
            -0.983744753048795646E-4,
             0.158088703224912494E-3,
            -0.210264441724104883E-3,
             0.217439618115212643E-3,
            -0.164318106536763890E-3,
             0.844182239838527433E-4,
            -0.261908384015814087E-4,
             0.368991826595316234E-5};

            double denominator = x;
            double series = 0.999999999999997092;
            double temp = x + 5.24218750000000000;
            temp = (x + 0.5) * Math.Log(temp) - temp;
            for (int j = 0; j < 14; j++)
                series += coef[j] / ++denominator;
            return (temp + Math.Log(2.5066282746310005 * series / x));
        }

        public static double Gamma(double x)
        {
            if (x <= 0) throw
                new Exception("Input value must be > 0");
            return Math.Exp(GammaLn(x));
        }

	    #endregion

        #region Beta Methods

		public static double Beta(double x, double y)
        {
            if ((x <= 0) || (y <= 0)) throw
            new Exception("Input values must both be > 0");

            return Math.Exp(GammaLn(x) + GammaLn(y) - GammaLn(x + y));
        } 

	    #endregion

        #region Erf and Erfc Methods

        static double erf(double x)
        {
            if (x >= 0)
                return 1.0 - ErfcCheb(x);
            else
                return ErfcCheb(-x) - 1.0;
        }

        static double erfc(double x)
        {
            if (x >= 0)
                return ErfcCheb(x);
            else
                return 2.0 - ErfcCheb(-x);
        }

        private static double ErfcCheb(double x)
        {
            int j;
            double tmp;
            double d = 0.0;
            double dd = 0.0;
            const int ncoef = 28;

            if (x < 0.0) throw new Exception("ErfcCheb requires nonnegative argument");

            double[] coef = new double[ncoef]{-1.3026537197817094, 
            6.4196979235649026e-1, 1.9476473204185836e-2,
            -9.561514786808631e-3, -9.46595344482036e-4,
            3.66839497852761e-4, 4.2523324806907e-5,
            -2.0278578112534e-5, -1.624290004647e-6,
            1.303655835580e-6, 1.5626441722e-8, -8.5238095915e-8,
            6.529054439e-9, 5.059343495e-9, -9.91364156e-10,
            -2.27365122e-10, 9.6467911e-11, 2.394038e-12,
            -6.886027e-12, 8.94487e-13, 3.13092e-13,
            -1.12708e-13, 3.81e-16, 7.106e-15, -1.523e-15,
            -9.4e-17, 1.21e-16, -2.8e-17};

            double t = 2.0 / (2.0 + x);
            double ty = 4.0 * t - 2.0;

            for (j = ncoef - 1; j > 0; j--)
            {
                tmp = d;
                d = ty * d - dd + coef[j];
                dd = tmp;
            }
            return t * Math.Exp(-x * x + 0.5 * (coef[0] + ty * d) - dd);
        }
		
	    #endregion

        #region Si and Ci Methods

        public static double Si(double x)
        {
            double sum = 0.0;
            double t = 0.0;
            const double epsilon = 1.0E-10;
            int n = 0;
            do
            {
                t = Math.Pow(-1, n) * Math.Pow(x, 2 * n + 1) / (2 * n + 1) / Gamma(2 * n + 2);
                sum += t;
                n++;
            }
            while (Math.Abs(t) > epsilon);
            return sum;
        }

        public static double Ci(double x)
        {
            double sum = 0.0;
            double t = 0.0;
            const double epsilon = 1.0E-10;
            int n = 1;
            do
            {
                t = Math.Pow(-1, n) * Math.Pow(x, 2 * n) / (2 * n) / Gamma(2 * n + 1);
                sum += t;
                n++;
            }
            while (Math.Abs(t) > epsilon);
            return 0.57721566490153286060 + Math.Log(x) + sum;
        }

	    #endregion

        #region Laguerre Polynomials

        public static double Laguerre(double x, int deg)
        {
            double L0 = 1.0;
            double L1 = 1.0 - x;
            double L2 = (x * x - 4 * x + 2.0) / 2.0;
            int n = 1;
            if (deg < 0)
                throw new Exception("Bad Laguerre polynomial: deg < 0");
            if (deg == 0)
                return L0;
            else if (deg == 1)
                return L1;
            else
            {
                while (n < deg)
                {
                    L2 = ((2.0 * n + 1.0 - x) * L1 - n * L0) / (n + 1);
                    L0 = L1;
                    L1 = L2;
                    n++;
                }
                return L2;
            }
        }

        #endregion

        #region Hermite Polynomials
		 
        public static double Hermite(double x, int deg)
        {
            double H0 = 1.0;
            double H1 = 2 * x;
            double H2 = 4 * x * x - 2;
            int n = 1;
            if (deg < 0)
                throw new Exception("Bad Hermite polynomial: deg < 0");
            if (deg == 0)
                return H0;
            else if (deg == 1)
                return H1;
            else
            {
                while (n < deg)
                {
                    H2 = 2.0 * x * H1 - 2.0 * n * H0;
                    H0 = H1;
                    H1 = H2;
                    n++;
                }
                return H2;
            }
        }

	    #endregion

        #region Legendre Polynomials
		 
        public static double Legendre(double x, int deg)
        {
            double P0 = 1.0;
            double P1 = x;
            double P2 = (3.0 * x * x - 1) / 2.0;
            int n = 1;
            if (deg < 0)
                throw new Exception("Bad Hermite polynomial: deg < 0");
            if (deg == 0)
                return P0;
            else if (deg == 1)
                return P1;
            else
            {
                while (n < deg)
                {
                    P2 = 2.0 * x * P1 - P0 - (x * P1 - P0) / (deg + 1);
                    P0 = P1;
                    P1 = P2;
                    n++;
                }
                return P2;
            }
        }

	    #endregion

        #region Chebyshev Polynomials
		 
        public static double ChebyshevT(double x, int deg)
        {
            double T0 = 1.0;
            double T1 = x;
            double T2 = 2.0 * x * x - 1.0;
            int n = 1;
            if (deg < 0)
                throw new Exception("Bad Hermite polynomial: deg < 0");
            if (deg == 0)
                return T0;
            else if (deg == 1)
                return T1;
            else
            {
                while (n < deg)
                {
                    T2 = 2.0 * x * T1 - T0;
                    T0 = T1;
                    T1 = T2;
                    n++;
                }
                return T2;
            }
        }

        public static double ChebyshevU(double x, int deg)
        {
            double U0 = 1.0;
            double U1 = 2.0 * x;
            double U2 = 4 * x * x - 1.0;
            int n = 1;
            if (deg < 0)
                throw new Exception("Bad Hermite polynomial: deg < 0");
            if (deg == 0)
                return U0;
            else if (deg == 1)
                return U1;
            else
            {
                while (n < deg)
                {
                    U2 = 2.0 * x * U1 - U0;
                    U0 = U1;
                    U1 = U2;
                    n++;
                }
                return U2;
            }
        }

	    #endregion

        #region Bessel Functions
		
        public static double BesselJ(double x, double a)
        {
            double sum = 0.0;
            double t = 0.0;
            const double epsilon = 1.0E-10;
            int n = 0;
            do
            {
                t = Math.Pow(-1, n) * Math.Pow(0.5 * x, 2 * n + a) / Gamma(n + 1) / Gamma(n + a + 1);
                sum += t;
                n++;
            }
            while (Math.Abs(t) > epsilon);
            return sum;
        }

        public static double BesselY(double x, double a)
        {
            return (BesselJ(x, a) * Math.Cos(a * Math.PI) - BesselJ(x, -a)) / Math.Sin(a * Math.PI);
        }

        public static double Besselj(double x, double a)
        {
            if (x == 0.0)
                throw new Exception("Attempted division by zero.");

            return Math.Sqrt(Math.PI / 2.0 / x) * BesselJ(x, a + 0.5);
        }

        public static double Bessely(double x, double a)
        {
            if (x == 0.0)
                throw new Exception("Attempted division by zero.");

            return Math.Sqrt(Math.PI / 2.0 / x) * BesselY(x, a + 0.5);
        }
 
	    #endregion


        static void Main(string[] args)
        {
            Console.WriteLine("Combination & Permutation Tests\n");

            string[] dataArray = new string[] { "A", "B", "C", "D", "E", "F" };
            int n = dataArray.Length;
            int k = 3;

            Console.WriteLine("Input data array = {A,B,C,D,E,F}");
            Console.WriteLine("Generating all lexicographical combinations for a set of " + dataArray.Length.ToString() + " elements taken " + k.ToString() + " at a time.\n");

            //Create a combination object to hold the data
            combinationlex combObj = new combinationlex(n, k);
            //and a string array to hold the current combination results
            string[] currComb = new string[n];

            long combCount = (long)combObj.Chose(n, k);
            Console.WriteLine("There are " + combCount + " possible combinations for a set of " + dataArray.Length.ToString() + " elements taken " + k.ToString() + " at a time and those combinations are:\n");

            //Loop through all possible combinations printing out 
            //all possible combination results
            int i = 0;
            while (combObj != null)
            {
                string[] result = new string[combObj.K];
                for (long j = 0; j < result.Length; ++j)
                    currComb[j] = dataArray[combObj.combArray[j]];

                Console.Write("[" + i + "] ");
                for (long m = 0; m < result.Length; ++m)
                {
                    Console.Write(currComb[m] + " ");
                }
                Console.WriteLine();
                combObj = combObj.next();
                ++i;
            }
            Console.WriteLine("Press any key to continue...");
            Console.ReadLine();

            Console.WriteLine("\nTesting lexicographical permutation for 3 elements");
            Permutation(3);
            Console.WriteLine("\n\nTesting lexicographical permutation for 4 elements");
            Permutation(4);
            Console.WriteLine("Press any key to continue...");
            Console.ReadLine();

            try
            {
                Console.WriteLine("\n\nTesting Special Functions\n");

                Console.WriteLine("\nTesting Factorial Function (version 1)\n");
                Console.WriteLine("Factorial(5) = {0}", Factorial(5.0));
                Console.ReadLine();

                Console.WriteLine("\nTesting Factorial Function (version 2)\n");
                Console.WriteLine("Factorial(15) = {0}", Factorial(5));
                Console.ReadLine();

                Console.WriteLine("\nTesting GammaLn Function\n");
                Console.WriteLine("GammaLn(0.50) = {0}", GammaLn(0.50));
                Console.ReadLine();

                Console.WriteLine("\nTesting Gamma Function\n");
                Console.WriteLine("Gamma(5) = {0}", Gamma(5.0));
                Console.WriteLine("Gamma(1/2) = {0}", Gamma(0.5));
                Console.ReadLine();

                Console.WriteLine("\nTesting Beta Function\n");
                Console.WriteLine("Beta(3,4) = " + Beta(3, 4).ToString());
                Console.ReadLine();

                Console.WriteLine("\nTesting Erf and Erfc Function\n");
                for (int j = 0; j < 15; j++)
                {
                    double x = (j - 10.0) / 2.0;
                    Console.WriteLine("x = {0,5:n2}, erf(x) = {1,20:e12}, erfc(x) = {2:e12}", x, erf(x), erfc(x));
                }
                Console.ReadLine();

                Console.WriteLine("\n\nTesting Si and Ci Function\n");
                for (int m = 1; m < 15; m++)
                {
                    double x = m / 2.0;
                    Console.WriteLine("x = {0,5:n2}, Si(x) = {1,18:e12}, Ci(x) = {2,20:e12}", x, Si(x), Ci(x));
                }
                Console.ReadLine();

                Console.WriteLine("\n\nTesting Laguerre Function\n");
                for (int w = 0; w < 5; w++)
                {
                    double x = 1.0 * w - 0.5;
                    Console.WriteLine("x = {0,5:n2}, L20(x) = {1,20:e12}", x, Laguerre(x, 20));
                }
                Console.ReadLine();

                Console.WriteLine("\n\nTesting Hermite Function\n");
                for (int p = 0; p < 5; p++)
                {
                    double x = 1.0 * p - 0.5;
                    Console.WriteLine("x = {0,5:n2}, H10(x) = {1,20:e12}", x, Hermite(x, 10));
                }
                Console.ReadLine();

                Console.WriteLine("\n\nTesting Chebychev (first and second kind) Function\n");
                for (int t = 0; t < 5; t++)
                {
                    double x = 0.15 * (t - 5.0);
                    Console.WriteLine("x = {0,5:n2}, T15(x) = {1,20:e12}, U15(x) = {2,20:e12}", x, ChebyshevT(x, 15), ChebyshevU(x, 15));
                }
                Console.ReadLine();

                Console.WriteLine("\n\nTesting Legendre Function\n");
                for (int s = 0; s < 5; s++)
                {
                    double x = 0.15 * (s - 4.0);
                    Console.WriteLine("x = {0,5:n2}, P10(x) = {1,20:e12}", x, Legendre(x, 10));
                 }
                Console.ReadLine();

                Console.WriteLine("\n\nTesting Bessel Functions\n");
                for (int c = 1; c < 15; c++)
                {
                    double x = 1.0 * c;
                    Console.WriteLine("x = {0,2:n0}, J0(x) = {1,18:e10}, Y0(x) = {2,18:e10}", x, BesselJ(x, 0),BesselY(x, 0.0+1.0e-5));
                }
                Console.ReadLine();
            }
            catch (Exception ex)
            {
                Console.WriteLine("Fatal error: " + ex.Message);
                Console.ReadLine();
            }
        }
    }
}
